# -*- coding: utf-8 -*-
#
# Copyright 2007, 2008 Intevation GmbH, Germany, <info@intevation.de>
#
# This file is part of mpuls WASKA (CoMPUter-based case fiLeS -
# Web-Anwendungs-Server fuer Kompetenzagenturen).
#
# mpuls WASKA is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# mpuls WASKA is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public
# License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with mpuls WASKA. If not, see <http://www.gnu.org/licenses/>.
#
# mpuls WASKA has been developed on behalf of the
# Projekttraeger im Deutschen Zentrum fuer Luft- und Raumfahrt e.V. (PT-DLR)
# within the programme Kompetenzagenturen (Durchfuehrungsphase) funded by
# the Bundesministerium fuer Familie, Senioren, Frauen und Jugend and
# European Social Fund resources.
#
# Authors:
# Torsten Irländer <torsten.irlaender@intevation.de>

import logging

from formencode import ForEach
from formencode.validators import Int

from mpulsweb.lib.search import CaseSearch as CaseBaseSearch
from mpulsweb.lib.search import EvaluationSearch as EvaluationBaseSearch
from mpulsweb.lib.search import CaseCountSearch


log = logging.getLogger(__name__)


class AppMixin(object):

    VALID_OPTIONS = ("gender", "type_ending", "migration")
    MULTI_OPTIONS = ("type_ending",)

    # In wasko, the values for male and female are swapped.
    wasko_gender_mapping = {1: 0, 0: 1}

    def extra_validators(self):
        validators = super(AppMixin, self).extra_validators()
        validators["type_ending"] = ForEach(Int(), convert_to_list=True)
        validators["migration"] = Int(if_missing=-2)
        return validators

    def convert_form_parameters(self, options, user, extended):
        search = super(AppMixin, self).convert_form_parameters(options, user,
                                                        extended=extended)
        if extended:
            type_ending = options.get("type_ending")
            if type_ending:
                search["type_ending"] = list(type_ending)
            if "migration" in options:
                search["migration"] = options["migration"]

        return search

    def get_ending_clause(self, options):
        """Return the where subclause that selects based on CM ending.
        If the type_ending option was not specified, any ending will be
        OK, so this method returns 'TRUE'.  Otherwise only endings
        contained in the list value of the type_ending option will be
        selected.  If that list is empty, no ending will be OK, so no
        case will be found.
        """
        return self.generic_int_choice_clause(options, "type_ending",
                                              "m.cm_end_art")

    def get_gender_clause(self, options):
        gender = options.get("gender")
        if gender is not None:
            return ("m.geschlecht = %s"
                    % self.wasko_gender_mapping.get(gender, gender))
        return "TRUE"

    def get_migration_clause(self, options):
        """Return search clause selecting cases based on migration background.
        """
        migration = options.get('migration')
        if migration is not None:
            migration_clause = ("(nat_muspra NOT IN (18, -1, -3)"
                                " OR nat_staat NOT IN (1, -1, -3)"
                                " OR vater_land NOT IN (29, -1, -3)"
                                " OR mutter_land NOT IN (29, -1, -3))")
            no_migration_clause = ("(nat_muspra = 18 AND nat_staat = 1"
                                   " AND vater_land = 29 AND mutter_land = 29)")
            if migration == 1:
                return migration_clause
            elif migration == 0:
                return no_migration_clause
            elif migration == -1:
                return ("(NOT %s AND NOT %s)"
                        % (migration_clause, no_migration_clause))
            elif migration == 2:
                return ("(nat_staat = 2"
                        " and (vater_staat != 2 or mutter_staat != 2))")
            elif migration == -2:
                return "TRUE"
            else:
                log.warning("unknown value for migration search: %d", migration)

        return "TRUE"

    def get_where_subclauses(self, options):
        return super(AppMixin, self).get_where_subclauses(options) + [
            self.get_ending_clause(options),
            self.get_migration_clause(options),
            ]


class CaseSearch(AppMixin, CaseBaseSearch):

    VALID_OPTIONS = (CaseBaseSearch.VALID_OPTIONS + AppMixin.VALID_OPTIONS)

    MULTI_OPTIONS = (CaseBaseSearch.MULTI_OPTIONS + AppMixin.MULTI_OPTIONS)


class EvaluationSearch(EvaluationBaseSearch):

    VALID_OPTIONS = (EvaluationBaseSearch.VALID_OPTIONS
                     + AppMixin.VALID_OPTIONS)

    MULTI_OPTIONS = (EvaluationBaseSearch.MULTI_OPTIONS
                     + AppMixin.MULTI_OPTIONS)


# vim:set ts=4 sw=4 si et sta sts=4 fenc=utf8:
