# -*- coding: utf-8 -*-
from datetime import datetime
import logging
import paste
import formencode

from pylons import request, tmpl_context as c, app_globals as g, url
from pylons.controllers.util import redirect

from formed.instance.tree import SetDataException

from mpulsweb.lib.translation import _, ungettext
from mpulsweb.lib.base import BaseController, render, session, h
from mpulsweb.lib.security import checkRole, check_feature
from mpulsweb.lib.dialogs import checkPrivacyStatement, checkSyncPermission, checkCaseLinked
from mpulsweb.lib.validators import SetStandinForm, SetEditorForm, NewCaseForm
from mpulsweb.lib.export import FormletterExport
from mpulsweb.lib.anonymize import determineAnonymizedCaseEndPhaseName, \
     determineAnonymizedCaseEndPhase
from mpulsweb.lib.db import  PostgresDBInterface
from mpulsweb.lib.metaclient import MetaException, MetaUnauthorized, \
     UnknownMetaCase, UnknownProjectPart, MetaCaseAnonymized, MetaCasePending
from mpulsweb.model.meta import get_meta_client

from mpulsweb.model.case import ConsistenceCheckException, SessionCase, \
     MPulsXMLImporter, XMLImportException
from mpulsweb.model.user import UserGroupList, UserListObject
from mpulsweb.model.logbook import LogbookEntry
from mpulsweb.model.statement import PrivacyStatement
from mpulsweb.model.agencysettings import Agency
from mpulsweb.model.phase import PhaseFactory

from mpulsweb.controllers.formularpage import get_rendered_page
from mpulsweb.controllers.case_bundle import build_case_list

log = logging.getLogger(__name__)


class CaseController(BaseController):

    @checkRole(('admin', 'cm', 'pt_dlr'))
    def index(self):
        redirect(url(controller="case_overview", action="index"))

    @checkRole(('admin', 'cm', 'pt_dlr'))
    @checkPrivacyStatement()
    @checkSyncPermission()
    @checkCaseLinked()
    def select(self, id, confirmed=0):
        return self.digest(id)

    @checkRole(('admin', 'cm', 'pt_dlr'))
    def digest(self, id):
        id = self._checkInt(id)
        case = self._loadCase(id, read_only=True)
        c.case = case
        c.last_access = case.getState().getAccessTime()
        c.daysToMaxSavetime = case.getState().daysToMaxSavetime()
        c.maxsavetime = case.getState().getMaxSavetime()
        c.finished = case.getState().isCmFinished()
        return render('/casemanagement/digest_body.mako')

    #
    # helper methods for dealing with meta server related errors.
    #

    def _meta_error(self, case, exc, title, text, ok_url, template):
        detail = ""
        if isinstance(exc, MetaUnauthorized):
            log.error("Meta server refused request with 401 Unauthorized")
            detail = _("The meta server did not accept the request"
                       " because of missing or incorrect username or"
                       " password. Please check whether the"
                       " meta-server is configured correctly.")

        elif isinstance(exc, MetaException):
            # the error parameters should be set already
            pass

        else:
            raise RuntimeError("_meta_error not called with MetaException"
                               " instance")

        if detail:
            text += "<br>" + detail

        return self._render_error(title, text, ok_url, template)

    def _render_error(self, title, text, ok_url, template):
        c.dialog_title = title
        c.dialog_text = text
        c.url_ok = ok_url
        return render(template)

    #
    # Delete
    #

    @checkRole('cm')
    def markForDelete(self, id, confirmed):
        id = self._checkInt(id)
        confirmed = self._checkBool(confirmed)
        case = self._loadCase(id)
        if confirmed == 1:
            case.markDelete()
            case.checkValidity()
            c.dialog_title = _("Case document deleted!")
            c.dialog_text = _("The case document is given to the administrator"
                              " for deletion.")
            c.url_ok = h.url(controller="case_overview", action='index')
            session['render_mode'] = 'ro'
            session.save()
            return render('/casemanagement/dialogs/success_delete_cm.mako')
        else:
            c.context = "../main.mako"
            c.dialog_title = ungettext('Delete case?', 'Delete cases?', 1)
            c.dialog_text = _("Do you really want to mark this case for"
                              " deletion?%s The case will not be editable"
                              " afterwards, and submited to the administration"
                              " for deletion.<br>") \
                            % "\n".join(build_case_list([case]))

            c.url_yes = h.url(controller='case', action='markForDelete',
                                  id=id, confirmed=1)
            c.url_no = h.url(controller='case_overview', action='index')
            return render('/casemanagement/dialogs/confirm.mako')

    @checkRole('admin')
    def delete(self, id, confirmed):
        id = self._checkInt(id)
        confirmed = self._checkBool(confirmed)

        case = self._loadCase(id)
        if not confirmed:
            c.context = "../main.mako"
            c.dialog_title = ungettext('Delete case?', 'Delete cases?', 1)
            c.dialog_text = _("Warning! On deleting the case all data will"
                              " be erased irrecoverable!<br>"
                              "Only delete the case if this case is"
                              " <b>not needed for evaluations</b> anymore."
                              "<br><br>Do you want to delete the case "
                              "from the server?%s<br>") \
                            % "\n".join(build_case_list([case]))
            c.url_yes = h.url(controller='case', action='delete',
                                  id=id, confirmed=1)
            c.url_no = h.url(controller='case_overview')
            return render('/casemanagement/dialogs/confirm_delete.mako')

        try:
            return self._do_delete_case(case)
        except MetaException, exc:
            log.exception("Could not delete case because of a problem with the"
                          " meta server")
            return self._meta_error(case, exc,
                                    _("Case document not deleted!"),
                                    _("The case can not be deleted because"
                                      " of an unexpected problem with the"
                                      " meta server."),
                                    h.url(controller="case_overview", action='index'),
                                    '/casemanagement/dialogs'
                                    '/failed_delete.mako')
        except Exception:
            log.exception("Exception while trying to delete case %d", case.id)
            c.dialog_title = _("Case document not deleted!")
            c.dialog_text = _("The case can not be deleted because of an"
                              " unexpected problem.")
            c.url_ok = h.url(controller="case_overview", action='index')
            return render('/casemanagement/dialogs/failed_delete.mako')

    def _do_delete_case(self, case):
        """Delete the case and return a HTML success or error message.
        Derived classes may override or extend this method if they need
        to handle specific exceptions raised by the case's delete method.
        """
        case.delete()
        if g.mpuls_config.is_enabled('module', 'meta-transmit-reportquota'):
            get_meta_client().transmit_reportquota()
        c.dialog_title = _("Case document deleted!")
        c.dialog_text = _("Click on OK to continue.")
        c.url_ok = h.url(controller="case_overview", action='index')
        return render('/casemanagement/dialogs/success_delete.mako')


    #
    # Anonymize
    #
    @checkRole('cm')
    def markForAnonymizeFromOverview(self, id, confirmed):
        url = h.url(controller="case_overview", action='index')
        return self._markForAnonymize(id, confirmed, url)

    @checkRole('cm')
    def markForAnonymize(self, id, confirmed):
        check_feature('case-module', 'anonymize')
        url = h.url(controller='case', action='digest', id=id)
        return self._markForAnonymize(id, confirmed, url)

    def _markForAnonymize(self, id, confirmed, url):
        id = self._checkInt(id)
        confirmed = self._checkBool(confirmed)
        case = self._loadCase(id)

        # First check if the case is in a state which allows anonymization
        try:
            case.check_anonymizeability()
        except ConsistenceCheckException, e:
            c.dialog_title = _(u"""Anonymising not possible.""")
            c.dialog_text = h.literal(e.value)
            c.url_ok = url
            return render('/casemanagement/dialogs/failed_markanonymize.mako')

        # Show confirmation dialog if necessary.
        if not confirmed:
            # Get description of the phase which will be set after anonymisation
            phase_name = []
            pdescriptions = g.mpuls_config.get('phases', 'description')[0]
            phase = pdescriptions.get(determineAnonymizedCaseEndPhaseName(case))
            end_phase = determineAnonymizedCaseEndPhase(case)
            for ph, pairs in g.mpuls_config.get('phases',
                                                'pairs')[0].iteritems():
                if str(end_phase) in pairs:
                    phase_name.append(phase)
                    if str(end_phase) == pairs[0]:
                        phase_name.append("(%s)" % _('Running'))
                    else:
                        phase_name.append("(%s)" % _('Finished'))

            c.dialog_title = _(u"Do make anonymous?")
            c.dialog_text = (_(u"<p>Do you really want to mark the case for"
                               u" anonymization?%s<br>"
                               "The case can then no longer be"
                               u" edited and will given to the administrators"
                               u" for final anonymisation.</p>"
                               u"<p><b>The case document is going to be"
                               u" anonymised in phase </b>%s<b>!</b><br>""") \
                             % ("\n".join(build_case_list([case])),
                                " ".join(phase_name)))
            c.url_yes = h.url(controller='case', action='markForAnonymize',
                                  id=id, confirmed=1)
            c.url_no = url
            return render('/casemanagement/dialogs/confirm_markanonymize.mako')

        # the action is confirmed, so call attempt to mark the case for
        # deletion.
        try:
            return self._do_mark_anonymize_case(case)
        except (UnknownMetaCase, MetaCaseAnonymized):
            log.exception("Could not mark case %d for anonymization.", case.id)
            log.warning("Resetting meta link for case %d.", case.id)
            case.getMeta().reset()
            return self._render_error(_("Case document not marked for"
                                        " anonymization!"),
                                      _("The case has not been marked for"
                                        " anonymization because the data could"
                                        " not be uploaded to the meta server"
                                        " because the meta case has been"
                                        " deleted or anonymized."
                                        " As a result, the link to the"
                                        " meta-case has been removed."
                                        " A second attempt to anonymize the"
                                        " case should succeed."),
                                      h.url(controller="case_overview",
                                            action='index'),
                                      '/casemanagement/dialogs'
                                      '/failed_markanonymize.mako')

        except UnknownProjectPart:
            log.exception("Could not mark case %d for anonymization.", case.id)
            log.warning("Resetting project uuid for case %d.", case.id)
            case.getMeta().reset_project_uuid()
            return self._render_error(_("Case document not marked for"
                                        " anonymization!"),
                                      _("The case has not been marked for"
                                        " anonymization because the data could"
                                        " not be uploaded to the meta server"
                                        " because the project part does"
                                        " not exist anymore."
                                        " The link to project part of the"
                                        " meta-case has been removed."
                                        " A second attempt to anonymize the"
                                        " case should succeed."),
                                      h.url(controller="case_overview",
                                            action='index'),
                                      '/casemanagement/dialogs'
                                      '/failed_markanonymize.mako')

        except MetaCasePending:
            log.exception("Could not anonymize case %d.", case.id)
            return self._render_error(_("Case document not marked for"
                                        " anonymization!"),
                                      _("The case has not been marked for"
                                        " anonymization because the data could"
                                        " not be uploaded to the meta server"
                                        " because the meta-case has been"
                                        " marked for deletion or anonymization."
                                        " The case may become editable again"
                                        " later if it is restored on the"
                                        " meta-server."),
                                      h.url(controller="case_overview",
                                            action='index'),
                                      '/casemanagement/dialogs'
                                      '/failed_markanonymize.mako')

        except MetaException, exc:
            log.exception("Could not anonymize case %d because of a problem"
                          " with the meta server", case.id)
            return self._meta_error(case, exc,
                                    _("Case document not marked for"
                                      " anonymization!"),
                                    _("The case has not been marked for"
                                      " anonymization because of an"
                                      " unexpected problem with the"
                                      " meta server."),
                                    h.url(controller="case_overview",
                                          action='index'),
                                    '/casemanagement/dialogs'
                                    '/failed_markanonymize.mako')

        except Exception:
            log.exception("Exception while trying to mark case %d for"
                          " anonymisation", case.id)
            c.dialog_title = _("Case document not marked for anonymisation!")
            c.dialog_text = _("The case can not be marked for anonymisation"
                              " because of an unexpected problem.")
            c.url_ok = h.url(controller="case_overview",
                             action='index')
            return render('/casemanagement/dialogs/failed_markanonymize.mako')

    def _do_mark_anonymize_case(self, case):
        """Mark the case for anonymization and handle SetDataException.
        Derived classes may override or extend this method if they need
        to handle more exceptions.

        The method returns a rendered HTML page with the success or
        error message.
        """
        try:
            case.make_anonymizable()
            case.markAnonymize()

            c.dialog_title = _(u"Case marked for anonymisation.")
            c.dialog_text = _(u"The case document has been marked for"
                              u" anonymisation and has been given to the"
                              u" administration for final anonymisation."
                              u" Please click on OK, to get back to the"
                              u" overview.")
            c.url_ok = h.url(controller="case_overview", action='index')
            session['render_mode'] = 'ro'
            session.save()
            return render('/casemanagement/dialogs/success_markanonymize.mako')
        except SetDataException, e:
            log.exception("SetDataException while trying to mark case %d for"
                          " anonymization", case.id)
            errors = []
            for k, values in e.errors.iteritems():
                for v in values:
                    errors.append(v.msg)
            c.dialog_title = _(u"Anonymisation not possible.")
            c.dialog_text = _(u"The case could not be anonymised for the"
                              u" following reason:</p>"
                              u"<p><i>%s</i></p>"
                              u"<p>Please click on OK to geht back to the"
                              u" overview.</p>""") % "<br>".join(errors)
            c.url_ok = h.url(controller='case', action='select',
                                 id=case.id, confirmed=1)
            return render('/casemanagement/dialogs/failed_markanonymize.mako')


    @checkRole('admin')
    def anonymize(self, id, confirmed):
        check_feature('case-module', 'anonymize')
        id = self._checkInt(id)
        confirmed = self._checkBool(confirmed)
        case = self._loadCase(id)

        # Check whether the case can actually be anonymized.
        try:
            case.check_anonymizeability()
        except ConsistenceCheckException, e:
            c.dialog_title = _(u"Anonymising not possible.")
            c.dialog_text = h.literal(e.value)
            c.url_ok = h.url(controller='case', action='digest', id=id)
            return render('/casemanagement/dialogs/failed_markanonymize.mako')

        # Show confirmation dialog if necessary.
        if not confirmed:
            c.context = "../main.mako"
            c.dialog_title = _(u"Do make anonymous?")
            c.dialog_text = _(u"Do you really want to anonymise the case and"
                              u" irrecovably lose the associated personal"
                              u" data? %s"
                              u" After it the case will not be available you."
                              u" The anonymised case will still be considered"
                              u" for the analysis.<br>") \
                            % "\n".join(build_case_list([case]))
            c.url_yes = h.url(controller='case', action='anonymize',
                                  id=id, confirmed=1)
            c.url_no = h.url(controller='case', action='digest', id=id)
            return render('/casemanagement/dialogs/confirm_anonymize.mako')

        try:
            return self._do_anonymize_case(case)
        except (UnknownMetaCase, MetaCaseAnonymized):
            log.exception("Could not anonymize case %d.", case.id)
            log.warning("Resetting meta link for case %d.", case.id)
            case.getMeta().reset()
            return self._render_error(_("Case document not anonymized!"),
                                      _("The case has not been anonymized"
                                        " because the data could not be"
                                        " uploaded to the meta server one last"
                                        " time because the meta case has been"
                                        " deleted or anonymized."
                                        " As a result, the link to the"
                                        " meta-case has been removed."
                                        " A second attempt to anonymize the"
                                        " case should succeed."),
                                      h.url(controller="case_overview",
                                            action='index'),
                                      '/casemanagement/dialogs'
                                      '/failed_anonymize.mako')
        except UnknownProjectPart:
            log.exception("Could not anonymize case %d.", case.id)
            log.warning("Resetting project uuid for case %d.", case.id)
            case.getMeta().reset_project_uuid()
            return self._render_error(_("Case document not anonymized!"),
                                      _("The case has not been anonymized"
                                        " because the data could not be"
                                        " uploaded to the meta server one last"
                                        " time because the project part does"
                                        " not exist anymore."
                                        " The link to project part of the"
                                        " meta-case has been removed."
                                        " A second attempt to anonymize the"
                                        " case should succeed."),
                                      h.url(controller="case_overview",
                                            action='index'),
                                      '/casemanagement/dialogs'
                                      '/failed_anonymize.mako')

        except MetaCasePending:
            log.exception("Could not anonymize case %d.", case.id)
            return self._render_error(_("Case document not anonymized!"),
                                      _("The case has not been anonymized"
                                        " because the data could not be"
                                        " uploaded to the meta server one last"
                                        " time because the meta-case has been"
                                        " marked for deletion or anonymization."
                                        " The case may become editable again"
                                        " later if it is restored on the"
                                        " meta-server."),
                                      h.url(controller="case_overview",
                                            action='index'),
                                      '/casemanagement/dialogs'
                                      '/failed_anonymize.mako')

        except MetaException, exc:
            log.exception("Could not anonymize case %d because of a problem"
                          " with the meta server", case.id)
            return self._meta_error(case, exc,
                                    _("Case document not anonymized!"),
                                    _("The case can not be anonymized because"
                                      " of an unexpected problem with the"
                                      " meta server."),
                                    h.url(controller="case_overview",
                                          action='index'),
                                    '/casemanagement/dialogs'
                                    '/failed_anonymize.mako')

        except Exception:
            log.exception("Exception while trying to anonymize case %s", id)
            c.dialog_title = _("Case document not anonymized!")
            c.dialog_text = _("The case can not be anonymized because of an"
                              " unexpected problem.")
            c.url_ok = h.url(controller="case_overview", action='index')
            return render('/casemanagement/dialogs/failed_anonymize.mako')

    def _do_anonymize_case(self, case):
        """Anonymize the case and return a HTML success or error message.
        Derived classes may override or extend this method if they need
        to handle specific exceptions raised by the case's anonymize method.
        """
        case.anonymize()
        c.dialog_title = _(u"""Case document anonymised!""")
        c.dialog_text = _(u"The personal data of the case document has been"
                          u" deleted and the case document is removed from"
                          u" the overview. Please click on OK, to get back to"
                          u" the overview.")
        c.url_ok = h.url(controller="case_overview", action='index')
        return render('/casemanagement/dialogs/success_anonymize.mako')


    #
    # Restore
    #

    @checkRole('admin')
    def restore(self, id, confirmed):
        id = self._checkInt(id)
        confirmed = self._checkBool(confirmed)
        case = self._loadCase(id)
        if confirmed == 1:
            case.restore()
            case.checkValidity()
            c.dialog_title = ungettext(u'Case restored!', u'Cases restored!', 1)
            c.dialog_text = _(u'Please click OK to continue.')
            c.url_ok = h.url(controller="case_overview", action='index')
            return render('/casemanagement/dialogs/success_restore.mako')
        else:
            c.context = "../main.mako"
            c.dialog_title = ungettext('Restore case?', 'Restore cases?', 1)
            c.dialog_text = _("Do you really want to restore the case?%s"
                              " The case will be editable afterwards again"
                              " for the users.<br>") \
                            % "\n".join(build_case_list([case]))
            c.url_yes = h.url(controller='case', action='restore', id=id,
                                  confirmed=1)
            c.url_no = h.url(controller='case_overview', action='index')
            return render('/casemanagement/dialogs/confirm_restore.mako')

    #
    # Printing
    #
    @checkRole(('admin', 'cm', 'pt_dlr'))
    def printdigest(self, id):
        c.print_version = 1
        return self.digest(id)

    @checkRole(('admin', 'cm', 'pt_dlr'))
    def printAll(self, id, empty='0'):
        c.print_version = 1
        out = []
        case = self._loadCase(id, read_only=True)
        c.case = case
        c.phases = PhaseFactory().load(id)
        render_mode = session['render_mode']
        session['render_mode'] = 'ro'
        session.save()
        printed_headings = set()
        navigation = session.get('navigation.tree')
        if navigation:
            for item in navigation.getAllItems():
                # create the headings for all hierarchy elements that
                # have no page of their own.  We can do that by looking
                # at all the parents of the current item that have not
                # yet been used.
                headings = []
                parent = item.parent
                while parent is not None:
                    if parent not in printed_headings:
                        if parent.description:
                            # Ideally, we should be using different
                            # levels of hierarchies, but currently all
                            # headings are H1 headings
                            headings.append("<h1>%s</h1>\n"
                                            % parent.description)
                        printed_headings.add(parent)
                    parent = parent.parent
                out.extend(reversed(headings))

                out.append(get_rendered_page(item.key,
                                             case.get_formed_instance_tree()))
        c.formular = "".join(out)
        session['render_mode'] = render_mode
        session.save()
        return render('/formularpage/formular_print.mako')

    #
    # Create new case
    #

    def _createCase(self, form_result):
        case = g.case_factory.createNew(data=form_result)

        statement = case.getPrivacyStatement()
        # 3. Do we want to sign the statement?
        if form_result.get('statement_signed'):
            statement.sign()

        if g.mpuls_config.is_enabled('case-module', 'meta-server'):
            meta = case.getMeta()
            if form_result.get('sync_permission'):
                meta.allow_sync()
            if g.mpuls_config.is_enabled('module', 'meta-transmit-reportquota'):
                get_meta_client().transmit_reportquota()

        # Reload case to load navigation and set session correct. Loading
        # overhead here is not nice but not critical as creating new cases
        # should not happen so often
        return self._loadCase(case.id)

    @checkRole('cm')
    def new(self):
        session_case = SessionCase()
        session['case'] = session_case
        session.save()
        redirect(url('new2'))

    @checkRole('cm')
    def new2(self):
        c.case = session.get('case')
        if c.case is None:
            redirect(url('new'))
        else:
            form_errors = {}
            form_defaults = c.case.getData()
            form = render('/casemanagement/new.mako')
            return formencode.htmlfill.render(form, defaults=form_defaults,
                                              auto_insert_errors=False,
                                              errors=form_errors)

    def _add_custom_fields(self):
        '''This function should be overwritten in specific applications to add
        custom fields like agencyname, and fkz to a new created case. On
        default this function returns an empty dictionary.'''
        return {}

    def _get_validator(self):
        '''This function should be overwritten in specific applications to add
        a custom validator'''
        return NewCaseForm()

    def _get_xmlimporter(self, xmlfile=None):
        '''This function should be overwritten in specific applications to add
        a custom XML Importer'''
        allow_documents = g.mpuls_config.is_enabled('case-module', 'documents')
        allow_logbook = g.mpuls_config.is_enabled('case-module', 'logbook')
        return MPulsXMLImporter(g.formedTree,
                                allow_logbook=allow_logbook,
                                allow_documents=allow_documents)

    def _convert_SetDataException2Formencode(self, err):
        """Convert a SetDataException into a formencode compatible dictionary"""

        # Expample SetDataException:
        # {'geburtsdatum:17': [type: 1, name: 'geburtsdatum:17', msg:
        # u"'Geburtsdatum' darf zeitlich nicht nach 'Datum Erstgespr\xe4ch'
        # liegen.", bad:  datetime.date(1990, 1, 1)], 'datum_ea_start:17':
        # [type: 1, name: 'datum_ea_start:17', msg:  u"'Geburtsdatum' darf
        # zeitlich nicht nach 'Datum Erstgespr\xe4ch' liegen.", bad:
        # datetime.date(1981, 12, 12)]}
        errors = {}
        for e in err.errors.keys():
            key = e.split(':')[0]
            errors[key] = err.errors.get(e)[0].msg
        return errors

    @checkRole('cm')
    def newAction(self):
        # TODO: Use htmlfill here!
        session_case = SessionCase()
        session_case.setData(request.params)

        session['case'] = session_case
        session.save()

        # Do we want to print the statement?
        if request.params.has_key('print'):
            c.print_version = 1
            statement = PrivacyStatement(session_case, Agency())
            c.content = statement.fillout()
            return render('privacy/default_statement.mako')

        # Or do we want to store the new case?
        elif request.params.has_key('save'):
            validator = self._get_validator()
            try:
                form_result = validator.to_python(request.params)
                # Add custom fields to the form_result which are not part of
                # the new case formular
                form_result.update(self._add_custom_fields())
                # Create a new case
                case = self._createCase(form_result)
                return self.select(str(case.id))
            except SetDataException, err:
                c.form_result = form_result
                c.form_errors = self._convert_SetDataException2Formencode(err)
                form = render('/casemanagement/new.mako')
                return formencode.htmlfill.render(form,
                                                  defaults=c.form_result,
                                                  auto_insert_errors=False,
                                                  errors=c.form_errors)
            except formencode.Invalid, error:
                log.debug("Invalid form data: %r", error)
                c.form_result = error.value
                c.form_errors = error.error_dict or {}
                form = render('/casemanagement/new.mako')
                return formencode.htmlfill.render(form,
                                                  defaults=c.form_result,
                                                  auto_insert_errors=False,
                                                  errors=c.form_errors)

    @checkRole('cm')
    def copy(self, id, confirmed):
        id = self._checkInt(id)
        confirmed = self._checkBool(confirmed)
        case = self._loadCase(id, read_only=True)
        if confirmed:
            case.copy()
            case.checkValidity()
            c.dialog_title = _('Case copied!')
            c.dialog_text = _('The case was succussfully copied and is now'
                              ' available in the case overview.'
                              ' Click on "OK" to open the case overview.')
            c.url_ok = h.url(controller="case_overview", action='index')
            return render('/casemanagement/dialogs/success_copy.mako')
        else:
            c.dialog_title = _('Copy case?')
            c.dialog_text = _('Do you really want to copy this case?'
                              ' All data will be copied in the new case.')
            c.url_yes = h.url(controller='case', action='copy',
                                  id=id, confirmed=1)
            c.url_no = h.url(controller='case', action='digest', id=id)
            return render('/casemanagement/dialogs/confirm_copy.mako')

    @checkRole("cm")
    def importCase(self):
        return render('/casemanagement/upload.mako')

    @checkRole('cm')
    def importXMLAction(self):
        try:
            xmlfile = request.POST['xmlfile']
            try:
                importer = self._get_xmlimporter(xmlfile.file)
                importer.import_xml(PostgresDBInterface(), xmlfile.file)
            finally:
                xmlfile.file.close()

            c.dialog_title = _("Case(s) imported!""")
            c.dialog_text = _("The file was imported successfully."
                              " Click on 'OK' to open the case overview.")
            c.url_ok = h.url(controller="case_overview", action='index')
            return render('/casemanagement/dialogs/success_xml_import.mako')
        except Exception, e:
            log.exception("Exception during xml import")
            c.dialog_title = _("Case(s) not imported!")
            if isinstance(e, XMLImportException):
                c.dialog_text = h.escape(e.message)
            else:
                c.dialog_text = _("An error occurred during the import."
                                  " Click 'OK' to return to the import dialog.")
            c.url_ok = h.url(controller='case', action='importCase')
            return render('/casemanagement/dialogs/failed_xml_import.mako')

    #
    # ORGANISATION (standin and editor settings)
    #

    @checkRole(('admin', 'cm'))
    def organisation(self, id):
        id = self._checkInt(id)
        ulist = UserListObject()
        list = UserGroupList()
        c.case = self._loadCase(id, read_only=True)
        c.standin = c.case.getStandin()
        c.editor = c.case.getEditor()
        c.current_user = session['USER_AUTHORIZED']

        # group list for standin
        filter = [u.gid for u in ulist.getAdminList()]
        filter.append(c.editor.gid)
        filter2 = filter[:]
        filter.append(c.current_user.gid)
        c.group_list = list.getGroupList(filter)
        c.group_list2 = list.getGroupList(filter2)

        # userlist for editor
        filter3 = [u.id for u in ulist.getAdminList()]
        filter3.append(c.current_user.id)
        filter3.append(c.editor.id)
        c.user_list = ulist.getUserList(filter3)
        return render('/casemanagement/organisation.mako')

    @checkRole('admin')
    def setEditorAction(self):
        validator = SetEditorForm()
        try:
            form_result = validator.to_python(request.params)
            case = self._loadCase(form_result.get('case_id'))
            case.setEditor(form_result.get('editor'))
            c.status_msg = h.status_message(_(
                "Editor successfully set."))
            return self.organisation(form_result.get('case_id'))
        except formencode.Invalid, error:
            c.form_result = error.value
            c.form_errors = error.error_dict or {}
            c.status_msg = h.status_message(_("Error! Editor cannot be set."))
            return render('/casemanagement/organisation.mako')

    @checkRole(('admin', 'cm'))
    def setStandinAction(self):
        validator = SetStandinForm()
        data = formencode.variabledecode.variable_decode(request.params)
        try:
            form_result = validator.to_python(data)
            case = self._loadCase(form_result.get('case_id'))
            standin = case.getStandin()
            standin.setGroups(form_result.get('standin'))
            c.status_msg = h.status_message(_("Standin successfully set."))
            return self.organisation(form_result.get('case_id'))
        except formencode.Invalid, error:
            c.form_result = error.value
            c.form_errors = error.error_dict or {}
            c.status_msg = h.status_message(_("Error! Standin is not set."))
            return render('/casemanagement/organisation.mako')

    @checkRole('cm')
    @checkPrivacyStatement('keepActive')
    @checkSyncPermission()
    @checkCaseLinked()
    def keepActive(self, id, confirmed):
        # If logbook is enabled let the user give the reason why he keeps the
        # case active.
        if g.mpuls_config.is_enabled('case-module', 'logbook'):
            return self._keep_active_with_logbook(id)
        # Just refresh last access date.
        else:
            return self._keep_active_without_logbook(id)

    def _keep_active_with_logbook(self, id):
        entry = LogbookEntry()
        entry.setDate(datetime.now())
        entry.setKind(-2)
        form_defaults = {}
        form_defaults['case_id'] = id
        form_defaults['date'] = entry.getDate()
        form_defaults['time'] = entry.getTime()
        form_defaults['kind'] = entry.getKindAsInt()
        form_defaults['short_notice'] = _('Case storage extended')
        form_defaults['notice'] = _('Please enter the reason for the case'
                                    ' storage extension.')
        id = self._checkInt(id)
        case = self._loadCase(id)
        c.entry = entry
        c.form_errors = {}
        form = render('/logbook/new.mako')
        return formencode.htmlfill.render(form, defaults=form_defaults,
                                          auto_insert_errors=False,
                                          errors=c.form_errors)

    def _keep_active_without_logbook(self, id):
        id = self._checkInt(id)
        case = self._loadCase(id)
        try:
            # Temporarily switch the state of the case to force
            # refreshing the last access date.
            state = case.getState().getState()
            case.getState().setState(2)
            case.getState().setState(state)
            c.dialog_title = _('Case storage extended')
            c.dialog_text = _('The max case storage has been extended.'
                              ' Please click on OK to continue')
            c.url_ok = h.url("/")
            return render('/casemanagement/dialogs/success_keepactive.mako')
        except:
            log.exception('Can not refresh case in'
                          ' "_keep_active_without_logbook"')
            raise

        pass

    ## Export

    def _download_data(self, data, mime_type, extension, **extra):
        disposition = 'attachment; filename="formletter.%s"' % extension
        data = paste.fileapp.DataApp(data,
                                     content_type=mime_type,
                                     content_disposition=disposition,
                                     cache_control='must-revalidate',
                                     pragma='public',
                                     **extra)
        return data(request.environ, self.start_response)

    def formletter(self, id, letterid=1):
        id = self._checkInt(id)
        case = self._loadCase(id, read_only=True)
        d, o = case.formletter(letterid)
        export = FormletterExport([d], o).export()
        return self._download_data(export, "application/msexcel", "xls")


# vim:set ts=4 sw=4 si et sta sts=4 fenc=utf8:
