# -*- coding: utf-8 -*-
# Copyright 2012 Intevation GmbH, Germany, <info@intevation.de>
#
# This file is part of MPULS (CoMPUter-based case fiLeS).
#
# MPULS is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# MPULS is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public
# License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with MPULS. If not, see <http://www.gnu.org/licenses/>.

import re
import logging

import formencode

from pylons import app_globals as g

from mpulsweb.lib.base import c, render, request, session, BaseController
from mpulsweb.lib.security import checkRole
from mpulsweb.model.user import UserListObject
from mpulsweb.model.case import MpulsCaseOverview
from mpulsweb.model.phase import PhaseFactory 


log = logging.getLogger(__name__)


def get_sort(params):
    '''Return sort field and sort order. First try to return a userdefined sort
    and order. If user does not provide one of there params, try to return the
    safed params in session. If session is also empty return default params.
    '''
    session_options = session.get('CASE_OVERVIEW_SEARCHOPTIONS', {})
    sort = params.get('sort_field',
                      session_options.get('sort_field',
                                          g.mpuls_config.get('search',
                                                            'default-sorting')))
    order = params.get('sort_order',
                       session_options.get('sort_order',
                                           g.mpuls_config.get('search',
                                                              'default-order')))
    return sort, order

# TODO: Move function. Is only used in other controllers. May be better placed
# in Model phase or helpers?
def default_phases():
    phases = []
    for pp in g.mpuls_config.get('phases', 'phases'):
        pairs = g.mpuls_config.get('phases', 'pairs')
        for p in pairs[0].get(str(pp), [-1]):
            phases.append(p)
    phases.append(-1)
    return phases

def parseSearchOptions(options, extended):
    return g.case_search.convert_form_parameters(options,
                                                 session['USER_AUTHORIZED'],
                                                 extended=extended)

def search_options_from_session():
    options = session.get('CASE_OVERVIEW_SEARCHOPTIONS')
    if options:
        return options

    user = session['USER_AUTHORIZED']
    if user.isAdmin():
        return g.mpuls_config.get('search', 'admin_defaults')[0]
    else:
        return g.mpuls_config.get('search', 'cm_defaults')[0]


class CaseOverviewController(BaseController):

    @checkRole(('admin', 'cm', 'pt_dlr'))
    def index(self):
        return self.overview()

    @checkRole(('admin', 'cm', 'pt_dlr'))
    def overview_extended(self):
        extended = session.get('EXTENDED_SEARCH')
        if extended is not None:
            extended = not extended
        else:
            extended = True
        session['EXTENDED_SEARCH'] = extended
        session.save()
        return self.overview()

    @checkRole(('admin', 'cm', 'pt_dlr'))
    def overview(self):
        # If reset parameter is set ti != 0 the delete all saved params and
        # cases on overview page
        reset = self._checkBool(request.params.get('reset', '0'))
        if reset:
            session['CASE_OVERVIEW_SEARCHOPTIONS'] = {}
            session['EXTENDED_SEARCH'] = False
            session.save()

        # Get saved search params from session.
        search_options = search_options_from_session()
        # Update search options with ordering
        sort, order = get_sort(request.params) 
        search_options['sort_order'] = order
        search_options['sort_field'] = sort 
        session['CASE_OVERVIEW_SEARCHOPTIONS'] = search_options
        session.save()
        return self.search(search_options)

    def _getCaseList(self, search_options):
        cases = MpulsCaseOverview()
        log.debug('search options: %s' % search_options)
        if search_options:
            options = self.parseSearchOptions(search_options,
                                        extended=session.get("EXTENDED_SEARCH"))
            cases.search(options)
            if hasattr(self, "get_pages_count"):
                cases.set_count(self.get_pages_count(options))
        return cases

    def _renderOverview(self, cases, form_results, form_defaults, form_errors):
        user = session['USER_AUTHORIZED']
        # Values used in the template
        c.extended_search = session.get('EXTENDED_SEARCH', False)
        ulo = UserListObject()
        filter = [u.id for u in ulo.getAdminList()]
        c.userlist = ulo.getUserList(filter)
        c.phases = PhaseFactory().build() 

        #count num of cases
        c.count_all = len([x for x in cases.getDatasets()])
        c.count_own = len([x for x in cases.getDatasets()
                           if x.user_is_editor(user.id)])
        c.count_standin = c.count_all - c.count_own
        c.count_anon = len([x for x in cases.getDatasets()
                                if x.getState().getState() == 5])
        c.count_markanon = len([x for x in cases.getDatasets()
                                if x.getState().getState() == 4])
        c.count_del = len([x for x in cases.getDatasets()
                           if x.getState().getState() == 3])

        # Hide evaluation option if someone did a search for a particular editor
        c.hide_evaluation = False
        if form_results.get('editor') != -1:
            c.hide_evaluation = True

        session["id_field_candidate"] = "m.id"
        form_defaults["all_ids"] = " ".join(unicode(case.id)
                                            for case in cases.getDatasets())
        c.cases = cases
        overview = render('/casemanagement/overview.mako')
        return formencode.htmlfill.render(overview, defaults=form_defaults,
                                          errors=form_errors)

    @checkRole(('admin', 'cm', 'pt_dlr'))
    def search(self, search_options=None):
        validator = g.case_search.get_form_validator()
        form_result = {}
        form_errors = {}
        form_defaults = None
        if search_options is None:
            search_options = {}
            form_result.update(search_options_from_session())
        else:
            form_result.update(search_options)

        if not search_options:
            try:
                form_result.update(validator.to_python(request.params))
                search_options.update(form_result)
                session['CASE_OVERVIEW_SEARCHOPTIONS'] = search_options 
                session.save()
            except formencode.Invalid, error:
                log.exception("search: formencode error %s" % error)
                form_defaults = error.value
                form_errors = error.error_dict or {}

        if form_defaults is None:
            form_defaults = validator.from_python(form_result)

        cases = self._getCaseList(search_options)
        return self._renderOverview(cases, form_result, form_defaults,
                                    form_errors)

    def parseSearchOptions(self, options, extended):
        return parseSearchOptions(options, extended)

# vim:set ts=4 sw=4 si et sta sts=4 fenc=utf8:
