# -*- coding: utf-8 -*-
import os
import logging
import formencode
import pylons

from pylons import app_globals as g, tmpl_context as c

from libmpuls.evaluation.config import EvaluationConfig
from libmpuls.evaluation.evaluation import EvaluationSet
from libmpuls.evaluation.export import EvaluationExportXML, \
     EvaluationExportCSV, EvaluationExportHTML

from mpulsweb.lib.translation import _
from mpulsweb.lib.base import BaseController, render, request, response, session
from mpulsweb.lib.db import db
from mpulsweb.lib.validators import EvaluationFormValidator
from mpulsweb.model.phase import PhaseFactory, get_phase_successors


log = logging.getLogger(__name__)


def get_configfile(id):
    for enabled_eval in g.mpuls_config.get('evaluations', 'enabled'):
        if enabled_eval.get('id') == str(id):
            filename = enabled_eval.get('config')
            eval_dir = pylons.config.get('mpuls.app.path.evaluation')
            path = os.path.join(eval_dir, filename)
            if os.path.exists(path):
                return path
            else:
                for d in pylons.config.get('pylons.paths').get('static_files'):
                    path = os.path.join(d, 'xml', filename)
                    if os.path.exists(path):
                        return path
    return None

def get_phase_dates(id):
    '''Helper function to get start and end dates of phases ready to use in
    SQL-Statements for evaluation based on the configuration. Note that
    enddates will have a coalesce clause.'''
    sdate, edate = None, None
    phaselist = PhaseFactory().build()
    phase = None
    for p in phaselist:
        if p.id == id:
            phase = p
    if phase:
        start, end = phase.start, phase.end
        sdate = start.datefield
        edate = "coalesce(%s, now())" % end.datefield
    log.debug('sdate: %s, edate: %s' % (sdate, edate))
    return sdate, edate


def get_default_options():
    """Return a dictionary with the default evaluation options"""
    return dict(phase=g.mpuls_config.get('evaluations', 'default-phases'),
                type_ending=g.mpuls_config.get('evaluations', 'default-ending'),
                start_date=g.mpuls_config.get_date('search',
                                                   'default-start-date'),
                end_date=g.mpuls_config.get_date('search', 'default-end-date'))

def get_common_options(evalconfig, id, selected_ids):
    """Return a dictionary with common evaluation options.

    The parameter id should be the id of the evaluation. It will be
    returned in the dictionary with the key 'id'. The information from
    the configuration file for the evaluation is taken from the
    evalconfig parameter which is expected to be an instance of
    EvaluationConfig.

    The parameter selected_ids should be a sequence of integers with the
    ids of the cases selected for evaluation. If no cases were selected,
    it should be an empty sequence.
    """
    return dict(id=id, selected_ids=selected_ids,
                typelist=evalconfig.get_evaluations())


def get_session_options():
    """Return a dictionary with the evaluation options stored in the session
    """
    options = dict()
    soptions = session.get('evaluation.options')
    if soptions:
        for name in ["start_date", "end_date", "phase", "type_ending"]:
            if name in soptions:
                options[name] = soptions[name]
    return options


def get_phase_filter_for_evaluation(eval_phases):
    """Helper function to get more human readable forms of the phase filter for
    the evaluations. """
    
    if eval_phases:
        phases = PhaseFactory().build()
        selected_phases = [phase.getDescription() for phase in phases \
                            if phase.id in eval_phases]
        selected_phases = ', '.join(selected_phases)
    else:
        selected_phases = 'keine'
    
    return selected_phases
    
def get_id_filter_for_evaluation(eval_ids):
    """Helper function to get more human readable forms of the id filter for
    the evaluations. """
    
    if eval_ids:
        selected_ids = eval_ids.replace(' ', ', ')
    else:
        selected_ids = 'keine'
    
    return selected_ids


class EvaluateController(BaseController):

    def index(self):
        return self.evaluate()

    def reset_filter(self, id):
        if session.has_key("evaluation_ids"):
            del session['evaluation_ids']
        if session.has_key("id_field"):
            del session['id_field']
        if session.has_key("sql_where"):
            del session['sql_where']
        session.save()
        return self.evaluate(id)

    def _get_validator(self):
        return EvaluationFormValidator()

    def _get_options(self, id, evalconfig):
        common_options = get_common_options(evalconfig, id,
                                            session.get("evaluation_ids", ()))
        return [get_default_options(), common_options, get_session_options()]

    def _render_options_disabled(self, id):
        for evaluation in g.mpuls_config.get('evaluations', 'enabled'):
            if evaluation.get("id") == str(id):
                return evaluation.get('options_disabled', "0") == "1"
        return False

    def evaluate(self, id=None):
        evalconfig = EvaluationConfig(get_configfile(id), None, None, None,
                                        None, None, None, None, None)

        evaloptions = dict()
        for options in self._get_options(id, evalconfig):
            evaloptions.update(options)
        validator = self._get_validator()
        defaults = validator.from_python(evaloptions)

        c.phases = PhaseFactory().build()
        c.eval_config_id = id
        c.evalconfig = evalconfig
        c.render_options_disabled = self._render_options_disabled(id)
        c.selected_ids = defaults.get('selected_ids', "")
        form = render('/evaluation/evaluate.mako')
        return formencode.htmlfill.render(form, defaults=defaults,
                                          errors={}, auto_insert_errors=False)

    def _get_evalparams(self, form_result, id_field, sql_where):
        """Turn parameters from the evaluation form into evaluation parameters.
        The parameter form_result should be a dictionary created from
        the raw form paramters with EvaluationFormValidator.

        The parameter id_field should be a string with an SQL expression
        giving the field which should match the ids in the selected_ids
        in form_result.

        The parameter sql_where should be a string with an SQL
        expression. It will be used directly in the sql_where expression
        created by this method and will be combined with the other
        clauses with 'AND'. If the sql_where is None or an empty string,
        it will be ignored.
        """
        params = {}
        params['id'] = form_result['id']
        params['start_date'] = str(form_result['start_date'])
        params['end_date'] = str(form_result['end_date'])
        params['typelist'] = form_result['typelist']
        params['phase'] = form_result['phase']
        params['type_ending'] = form_result['type_ending']

        where_clauses = []

        # Dates
        # Build timeframes bases on selected phases.
        dates = []
        all_phases = []
        tdates = []
        phase = params.get('phase')
        dates.append('( ')
        if (params.get('start_date') != 'None'
            and params.get('end_date') != 'None'):
            if phase:
                for p in phase:
                    if int(p) >= 0:
                        # phase is unknown -> has no start and endphase
                        sdf, edf = get_phase_dates(p)
                        suc = get_phase_successors(p)
                        all_phases.append(p)
                        tdates.append("('%s'::date <= %s AND '%s'::date >= %s"
                                      " AND phase IN (%s) )"
                                      % (params.get('start_date'), edf,
                                         params.get('end_date'), sdf,
                                         ",".join(["%s" % s for s in suc])))
                    else:
                        tdates.append('phase IN (-1)')
                dates.append(" OR ".join(tdates))
            else:
                dates.append("FALSE")
        else:
            dates.append("TRUE")
        dates.append(' )')

        where_clauses.append("".join(dates))

        selected_ids = "TRUE"
        if form_result.get("selected_ids"):
            selected_ids = ("(%s in (%s))"
                            % (id_field,
                               ",".join(map(str, form_result["selected_ids"]))))

        where_clauses.append(selected_ids)

        if sql_where:
            # The sql_where parameter is currently (2011-10-14) only set
            # when running as an evaluation server (it's set by
            # EvaluationSearch). The database of the evaluation server
            # has an additional table, fkz_tbl, which may be referenced
            # by the expression in sql_where. That table doesn't exist
            # in a normal mpuls application database, so we can only add
            # the join clause for the fkz_tbl on evaluation servers.
            # FIXME: There should be a better way to communicate the
            # requirements of the sql_where clause, especially if we
            # ever extend it to the search on the normal application
            # server.
            where_clauses.append(sql_where)
            join_fkz = "JOIN fkz_tbl f ON f.id::integer = m.fkz::integer"
        else:
            join_fkz = ""
        where_clauses.append(self.type_ending_clause(params))
        where_clauses.append(self.appspecific_clause(form_result))

        params['sql'] = ("SELECT %%(fields)s from master_tbl_eval_total_view m"
                         " JOIN ka_status_tbl_view st on st.master_id = m.id"
                         " %s"
                         " WHERE %s "
                         % (join_fkz, " AND ".join(where_clauses)))

        #Datefields
        if all_phases:
            params['start_date_field'] = get_phase_dates(min(all_phases))[0]
            params['end_date_field'] = get_phase_dates(max(all_phases))[1]
        else:
            params['start_date_field'] = None
            params['end_date_field'] = None
        return params

    def type_ending_clause(self, params):
        """Return the condition to search based on the type of ending.
        Derived classes should override this if necessary as the details
        are usually application specific.
        """
        return "TRUE"

    def appspecific_clause(self, params):
        """Return app-specific conditions to search based on different criteria.
        Derived classes should override this if necessary as the details
        are usually application specific.
        """
        return "TRUE"

    def evaluateAction(self):
        """Return an HTML file containing the result of one or more evaluations.
        """
        params = formencode.variabledecode.variable_decode(request.params)
        validator = self._get_validator()

        id = int(params['id'])
        config_file = get_configfile(id)
        evalconfig = EvaluationConfig(config_file, None, None, None,
                                      None, None, None, None, None)

        # Check values
        try:
            form_result = validator.to_python(params)
        except formencode.Invalid, error:
            log.exception("Error while decoding form parameters")
            form_result = error.value

            # format selected ids as a work-around for an inconsistency
            # in formencode. Depending on what the reason for the
            # Invalid exception actually is, the value of "selected_ids"
            # may have been converted to a list of ints or not. If it's
            # still a string, it be used directly as the value of the
            # hidden input field in the form by formfill. However, if it
            # has been converted to a list of ints already the
            # formatting will be wrong and the validator will not be
            # able to parse it anymore. So as a work-around, we
            # explicitly reformat the value again here if it has been
            # converted already.
            selected_ids = form_result.get("selected_ids")
            if isinstance(selected_ids, list):
                form_result["selected_ids"] = \
                    " ".join([repr(i) for i in selected_ids])

            form_errors = error.error_dict or {}
            c.phases = PhaseFactory().build()
            c.eval_config_id = id
            c.evalconfig = evalconfig
            form = render('/evaluation/evaluate.mako')
            return formencode.htmlfill.render(form, defaults=form_result,
                                              errors=form_errors,
                                              auto_insert_errors=False)

        if self._render_options_disabled(id):
            # when doing evaluations with disabled options (like adele), always
            # use the defaults. Since the options are shown as disabled in the
            # HTML form, form_result will usually not contain them anyway.
            form_result.update(get_default_options())
        # Build evaluation
        eval_params = self._get_evalparams(form_result, session.get("id_field"),
                                           session.get("sql_where"))

        conn = db.getConnection()
        try:
            try:
                evalconfig = EvaluationConfig(config_file, conn,
                                              eval_params['start_date'],
                                              eval_params['end_date'],
                                              eval_params['start_date_field'],
                                              eval_params['end_date_field'],
                                              None,
                                              eval_params['sql'],
                                              eval_params['typelist'])
                
                
                # Massage the filter to make it human readable
                selected_phases = get_phase_filter_for_evaluation(params.get('phase'))
                selected_ids = get_id_filter_for_evaluation(params.get('selected_ids'))
                
                evalconfig.set_filter({"phases": selected_phases,
                                       "items": selected_ids}) # Cases or agencies
                
                # Set second parameter to True to enable debugging
                evalset = EvaluationSet(evalconfig)
                evalset.evaluate()
            except:
                log.exception(_('Error: Evaluation failed'))
        finally:
            db.recycleConnection(conn, None)

        exporter = EvaluationExportHTML(show_percent=form_result['show_percent'])
        c.result = evalset.export(exporter)
        session['evaluation.params'] = eval_params
        session['evaluation.options'] = form_result
        session.save()
        return render('/evaluation/result.mako')


    def exportXML(self):
        '''Return an XML file containing the result of one or more evaluations.
        '''
        response.headers['Content-Type'] = 'application/xml; charset=utf8'
        response.headers['Content-Disposition'] = \
                                  'attachment; filename=evaluation-export.xml'
        return self._do_export(EvaluationExportXML)

    def exportCSV(self):
        '''Return an CSV file containing the result of one or more evaluations.
        '''
        response.headers['Content-Type'] = 'application/csv; charset=utf8'
        response.headers['Content-Disposition'] = \
                                  'attachment; filename=evaluation-export.csv'
        return self._do_export(EvaluationExportCSV)

    def _do_export(self, export_class):
        params = session.get('evaluation.params')
        try:
            try:
                conn = db.getConnection()
                cur = None
                evalconfig = EvaluationConfig(get_configfile(params['id']),
                                              conn,
                                              params['start_date'],
                                              params['end_date'],
                                              params['start_date_field'],
                                              params['end_date_field'],
                                              None,
                                              params['sql'],
                                              params['typelist'])
                
                # Massage the filter to make it human readable
                selected_phases = get_phase_filter_for_evaluation(params.get('phase'))
                selected_ids = get_id_filter_for_evaluation(params.get('selected_ids'))
                
                evalconfig.set_filter({"phases": selected_phases,
                                       "items": selected_ids}) # Cases or agencies
                
                # Set second parameter to True to enable debugging
                evalset = EvaluationSet(evalconfig)
                evalset.evaluate()
                return evalset.export(export_class())
            except:
                log.exception("Export with %r failed", export_class)
        finally:
            db.recycleConnection(conn, cur)
