# -*- coding: utf-8 -*-

import logging

from pylons import request, tmpl_context as c, app_globals as g
from pylons.controllers.util import abort, redirect
from pylons.i18n.translation import get_lang

from formed.instance.tree import SetDataException, RepeatInstanceContainerNode
from formed.renderer.html import apply_hidden_booleans, RepeatGroupRenderer, \
     ErrorRenderer, TreeFormData

from mpulsweb.lib.translation import _
from mpulsweb.lib.security import checkRole
from mpulsweb.lib.navigation import mark_folderopen, get_navigation
from mpulsweb.lib.base import BaseController, render, session
from mpulsweb.lib.helpers import literal, url
from mpulsweb.lib.formedrenderer import ViewRenderer, ViewPrintRenderer, \
        UrlFactory

from mpulsweb.model.phase import PhaseFactory, PC_MANUAL, PC_FULLAUTO


log = logging.getLogger(__name__)


def convert_error_items(error_items):
    errors = {}
    defaults = {}
    for items in error_items:
        for e in items:
            errors[e.name] = e.msg
            defaults[e.name] = e.bad
    return errors, defaults


def convert2dic(params):
    result = {}
    for i in params:
        # libformed
        if (i not in ['save', 'savenext', 'page']
            and not i.startswith(('add_rg', 'del_rg'))):
            result[i] = params[i]
    return result


class PageRenderer(object):

    def __init__(self):
        self.set_viewrenderer(ViewRenderer)
        self.set_errorrenderer(ErrorRenderer)
        self.set_printrenderer(ViewPrintRenderer)
        self.set_rgrenderer(RepeatGroupRenderer)

    def set_viewrenderer(self, viewrenderer, **kw_args):
        self.viewrenderer = viewrenderer
        self.viewrenderer_args = kw_args

    def set_errorrenderer(self, errorrenderer, **kw_args):
        self.errorrenderer = errorrenderer
        self.errorrenderer_args = kw_args

    def set_printrenderer(self, printrenderer, **kw_args):
        self.printrenderer = printrenderer
        self.printrenderer_args = kw_args

    def set_rgrenderer(self, rgrenderer, **kw_args):
        self.rgrenderer = rgrenderer
        self.rgrenderer_args = kw_args

    def get_rendered_page(self, id, instance_tree=None):
        if instance_tree is None:
            factory = g.get_instance_factory()
            instance_tree = factory.loadInstanceTreeByIdentifier(
                session['case'].id,
                session.get('uncommited_fields'))

        log.debug('Getting page "%s"' % id)
        page = instance_tree.getItem(id)

        # Abort with a 404 HTTP response if the page wasn't found. abort
        # raises an exception, so in the code after this branch we can
        # assume that the page was found.
        if not page:
            log.error('Page not found %s' % id)
            abort(404)

        page_label = page.getMeta().getDescription()

        url_factory = UrlFactory()
        renderer_kw_args = dict(read_only=session.get('render_mode', 'ro') == 'ro',
                                languages=get_lang())
        if isinstance(page, RepeatInstanceContainerNode):
            formdata = TreeFormData(instance_tree)
            renderer_kw_args.update(self.rgrenderer_args)
            renderer = self.rgrenderer(formdata, page.getMeta(), url_factory,
                                       **renderer_kw_args)
            errors = {}
            defaults = {}
        else:
            errors, defaults = convert_error_items(page.getErrors())
            formdata = TreeFormData(instance_tree, shadow=defaults,
                                    errors=errors, node=page)
            if c.print_version:
                renderer_kw_args.update(self.printrenderer_args)
                renderer = self.printrenderer(formdata, page.getMeta(),
                                              url_factory, **renderer_kw_args)
            else:
                renderer_kw_args.update(self.viewrenderer_args)
                renderer = self.viewrenderer(formdata, page.getMeta(),
                                             url_factory, **renderer_kw_args)

        navigation_item = session['navigation.tree'].getTreeItem(id)
        mark_folderopen(navigation_item)

        c.formularheader = renderer.render_page_headings()
        c.formular = renderer.render(show_next=navigation_item.nextKey(),
                                    hidden_fields=[("page",
                                                    page.getIdentifier())])

        error_renderer = self.errorrenderer(formdata, g.formedTree, errors,
                                            renderer.warnings, renderer.notices,
                                            renderer.widget_order, url_factory,
                                            languages=get_lang())
        c.formularerrors = error_renderer.render()
        c.page_id = page.getIdentifier()
        session['navigation.selectedpage'] = id
        session.save()

        return render('/formularpage/content.mako')


get_rendered_page = PageRenderer().get_rendered_page

class FormularpageController(BaseController):

    def index(self):
        # Return a rendered template
        # return render('/formularpage.mako')
        # or, return a response
        return 'Hello World'

    def _render_page(self, case, page_id):
        c.formular = get_rendered_page(page_id,
                                       case.get_formed_instance_tree())
        return render('/formularpage/formular.mako')

    @checkRole(('admin', 'cm', 'pt_dlr'))
    def select(self, id):
        return self._render_page(self._load_current_case(read_only=True), id)

    @checkRole(('cm'))
    def edit(self, id):
        case = self._load_current_case()
        session['render_mode'] = 'rw'
        session.save()
        return self._render_page(case, id)

    @checkRole(('admin', 'cm', 'pt_dlr'))
    def show(self, id):
        session['render_mode'] = 'ro'
        session.save()
        return self.select(id)

    @checkRole(('cm'))
    def undo(self, id):
        unc = session.get('uncommited_fields')
        if unc:
            try:
                del unc[id]
            except KeyError:
                pass
            else:
                session['uncommited_fields'] = unc
                session.save()
        return self.select(session['navigation.selectedpage'])

    @checkRole(('cm'))
    def save(self):
        case = self._load_current_case()
        case_id = case.id
        instance_tree = case.get_formed_instance_tree()

        page_id = request.params['page']

        form_result = convert2dic(request.params)
        apply_hidden_booleans(form_result)

        self._before_set_data_hook(case_id, form_result)

        try:
            instance_tree.setData(form_result)
        except SetDataException, err:
            # We can ignore this exception because errors that occurred
            # will be shown to the user on the form page anyway via the
            # uncommitted fields data stored in the session.
            pass
        session['uncommited_fields'] = instance_tree.commit()

        if session['navigation.tree'].getTreeItem(page_id).parent_repeat():
            session['navigation.tree'] = get_navigation(
                case_id,
                session['navigation.openfolders'])
        session.save()

        # How to proceed after saving the data?
        self._handle_repeat_group_redirects()

        if request.params.has_key('savenext'):
            ti = session['navigation.tree'].getTreeItem(page_id)
            if ti.parent_repeat():
                page_id = ti.parent.key
            elif ti.nextKey().parent_repeat():
                page_id = ti.nextKey().parent.key
            else:
                page_id = ti.nextKey().key

        # Check consistence of case and retag
        self._check_case_validity(case)

        phase_changed = self._handle_automatic_phase_change(page_id)
        if phase_changed:
            return phase_changed

        redirect(url(controller="formularpage", action='select', id=page_id))

    def _before_set_data_hook(self, case_id, form_result):
        """Modify the form_result dict of the save method before it's applied.
        Override in derived classes to handle application specific
        special cases.  The default implementation does nothing.
        """

    def _check_case_validity(self, case):
        """Check validity of the case given by case_id."""
        case.checkValidity()

    def _handle_automatic_phase_change(self, page_id):
        """Change Phase automatically depending on the state of the case"""

        fullautomatic_title = _("Phase automatically changed")
        halfautomatic_title = _("Change phase?")

        case = g.case_factory.loadById(session['case'].id)
        # Depending if the meta-server is enabled we need different dialogs:
        if g.mpuls_config.is_enabled('case-module', 'meta-server') \
        and case.getMeta().is_linked():
            fullautomatic_text = _("The phase of the case was changed automatically"
                                   " after checking the filled out required fields."
                                   " The phase changed from '%s' to '%s'. The"
                                   " case has been synced with the Meta-Server"
                                   " Please click on 'Ok', to proceed editing"
                                   " the case.")
            halfautomatic_text = _("After checking the filled out required fields,"
                                   " it seems that a phase change can be done from"
                                   " phase '%s' to '%s'."
                                   " Please click on 'Yes' to perform the"
                                   " change and sync the case with the"
                                   " Meta-Server." " Click on 'No' to perform the"
                                   " change later" " manually on the phase"
                                   " overview page.")
        else:
            fullautomatic_text = _("The phase of the case was changed automatically"
                                   " after checking the filled out required fields."
                                   " The phase changed from '%s' to '%s'."
                                   " Please click on 'Ok', to proceed editing"
                                   " the case.")
            halfautomatic_text = _("After checking the filled out required fields,"
                                   " it seems that a phase change can be done from"
                                   " phase '%s' to '%s'."
                                   " Please click on 'Yes' to perform the change."
                                   " Click on 'No' to perform the change later"
                                   " manually on the phase overview page.")

        changemode = g.mpuls_config.get('phases', 'changemode')
        if changemode != PC_MANUAL:
            phases = PhaseFactory().load(session['case'].id)
            log.debug("current phase: %s" % phases.get_current_phase_id())
            log.debug("max phase? %s" % phases.get_max_phase_id())
            current_phase = phases.get_current_phase_id()
            log.debug(g.mpuls_config.get('phases', 'description'))
            reachable_phase = phases.get_max_phase_id()
            if current_phase != reachable_phase:
                cpd = phases.get_phase(current_phase).getDescription()
                rpd = phases.get_phase(reachable_phase).getDescription()
                if changemode == PC_FULLAUTO:
                    log.debug("fullautomatic phase change!")
                    # load case and set phase
                    case = g.case_factory.loadById(session['case'].id)
                    case.setPhase(reachable_phase)
                    # build dialog
                    c.dialog_title = fullautomatic_title
                    c.dialog_text = literal((fullautomatic_text % (cpd, rpd)))
                    c.url_ok = url(controller="formularpage",
                                       action="select", id=page_id)
                    return render('/phase/dialogs/success_fullautomatic_set_phase.mako')
                else:
                    log.debug("halfautomatic phase change!")
                    # build dialog
                    c.dialog_title = halfautomatic_title
                    c.dialog_text = literal((halfautomatic_text % (cpd, rpd)))
                    c.url_ok = url(controller="phase", action="setPhase",
                                       id=session['case'].id,
                                       value=reachable_phase)
                    c.url_no = url(controller="formularpage",
                                       action="select", id=page_id)
                    return render('/phase/dialogs/success_halfautomatic_set_phase.mako')
        return None

    def _handle_repeat_group_redirects(self):
        """Redirect the appropriate repeat group action implied in the request
        """
        for param in request.params:
            if param.startswith('add_rg') and not param.endswith(('x', 'y')):
                if param == 'add_rg':
                    redirect(url(controller='repeatgroup', action="new",
                                id=request.params.get('add_rg')))
                else:
                    redirect(url(controller='repeatgroup', action="new",
                                id=param.split('|')[1]))
            elif param.startswith('del_rg') and not param.endswith(('x', 'y')):
                redirect(url(controller='repeatgroup', action="delete",
                            id=param.split('|')[1]))


    @checkRole(('admin', 'cm', 'pt_dlr'))
    def printout(self, id):
        c.print_version = 1
        render_mode = session['render_mode']
        session['render_mode'] = 'ro'
        session.save()
        page = self.select(id)
        session['render_mode'] = render_mode
        session.save()
        return page
