# -*- coding: utf-8 -*-

import logging

from pylons import app_globals as g

from mpulsweb.lib.translation import _

from formed.instance.tree import SetDataException

from mpulsweb.lib.security import checkRole
from mpulsweb.lib.base import BaseController, render, c, h, session
from mpulsweb.lib.anonymize import determineAnonymizedCaseEndPhase
from mpulsweb.model.phase import ConsistenceCheckException, PhaseAnonymizableCheckException

log = logging.getLogger(__name__)


class PrivacyController(BaseController):

    '''
    Controller to handle privacy related things within mpuls. This mainly
    includes accepting or declining the 'Childs declaration of consent' (known
    in WASKA/WASKO) or similar. Additionally it is possible to list available
    static statements.
    '''
    @checkRole('cm')
    def index(self):
        return self.listStatements()

    @checkRole('cm')
    def listStatements(self):
        """Render a list of available static statements related to privacy"""
        return render('privacy/list_statements.mako')

    @checkRole('cm')
    def printStatement(self, id):
        id = self._checkInt(id)
        case = self._loadCase(id, read_only=True)
        statement = case.getPrivacyStatement()
        c.content = statement.fillout()
        return render('privacy/default_statement.mako')

    @checkRole('cm')
    def set(self, id):
        c.targeturl = h.url(controller="case", action="digest", id=id)
        c.ds_id = id
        return render('privacy/edit_statement.mako')

    @checkRole('cm')
    def accept(self, id, confirmed):
        id = self._checkInt(id)
        confirmed = self._checkBool(confirmed)
        case = self._loadCase(id)
        if confirmed:
            statement = case.getPrivacyStatement()
            statement.sign()
            # TODO: Think about refreshing the state of the privacy
            # statement in the statement model
            session_case = session.get('case')
            session_case.privacy_statement = True
            session.save()
            c.dialog_title = _(u"Privacy statement set!")
            c.dialog_text = _(u'Please click Bitte klick "OK" to continue work'
                    ' on the case')
            c.url_ok = h.url(controller="case", action="select",
                                 id=case.id)
            return render('privacy/dialogs/accept_success.mako')
        else:
            c.dialog_title = _("Set declaration of consent?")
            c.dialog_text =  _("You have stated that there is a declaration of "
                    "consent from\nthe young adult to the acquisition, handling "
                    "and storage of personal data.\n"
                    "Are you sure this is correct?")
            c.url_yes = h.url(controller="privacy", action="accept",
                                  id=id, confirmed="1")
            c.url_no = h.url(controller="case", action="select",
                                 id=id, confirmed=0)
            return render('privacy/dialogs/confirm.mako')

    @checkRole('cm')
    def decline(self, id, confirmed=False):
        id = self._checkInt(id)
        confirmed = self._checkBool(confirmed)
        case = self._loadCase(id)
        try:
            case.check_anonymizeability()
        except PhaseAnonymizableCheckException, e:
            c.dialog_title = _('Decline privacy statement not possible')
            c.dialog_text = h.literal( _('The case is in a non consistent state und it will not be possible to anonimise it after retracting the privacy statement. The reason for this is listed below:%s') % e.value)
            c.url_ok = h.url(controller='case', action='digest',
                                 id=case.id)
            return render('/casemanagement/dialogs/failed_markanonymize.mako')

        except ConsistenceCheckException, e:
            c.dialog_title = _('Decline privacy statement not possible')
            c.dialog_text = h.literal(_('<p>The case is in a non consistent state und it will not be possible to anonimise it after retracting the privacy statement. The reason for this is listed below:</p><p>%s</p><p>Please refer to the documentation in case that the privacy statement has been declined.</p>') % e.value)
            c.url_ok = h.url(controller='case', action='digest',
                                 id=case.id)
            return render('/casemanagement/dialogs/failed_markanonymize.mako')

        if confirmed:
            # Mark case for anonymizion
            try:
                case.make_anonymizable()
                case.markAnonymize()
                c.dialog_title = _(u"""Freed case document to anonymise.""")
                c.dialog_text = _(u"The case document has been freed to be"
                        u" anonymised and has been given to\n"
                        u"the admininstration. Please click on OK, to get back "
                        u"to the overview.")
                c.url_ok = h.url(controller='case_overview', action='index')
                statement = case.getPrivacyStatement()
                statement.decline()
                return render('/privacy/dialogs/decline_success.mako')
            except SetDataException, e:
                errors = []
                for k, values in e.errors.iteritems():
                    for v in values:
                        errors.append(v.msg)
                c.dialog_title = _(u"""Anonymising not possible.""")
                c.dialog_text = _("To anonymise this case document was not "
                        "possible, because of following \nreason:"
                        "</p><p><i>%s</i></p><p>Please click on OK to get"
                        " back to the overview.\n</p>") % "<br>".join(errors) 
                c.url_ok = h.url(controller='case', action='select',
                                     id=case.id, confirmed=1)
                return render('/privacy/dialogs/decline_failed.mako')

        else:
            # Get description of the phase which will be set after anonymisation
            pdescriptions = g.mpuls_config.get('phases', 'description')[0]
            c.dialog_title =  _(u"Retract privacy statement?")
            c.dialog_text = _(u"The case will be anonymised if you retract the privacy statement.""")
            c.url_yes = h.url(controller="privacy", action="decline",
                                  id=id, confirmed="1")
            c.url_no = h.url(controller="case", action="select",
                                 id=id, confirmed=0)
            return render('privacy/dialogs/confirm.mako')

# vim:set ts=4 sw=4 si et sta sts=4 fenc=utf-8:
