# -*- coding: utf-8 -*-
import logging
import formencode
from subprocess import Popen, PIPE
from email.mime.text import MIMEText

from pylons import request, tmpl_context as c, app_globals as g
from mpulsweb.lib.translation import _

from mpulsweb.lib.helpers import url, format_number
from mpulsweb.lib.base import BaseController, render
from mpulsweb.lib.validators import EditSettingsForm, Number, MaxMinInt
from mpulsweb.lib.security import checkRole, get_unmapped_db_name
from mpulsweb.model.agencysettings import Agency


log = logging.getLogger(__name__)


class SettingsController(BaseController):

    @checkRole('admin')
    def index(self):
        return self.show()

    def _get_agency(self):
        return Agency()

    def _get_settings_validator(self, param):
        return EditSettingsForm(**param)

    def show(self):
        c.form_errors = {}
        c.form_result = {}
        c.agency = self._get_agency()
        return render('/settings/show.mako')

    @checkRole('admin')
    def edit(self):
        c.agency = self._get_agency()
        c.form_errors = {}
        c.form_result = c.agency.conf
        form = render('/settings/edit.mako')
        c.form_result['job_percentage'] = str(format_number(float(c.form_result['job_percentage'])).replace(',', '.'))
        return formencode.htmlfill.render(form,
                                          defaults=c.form_result,
                                          auto_insert_errors=False,
                                          errors=c.form_errors)

    @checkRole('admin')
    def editAction(self):
        param = {"max_speicherdauer":MaxMinInt(min=g.mpuls_config.get('case',
                                                         'remindermindays'),
                                  max=g.mpuls_config.get('case',
                                                         'remindermaxdays'),
                                  not_empty=True)
                }
        # Additional validator if the job_percentage is enabled, otherwise it
        # will complain about missing values on applications that don't have
        # the field
        if g.mpuls_config.is_enabled('job_percentage', 'enabled'):
            param["job_percentage"] = Number(min=0)

        validator = self._get_settings_validator(param)
        c.agency = self._get_agency()
        try:
            c.form_result = validator.to_python(request.params)
        except formencode.Invalid, error:
            c.form_result = error.value
            c.form_errors = error.error_dict or {}
            c.form_error_msg = error
            form = render('/settings/edit.mako')
            return formencode.htmlfill.render(form,
                                              defaults=c.form_result,
                                              auto_insert_errors=False,
                                              errors=c.form_errors)

        # Try to get old and new job percentage value.
        if g.mpuls_config.is_enabled('job_percentage', 'enabled'):
            old_job_perc = c.agency.getJobPercentage()
            new_job_perc= c.form_result['job_percentage']

        # Store the values.
        self.pre_save_hook(c.form_result)
        c.agency.store(c.form_result)

        # If the job_percentage is enabled and email adresses are configured,
        # send an email each time the value is saved and changed
        if (g.mpuls_config.is_enabled('job_percentage', 'enabled')
           and g.mpuls_config.get('job_percentage', 'from_email') != '0'
           and g.mpuls_config.get('job_percentage', 'to_email') != '0'):
            self.send_job_percentage_email(old_job_perc, new_job_perc)

        c.dialog_title = _('Configurations saved!')
        c.dialog_text = _('Please click on ok to continue.')
        c.url_ok = url(controller='settings', action='index')
        return render ('/settings/dialogs/success_edit.mako')

    def pre_save_hook(self, form_data):
        """
        Customizable function, is called before saving the agency settings
        """
        pass

    def send_job_percentage_email(self, old_value, new_value):
        """
        Sends an email if the value of the job percentage was changed
        """
        if str(old_value) != str(new_value):
            # Prepare the contents of the email
            from_adress = g.mpuls_config.get('job_percentage', 'from_email')
            to_adress = g.mpuls_config.get('job_percentage', 'to_email')
            program_name = g.mpuls_config.get('common', 'name')
            agency_id = get_unmapped_db_name()  #certnumber/db
            agency_account = c.agency.getFKZ()
            msg_values = {'agency': c.agency.getName(),
                          'program': program_name,
                          'agency_account': agency_account,
                          'agency_id': agency_id,
                          'old': format_number(float(old_value), precision=1),
                          'new': format_number(float(new_value), precision=1)}
            message_text = _('The job percentage in agency '
                             '%(agency)s in program "%(program)s"'
                             ' changed from %(old)s to %(new)s.'
                            )% msg_values
            # Set the values for the email
            msg = MIMEText(message_text, _charset="utf-8")
            msg['Subject'] = (_('[%(program)s]: Change of jobpercentage in %(agency)s (%(agency_account)s/%(agency_id)s)') % msg_values)
            msg['From'] = from_adress
            msg['To'] = to_adress
            # Send the email
            try:
                log.debug('Sendig email: %s' % msg.as_string())
                p = Popen(["/usr/sbin/sendmail", "-t"], stdin=PIPE)
                p.communicate(msg.as_string())
            except:
                log.error("Error while sending Email")

# vim:set ts=4 sw=4 si et sta sts=4 fenc=utf8:
