# Copyright 2007, 2008 Intevation GmbH, Germany, <info@intevation.de>
#
# This file is part of mpuls WASKA (CoMPUter-based case fiLeS -
# Web-Anwendungs-Server fuer Kompetenzagenturen).
#
# mpuls WASKA is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# mpuls WASKA is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public
# License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with mpuls WASKA. If not, see <http://www.gnu.org/licenses/>.
#
# mpuls WASKA has been developed on behalf of the
# Projekttraeger im Deutschen Zentrum fuer Luft- und Raumfahrt e.V. (PT-DLR)
# within the programme Kompetenzagenturen (Durchfuehrungsphase) funded by
# the Bundesministerium fuer Familie, Senioren, Frauen und Jugend and
# European Social Fund resources.

"""The application's Globals object"""

import os
import re
import logging
import ConfigParser

from paste.reloader import watch_file
from paste.registry import StackedObjectProxy

from formed.instance.backends.postgres import DBFactory as InstanceFactory
from formed.meta.io.document import openDocument

from libmpuls.service.tags import TagSetterConfig

from mpulsweb.config.importer import import_overridable_module
from mpulsweb.lib.config import MpulsAppConfig, MpulsDBMapping, MpulsDBList
from mpulsweb.lib.db import PostgresDBInterface
from mpulsweb.lib.metaclient import get_ssl_context
from mpulsweb.model.annotations import AnnotationsProvider



log = logging.getLogger(__name__)


# MPuls session object.  Application specific wrapper around the pylons
# session object.  The actual functionality is defined in
# mpulsweb/lib/session.py.  It's set on a per-request basis in
# mpulsweb/lib/base.py.  It's located here in app_globals.py to avoid
# excessive problems with circular imports.
session = StackedObjectProxy()


def ssl_option(config, key, default=None):
    """Return the SSL option named by key.

    The SSL option is taken from pylons config with a key formed by
    prepending "mpuls.meta.ssl." to the key passed to this function. If
    there's no value for that key, the value of the parameter default is
    returned.
    """
    return config.get("mpuls.meta.ssl." + key, default)


def ssl_file_option(config, key, default=None):
    """Like ssl_option, but for values that are filenames.
    The filenames are added to the paster reloader watch list, so that
    when the application runs with 'paster serve --reload' paster
    restarts automatically when the file changes.
    """
    filename = ssl_option(config, key, default=None)
    watch_file(filename)
    return filename


class Globals(object):

    """Globals acts as a container for objects available throughout the
    life of the application
    """

    def __init__(self, config):
        """One instance of Globals is created during application
        initialization and is available during requests via the 'g'
        variable
        """
        def filename_from_config(setting, default):
            return os.path.join(config.get('pylons.paths').get('root'),
                                '..', config.get(setting, default))

        # Load mpuls configuration
        config_file = filename_from_config('mpuls.app.config',
                                           'default_config.json')
        mapping_file = filename_from_config('mpuls.app.dbmapping',
                                            'default_dbmapping.json')
        tagging_file = filename_from_config('mpuls.app.tagging',
                                            'default_tagging.xml')
        dblist_file = filename_from_config('mpuls.app.dblist',
                                            'default_dblist.json')


        # Load ini file and make all sections available a global var.
        apps = re.compile('app:')
        self.mpuls_ini = ConfigParser.ConfigParser()
        path = os.path.join(config['here'], config['__file__'])
        self.mpuls_ini.readfp(open(path))

        self.config_path = os.path.dirname(config_file)
        self.mpuls_config = self._get_config(config_file)
        self.mpuls_dbmapping = MpulsDBMapping(mapping_file)
        self.mpuls_dblist = MpulsDBList(dblist_file)

        self._load_form(config)

        # Load taggingconfig
        log.debug("Loading configuration for tagging from %s" % tagging_file)
        self.taggingConfFile = tagging_file
        self.taggingConf = TagSetterConfig(tagging_file) 

        # tell paste to restart the server if the formed files change
        watch_file(tagging_file)

        # meta-server settings
        if ssl_option(config, "enabled", "false") == "true":
            self.ssl_context = get_ssl_context(ssl_file_option(config,
                                                               "cacerts"),
                                               ssl_file_option(config,
                                                               "client_cert"),
                                               ssl_file_option(config,
                                                               "client_key"))
        else:
            self.ssl_context = None

        self.meta_converter = config.get("mpuls.meta.converter")
        if self.meta_converter:
            watch_file(self.meta_converter)


        # TODO: These setups should be while application intialisation, as they
        # are not related to any global vars or functions. Move to
        # config/environment.py

        self._init_case(config)

    def _init_case(self, config):
        from mpulsweb.model.agency import MpulsAgencyFactory, MpulsAgency
        from mpulsweb.model.case import MpulsCase, MpulsCaseFactory
        """Set the mpuls application specific case classes"""
        try:
            case_module = import_overridable_module("model.case", config)
            Case = case_module.Case
            self.case = Case
            CaseFactory = case_module.CaseFactory
            self.case_factory = case_module.CaseFactory(self.case)
        except Exception, e:
            # it would be better to not catch all exceptions here
            # an application should eventually not start if a case module
            # couldn't be loaded. At least log the reason as error
            log.error("Error loading 'model.case' from overridable module."
                      "Taking case from base. Exception was %s", e)
            self.case = MpulsCase
            self.case_factory = MpulsCaseFactory(self.case)

        self.agency = MpulsAgency 

        self.case.set_fields(self.mpuls_config.get("case", "fields"))
        self.agency.set_fields(self.mpuls_config.get("agency", "fields"))

        self.agency_factory = MpulsAgencyFactory(self.agency)

        search_module = import_overridable_module("lib.search", config)
        self.case_search = search_module.CaseSearch(
            self.case_factory.get_search_retrieve_fields(),
            self.case_factory.get_search_match_fields())
        self.casecount_search = search_module.CaseCountSearch(
            self.case_factory.get_search_match_fields())
        self.eval_search = search_module.EvaluationSearch(
            self.agency_factory.get_search_retrieve_fields(),
            self.agency_factory.get_search_match_fields())


    def get_instance_factory(self):
        return InstanceFactory(self.formedTree, PostgresDBInterface())

    def _get_config(self, filename):
        return MpulsAppConfig(filename)

    def _load_form(self, config):
        # Load formed from defined location. If no location is defined. Take
        # the default which is the formed dir under public.
        formed_dir = config.get('mpuls.app.path.formed',
            os.path.join(config.get('pylons.paths').get('static_files')[0],
                         'formed'))
        formed_file = os.path.join(formed_dir,
                                   self.mpuls_config.get('formed', 'treefile'))
        annotation_file = os.path.join(formed_dir,
                                       self.mpuls_config.get('formed',
                                                             'annotation'))
        self.formedTree = self._load_formedtree_from_file(formed_file)
        self.helpData = AnnotationsProvider(annotation_file)

        watch_file(formed_file)
        watch_file(annotation_file)

    def _load_formedtree_from_file(self, formed_file):
        # Build FormedTree. This tree will work as template for all cases
        log.debug("loading formed definition from %r", formed_file)
        return openDocument(formed_file)
