# -*- coding: utf-8 -*-
#
# Copyright 2007, 2008, 2010 Intevation GmbH, Germany, <info@intevation.de>
#
# This file is part of mpuls WASKA (CoMPUter-based case fiLeS -
# Web-Anwendungs-Server fuer Kompetenzagenturen).
#
# mpuls WASKA is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# mpuls WASKA is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public
# License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with mpuls WASKA. If not, see <http://www.gnu.org/licenses/>.
#
# mpuls WASKA has been developed on behalf of the
# Projekttraeger im Deutschen Zentrum fuer Luft- und Raumfahrt e.V. (PT-DLR)
# within the programme Kompetenzagenturen (Durchfuehrungsphase) funded by
# the Bundesministerium fuer Familie, Senioren, Frauen und Jugend and
# European Social Fund resources.
# Authors:
# Torsten Irländer <torsten.irlaender@intevation.de>
# Bernhard Herzog <bernhard.herzog@intevation.de>
#

import logging
from datetime import datetime
from sets import Set
from xml.etree import ElementTree

import psycopg2.extras

from pylons import config

from formed.meta.data import RepeatNode, DateLeaf, IntLeaf, TextLeaf,\
     ChoiceNode, RadioNode, RuleLeaf
from formed.instance.backends.xmlimport import XMLImportException
from formed.instance.backends.dbxmlimport import DBXMLImporter
from formed.instance.backends.common import RuleEvalContext
from formed.instance.tree import SetDataException

from mpulsweb.lib.base import g, h, session
from mpulsweb.lib.export import XLSExport, CSVExport
from mpulsweb.lib.anonymize import determineAnonymizedCaseEndPhase
from mpulsweb.lib.helpers import ensure_unicode, format_date, get_TagSetter, \
dd_mm_YYYY
from mpulsweb.lib.db import PostgresDBInterface, db

from mpulsweb.model.document import listDocuments, export_documents, \
     get_documents_import_factory, import_documents
from mpulsweb.model.statement import PrivacyStatement
from mpulsweb.lib.translation import _
from mpulsweb.model.phase import PhaseFactory, StartPhasePart, \
    PhaseAnonymizableCheckException, ConsistenceCheckException, PHASEFIELD, PC_MANUAL
from mpulsweb.model.user import UserObject
from mpulsweb.model.agencysettings import Agency
from mpulsweb.model.logbook import export_logbook, get_logbook_import_factory, \
     import_logbook
from mpulsweb.model.meta import Meta, get_meta_client


log = logging.getLogger(__name__)

PRESET_FIELDS_IN_MASTERTABLE = ['bearbeiter_id']

LOAD_CASE_SQL = """\
SELECT
  %(fields)s,
  ARRAY(SELECT tag_id FROM nm_tags_cases_view WHERE case_id = %%(id)s) AS tags
FROM master_tbl_view WHERE id = %%(id)s"""

#SQL
LOAD_STANDIN_SQL = """
SELECT n.benutzer_id
FROM nm_benutzer_master_tbl_view n
JOIN master_tbl_view m ON m.id = n.master_id
WHERE m.id = %(id)s"""
ADD_STANDIN_SQL = """
SELECT create_nm_benutzer_master_ds(%(groupid)s, %(caseid)s)"""
DELETE_STANDIN_SQL = """
SELECT delete_nm_benutzer_master_ds(%(groupid)s, %(caseid)s)"""
DELETE_CASE_SQL = """select delete_master_ds(%(id)s)"""
SET_EDITOR_SQL = """SELECT set_case_editor(%(userid)s, %(caseid)s)"""
LOAD_CASESTATE_SQL = """
SELECT
    s.id,
    s.status,
    s.zugriff,
    getLastCaseAppointment(m.id) AS letzer_termin,
    m.phase AS phase
FROM ka_status_tbl_view s JOIN master_tbl_view m ON m.id = s.master_id
WHERE s.master_id = %(id)s
"""
SET_CASESTATE_SQL = """
UPDATE ka_status_tbl_view
    SET status = %(status)s
    WHERE id = %(id)s"""

SET_CASEPHASE_SQL = """
    UPDATE master_tbl_view
    SET phase = %(phase)s
    WHERE id IN (SELECT master_id FROM ka_status_tbl_view WHERE id = %(id)s)"""
ANONYMIZE_CASE = """SELECT anonymize_case_wrapper(%(id)s)"""
CHECK_REMIND_CONDITION = """\
SELECT id FROM master_tbl_view m WHERE id = %%(id)s AND (%s)"""

def save_getone(params, k, v):
    try:
        return params.getone(k)
    except KeyError:
        return v


class LoadCaseNotExistsError(Exception):

    def __init__(self, value):
        self.value = value

    def __str__(self):
        return repr(self.value)

def get_unknown_value(item):
    """Will return the default value for a instance tree item"""
    meta = item.getMeta()
    if isinstance(meta, DateLeaf):
        default = g.mpuls_config.get('case', 'default-autocomplete-date')
        if default == "NOW":
            default = dd_mm_YYYY(datetime.today())
        value = default
    elif isinstance(meta, IntLeaf):
        value = g.mpuls_config.get('case', 'default-autocomplete-integer')
    elif isinstance(meta, TextLeaf):
        value = g.mpuls_config.get('case', 'default-autocomplete-text')
    elif isinstance(meta, ChoiceNode):
        value = g.mpuls_config.get('case', 'default-autocomplete-choice')
    elif isinstance(meta, RadioNode):
        value = g.mpuls_config.get('case', 'default-autocomplete-choice')
    else:
        value = None
    return value

def get_field_identifier(id, name, it=None, idx=0):
    """Will return the identifier of a field. If the field in in a rg the
    identifier of the first occourance will be returned on default."""
    # Map generic fieldnames to specific fieldnames
    # get name of the field
    nc = g.formedTree.findByName(name)
    if not nc:
        return None

    # is the field within a rg. if so get name
    parent = nc.getParent()
    rg = None
    while parent:
        if isinstance(parent, RepeatNode):
            rg = parent
            break
        parent = parent.getParent()

    if not rg:
        return "%s:%d" % (name, id)

    # get instance of the rg
    rgc = it.getItem("%s:%d" % (rg.getName(), id))
    if not rgc:
        return None
    rgs = rgc.getChildren()
    if not rgs:
        log.debug("No repeat group found for %r: create new", name)
        rg = rgc.createInstance()
    else:
        rg = rgs[idx]
    return '%s:%s:%s' % (name, rg.getInternalIdentifier(), id)


class Field(object):

    """Class providing extra information about an attribute of the case.

    The constructor parameters and public instance variables are:

       name -- The name of the attribute.  This should be the name used
               in the master table in the database which usually means
               that it's the name in the formed meta tree (but not in a
               repeat group).
               The value of the field can be fetched with the case
               object's get_value method

       digest -- If true (default false), the field will be listed in
               the default case digest.

       overview -- If true (default false), the field will be listed in the
               default caseoverview

       info -- If true (default false), the field will be listed in the
               default infofield

       info_label --  Defines a label to be used in the infofield.
               Default will be the name of the field.

       overview_label -- Defines a label to be used in the overviewpage.
               Default will be the name of the field.

       overview_style -- Defines a css style for the fields in the overview
               (default table_name)

       overview_length -- Defines the maximum length of the displayed field.
               Field will be shortend after set number of chars (default 15)

       search_retrieve -- If true (default false), the field will be
               returned from a search.  This only works if the name is a
               column of the master table.

       search_match -- If true (default false), the search string
               entered by the user will be matched against the value of
               this field.  This will only work if the name is a column
               of the master table.  search_match and search_retrieve
               are completely independent settings.

       preload -- If true (default false), the field will be loaded from
               the database explicitly when creating a Case object from
               an id using CaseFactory.loadById.  This flag must be set
               for all fields from the master table that cannot be
               loaded using the formed tree, because otherwise the
               fields won't be available through the case object at all.
               For other fields it can be an optimization to set this
               flag.

       alias -- Additional name under which the attribute can be
               accessed directly as an attribute of the case object.
               Default is None, which means that the field is not
               available under an alias.

       alias_force_string -- If true (default), the value will be forced
               to be a unicode string using the ensure_unicode function
               when it is accessed through the alias name.
               This setting only has an effect if alias is not None.

       session -- If true (default false) the field is available with
               the alias name from the SessionCase object too.
               This setting only has an affect if alias is not None.
               
       order_by -- if true (default false) will be shown as an option in a drop
               down list in the extended search page for sorting
    """

    def __init__(self, name, digest=False,
                 info=False, info_label=None,
                 search_retrieve=False, search_match=False,
                 session=False, preload=False,
                 alias=None, alias_force_string=True,
                 overview=False, overview_label=None,
                 overview_style="table_name",
                 overview_length=15,
                 order_by=False):

        self.name = name
        self.digest = digest
        self.info = info
        self.info_label = info_label or name
        self.search_retrieve = search_retrieve
        self.search_match = search_match
        self.order_by = order_by

        self.session = session
        self.preload = preload
        self.alias = alias
        self.alias_force_string = alias_force_string

        self.overview = overview
        self.overview_label = overview_label or name
        self.overview_style = overview_style
        self.overview_length = overview_length

    def __repr__(self):
        return "Field(%r)" % (self.name,)

    @property
    def default(self):
        """Return the default value when accessed via the alias name"""
        if self.alias_force_string:
            return ""
        else:
            return None

    def convert(self, value):
        """Convert the value when accessed via the alias name.
        If alias_force_string is true, the value is forced to a unicode
        string with the ensure_unicode method.  Otherwise it is returned
        as-is.
        """
        if self.alias_force_string:
            value = ensure_unicode(value)
        return value


class MpulsCaseList:

    def __init__(self, case_ids=None):
        self.case_list = case_ids or []

    def __nonzero__(self):
        # for backwards compatibility even empty MpulsCaseList instances
        # are not considered false.  There's probably no place where an
        # MpulsCaseList instance is used in a boolean context, though,
        # so we shouldn't actually need this method and instances where
        # case_list is empty, should be considered False
        return True

    def __len__(self):
        return len(self.case_list)

    def listDatasetIds(self):
        return self.case_list

    def isEmpty(self):
        return len(self.case_list) == 0

    def numDatasets(self):
        return len(self.case_list)

    def getCases(self, filter=None):
        list = []
        if filter:
            cases = [id for id in self.case_list if id not in filter]
        else:
            cases = self.case_list
        for id in cases:
            list.append(g.case_factory.loadById(id))
        return list


class MpulsCaseBundle(MpulsCaseList):

    def __init__(self, case_ids=None):
        MpulsCaseList.__init__(self, case_ids)

    def setEditor(self, user_id):
        factory = g.case_factory
        num = 0
        for id in self.case_list:
            case = factory.loadById(id)
            try:
                case.setEditor(user_id)
                num += 1
            except Exception, e:
                log.exception(e)
        return num

    def setStandin(self, group_ids):
        factory = g.case_factory
        num = 0
        for id in self.case_list:
            case = factory.loadById(id)
            case_group_ids = group_ids[:]
            if case.getEditor().gid in case_group_ids:
                case_group_ids.remove(case.getEditor().gid)
            try:
                standin = case.getStandin()
                standin.setGroups(case_group_ids)
                num += 1
            except Exception, e:
                log.exception(e)
        return num

    def delete(self):
        factory = g.case_factory
        num = 0
        for id in self.case_list:
            case = factory.loadById(id)
            try:
                case.delete()
                num += 1
            except Exception, e:
                log.exception(e)
        return num

    def anonymize(self):
        factory = g.case_factory
        num = 0
        for id in self.case_list:
            case = factory.loadById(id)
            try:
                case.anonymize()
                num += 1
            except Exception, e:
                log.exception(e)
        return num

    def restore(self):
        factory = g.case_factory
        num = 0
        for id in self.case_list:
            case = factory.loadById(id)
            try:
                case.restore()
                num += 1
            except Exception, e:
                log.exception(e)
        return num

    def markAnonymize(self):
        factory = g.case_factory
        num = 0
        for id in self.case_list:
            case = factory.loadById(id)
            try:
                case.markAnonymize()
                num += 1
            except Exception, e:
                log.exception(e)
        return num

    def markDelete(self):
        factory = g.case_factory
        num = 0
        for id in self.case_list:
            case = factory.loadById(id)
            try:
                case.markDelete()
                num += 1
            except Exception, e:
                log.exception(e)
        return num

    def export(self, type='XML', anonym=True, selection=None):
        factory = g.get_instance_factory()
        num = 0

        xmltree = factory.toXML(self.case_list, selection=selection,
                                anonymize=anonym,
                                external=self._export_additional_data)
        num = len(self.case_list)

        if type == "XML":
            export = self._exportXML(xmltree)
        elif type == "XLS":
            export = self._exportXLS(xmltree, selection)
        elif type == "CSV":
            export = self._exportCSV(xmltree, selection)

        return num, export

    def _export_additional_data(self, case, case_id, cursor, selection,
                                anonymize):
        """Export logbook and documents if selected and not anonymizing.
        """
        if anonymize:
            return
        if "logbook" in selection:
            export_logbook(case, case_id, cursor, selection, anonymize)
        if "documents" in selection:
            export_documents(case, case_id, cursor, selection, anonymize)

    def formletter(self, letterid):
        factory = g.case_factory
        l = []
        o = None
        for id in self.case_list:
            case = factory.loadById(id)
            try:
                d,o = case.formletter(letterid)
                l.append(d)
            except Exception, e:
                log.exception(e)
        return l,o 

    def _exportXML(self, tree):
        root = tree.getroot()
        xml = ElementTree.tostring(root)
        return xml

    def _exportXLS(self, tree, selection):
        return XLSExport(tree, selection).export()

    def _exportCSV(self, tree, selection):
        return CSVExport(tree, selection).export()


class MpulsCaseStandin:

    """This class represents the standin for a case"""

    def __init__(self, case_id=None):
        self.case_id = case_id
        self.group_ids = []
        self._loadFromDB()

    def _loadFromDB(self):
        fields = {'id': self.case_id}
        con, cur = None, None
        try:
            con = db.getConnection()
            cur = con.cursor(cursor_factory=psycopg2.extras.DictCursor)
            cur.execute(LOAD_STANDIN_SQL, fields)
            rows = cur.fetchall()
            for row in rows:
                self.group_ids.append(row['benutzer_id'])
        finally:
            db.recycleConnection(con, cur)

    def getGroups(self):
        """Returns a list of group ids"""
        return self.group_ids

    def setGroups(self, groupid_list):
        """Sets the Standin to the provided list of group ids"""
        old = Set(self.getGroups())
        new = Set(groupid_list)
        self._delete(old - new)
        self._add(new - old)

    def _delete(self, groupid_list):
        con, cur = None, None
        fields = {'caseid': self.case_id, 'groupid': None}
        try:
            con = db.getConnection()
            for uid in groupid_list:
                fields['groupid'] = uid
                cur = con.cursor()
                try:
                    cur.execute(DELETE_STANDIN_SQL, fields)
                except:
                    con.rollback()
                cur.close()
                cur = None
            con.commit()
        finally:
            db.recycleConnection(con, cur)

    def _add(self, groupid_list):
        con, cur = None, None
        fields = {'caseid': self.case_id, 'groupid': None}
        try:
            con = db.getConnection()
            for uid in groupid_list:
                fields['groupid'] = uid
                cur = con.cursor()
                try:
                    cur.execute(ADD_STANDIN_SQL, fields)
                except:
                    con.rollback()
                cur.close()
                cur = None
            con.commit()
        finally:
            db.recycleConnection(con, cur)


class MpulsCase(object):

    def __init__(self, id=None, preset=None, state_id=None, state_status=None,
                 state_access=None, state_phase=None, tags=None,
                 uncommitted=None, editor_id=None):
        self.id = id
        if preset is None:
            preset = dict()
        self.preset = preset
        self.formed_instance = None
        self.phaseslist = None
        self.standin = None
        self.documents = []
        self.appointments = []
        self.tags = tags
        self.uncommitted = uncommitted
        self.privacy_statement = None
        self.editor_id = editor_id
        self.init_state(state_id, state_status, state_access, state_phase)

    def init_state(self, state_id, status, access, phase):
        self.state = MpulsCaseState(state_id, self.id, status, access, phase)

    @classmethod
    def set_fields(cls, raw_fields):
        fields = [Field(**kw) for kw in raw_fields]
        cls.fields = fields
        cls.alias_dict = dict((field.alias, field) for field in fields
                                                   if field.alias is not None)

    @classmethod
    def get_preset_fields(cls):
        """Return a list with the names of formed fields that can be preset.
        The list is intended to be used when populating the preset
        mapping passed to the constructor.
        """
        return [field.name for field in cls.fields if field.alias is not None]

    @classmethod
    def get_preload_fields(cls):
        """Return a list with the names of the fields to preload."""
        return [field.name for field in cls.fields if field.preload]

    @classmethod
    def get_overview_fields(cls):
        """Return a list fields to be displayed in overviewpage."""
        return [field for field in cls.fields if field.overview]

    @classmethod
    def get_session_fields(cls):
        """Return a list describing the fields to store in the session case.
        The items of the list are those items of cls.alias whose
        in_session attribute is true.
        """
        return [field for field in cls.fields if field.session and field.alias]

    @classmethod
    def get_digest_fields(cls):
        """Return a list describing all fields shown in the digest."""
        return [field for field in cls.fields if field.digest]

    @classmethod
    def get_infofield_fields(cls):
        """Return a list describing all fields shown in the infofield."""
        return [field for field in cls.fields if field.info]

    @classmethod
    def get_search_retrieve_fields(cls):
        """Return a list with all names of fields to retrieve from the search"""
        return [field.name for field in cls.fields if field.search_retrieve]

    @classmethod
    def get_search_match_fields(cls):
        """Return a list with all names of fields to match in the search"""
        return [field.name for field in cls.fields if field.search_match]

    def __getattr__(self, name):
        """Provide attribute like access to some fields for compatibility"""
        field = self.alias_dict.get(name)
        log.debug('accessing field %r using name %r' % (field, name))
        if field:
            try:
                value = self.get_value(field.name)
            except KeyError:
                value = field.default
            return field.convert(value)
        raise AttributeError(name)

    def _load_formed_instance_tree(self):
        if self.formed_instance is None:
            factory = self._get_instance_factory()
            self.formed_instance = factory.loadInstanceTreeByIdentifier(
                self.id, uncommited=self.uncommitted)

    def _get_instance_factory(self):
        return g.get_instance_factory()

    def get_formed_instance_tree(self):
        """Return the formed instance tree associated with the case object.
        The instance tree will be loaded if hasn't been loaded yet.

        This method is meant as a way to improve older code that
        directly deals with the formed instance tree. By using the case
        object's instance tree, it's possible in some cases to avoid
        explicitly loading the instance tree separately and redundantly.

        Also, it's possible to perform additional access permission
        checks using the user_may_read and user_may_edit methods.
        """
        self._load_formed_instance_tree()
        return self.formed_instance

    def get_formed_item(self, name):
        self._load_formed_instance_tree()
        # Some fields are not part of the formed tree because they are preset
        # fields of the master_tbl. Do not try to get a field identifier/formed
        # item for these fields as this would cause an error anyway. Ignore an
        # return None 
        if name not in PRESET_FIELDS_IN_MASTERTABLE:
            fid = get_field_identifier(self.id, name, self.formed_instance)
            if fid:
                return self.formed_instance.getItem(fid)
            log.error('(get_formed_item): Can not find %s' % name)
        return None

    def get_description(self, name):
        """Return the description of the formed field given by name"""
        if self.formed_instance is None:
            self._load_formed_instance_tree()
        item = self.get_formed_item(name)
        if item:
            return item.getMeta().getDescription()
        else:
            raise KeyError("Cannot find formed item for %r" % name)

    def get_value(self, name, **kw):
        """Return the value of the formed field given by name.
        If the formed instance of the case has not yet been loaded, the
        value is taken from the preset dictionary the case was
        instantiated with.  If it's not in the preset dictionary, the
        formed tree is loaded.  Once the formed instance has been
        loaded, the value will always be taken from there first. Only if this
        does not work the method takes the value from the preset dictionary.

        This behavior is intended as an optimization so that the case
        can be instantiated from e.g. a search result which returns the
        commonly used fields without any further database queries.  The
        other fields will be loaded on demand when they're actually
        needed.  Once the formed instance has been created it takes
        precedence over the preset fields.
        """
        # always use formed instance tree if available
        if self.formed_instance is not None:
            item = self.get_formed_item(name)
            if item:
                return item.getValue()
            else:
                # Item not found in formed. This my happen for fields
                # which are not defined in the formed description like
                # "bearbeiter_id". Try to look if this value has been
                # preloaded and is located in the preset dict.
                try:
                    return self.preset[name]
                except:
                    raise KeyError("Cannot find formed item for %r" % name)

        # if the instance tree has not yet been loaded, look in preset
        if name in self.preset:
            return self.preset[name]

        # otherwise actually load the formed instance tree and repeat
        # We should only get here if self.formed_instance is not set yet.
        assert self.formed_instance is None
        self._load_formed_instance_tree()
        return self.get_value(name, **kw)

    def get_phaseslist(self):
        if self.phaseslist is None:
            self.phaseslist = PhaseFactory().load(self.id)
        return self.phaseslist

    def getPrivacyStatement(self):
        """Returns the privacy statement of the case"""
        if not self.privacy_statement is None:
            return self.privacy_statement
        try:
            # Create privacy_statement
            self.privacy_statement = PrivacyStatement(self, Agency())
            return self.privacy_statement
        except:
            self.privacy_statement = None
            raise
        return None

    def getSessionCase(self):
        """Return a SessionCase object initialized from the current case"""
        fields = dict()
        for field in self.get_session_fields():
            # FIXME: Ideally we would use
            # self.get_value(field.name) to get the value.  We use
            # getattr instead because this compatibility layer
            # translates None values into strings which some code
            # (e.g. buildInfofield in
            # templates/casemanagement/main.mako) relies on.
            fields[field.alias] = getattr(self, field.alias)
        if g.mpuls_config.is_enabled('case-module', 'privacy'):
            signed = self.getPrivacyStatement().isSigned()
        else:
            signed = True
        return SessionCase(self.id, state=self.state,
                           privacy_statement=signed,
                           **fields)

    def check_consistence(self, fields=None, phase=None):
        """Raises ConsistenceCheckException if the case is not completly
        consistent up to the current phase. A case is consistent if no required
        field are missing and all relevant dates on phases are set.

        fields: List of fieldnames which will be checked. On default all fields
                will be checked
        phase:  Id id phase for which the check will be done.
        """
        self.get_phaseslist().check_consistence(fields, phase)

    def check_anonymizeability(self, fields=None, phase=None):
        """Raises PhaseAnonymizableCheckException or ConsistenceCheckException
        if the case is not in a anonymizable state.
        """

        # 1. Check if anonymisation is allowed in the current phase
        if phase is not None:
            phasepart = self.get_phaseslist().get_phase(phase)
        else:
            phasepart = self.get_phaseslist().get_current_phase()

        # Check if anonymization is allowed in the current phase
        desc = g.mpuls_config.get('phases',
                                  'description')[0].get(phasepart.phase)
        if not phasepart.is_anonymizable():
            raise PhaseAnonymizableCheckException(h.literal(_( \
            u"<p>An anonymisation of the case is not possible in phase %s." \
            u"</p><p>Please remove this case.</p>") % desc))

        # 2. Now determine if we will enforce full consistence or minimal
        # consistence (e.g on abort)
        self._load_formed_instance_tree()
        minimal_condition = g.mpuls_config.get('phases',
                                               'minimal_consistence_conditon')
        context = RuleEvalContext(self.formed_instance, "XXX:%s" % self.id)
        minimal_rule = RuleLeaf()
        minimal_rule.setAttribute('expr', minimal_condition)
        if minimal_rule.getExpr().evaluate(context):
            # Also running cases can be anonymized, but check minimal
            # consistence of case for the phase it will be anonymized.
            min_fields = g.mpuls_config.get('phases',
                                            'minimal_fields_to_anonymize')
            end_phase = determineAnonymizedCaseEndPhase(self)
            self.check_consistence(min_fields, end_phase)
        else:
            #Check full consistence
            self.check_consistence()
            # 3. Check if the case is currently running
            if isinstance(phasepart, StartPhasePart):
                raise ConsistenceCheckException(
                    h.literal(_(
                    "<p>An anonymisation of this case is not possible: " \
                    "It is only possible to end cases, which current phase" \
                    " is finished. The phase %s is not finished. Please " \
                    "finish the phase and try to anonymise again.</p>") % desc))

    def make_consistent(self, endphase=None):
        """Will ensure that the case is in a (minimal) consistent state. That
        means that all missing required fields up to the given or current phase
        are set to 'unknown'.  Missing relevant datefields (start, end) are set
        to the current date.
        """
        phaseslist = self.get_phaseslist()
        curphase = phaseslist.get_current_phase_id()
        self._load_formed_instance_tree()
        data = {}
        for phase in phaseslist:

            start, end = phase.getStart(), phase.getEnd()
            if start.datefield is None or end.datefield is None:
                # ignore phases without datefields.  This is usually
                # only the 'unknown' phase.
                continue

            if endphase is not None:
                if int(start.id) > int(endphase):
                    break
            else:
                if int(start.id) > int(curphase):
                    break

            # set relevant start and enddate to the current date if not already
            # set
            if start.date in [None, ""]:
                key = "%s:%s" % (start.datefield, self.id)
                data[key] = format_date(datetime.today())
            if end.date in [None, ""]:
                key = "%s:%s" % (end.datefield, self.id)
                data[key] = format_date(datetime.today())

            # Fill all other fields with "unknown" values.
            for field in start.fields:
                if data.has_key("%s" % (field.id)):
                    continue
                item = self.formed_instance.getItem(field.id)
                try:
                    value = item.getValue()
                    log.debug('Current value of %s: %s' % (field.id, value))
                    if value in [-1, "-1", None, ""]:
                        def_value = get_unknown_value(item)
                        if def_value:
                            log.debug("Setting %s to %s"
                                      % (field.id, def_value))
                            data["%s" % (field.id)] = def_value
                except:
                    log.warning('Tried to get Value for unknown field %s'
                                % item.getIdentifier())
            for field in end.fields:
                if data.has_key("%s" % (field.id)):
                    continue
                try:
                    item = self.formed_instance.getItem(field.id)
                    value = item.getValue()
                    log.debug('Current value of %s: %s' % (field.id, value))
                    if item.getValue() in [-1, "-1", None, ""]:
                        def_value = get_unknown_value(item)
                        if def_value:
                            log.debug("Setting %s to %s"
                                      % (field.id, def_value))
                            data["%s" % (field.id)] = def_value
                except:
                    log.warning('Tried to get Value for unknown field %s'
                                % item.getIdentifier())

        # Set values
        self.formed_instance.setData(data)
        self.formed_instance.commit()

    def make_anonymizable(self):
        """Will ensure that the case is in a anonymizable state. This means
        that current phase is finished and the case is in a consistent state.
        """
        self._load_formed_instance_tree()
        data = {}

        end_phase = determineAnonymizedCaseEndPhase(self)
        # Ensurse consistence of the case
        self.make_consistent(end_phase)
        # Check if the current phase is already finished. If not then set to
        # finished
        data['%s:%s' % (PHASEFIELD, self.id)] = str(end_phase)
        self.formed_instance.setData(data)
        self.formed_instance.commit()

    def get_subject(self):
        """Returns a short string identifier for the case built from info fields."""
        values = []
        for f in self.get_infofield_fields():
            x = self.get_value(f.name)
            if x:
                values.append(unicode(x))
        return ", ".join(values) 

    def getPhase(self):
        return self.getState().getPhase()

    def set_reachable_phase(self):
        """Will set the maximum reachable phase of the case"""
        phases = self.get_phaseslist()
        current_phase = phases.get_current_phase_id()
        reachable_phase = phases.get_max_phase_id()
        if current_phase != reachable_phase:
            self.setPhase(reachable_phase)

    def setPhase(self, phase):
        r = self.getState().setPhase(phase)
        self.checkValidity()
        return r

    def getState(self):
        return self.state

    def setState(self, value):
        self.state.setState(value)

    def getStandin(self):
        """Returns the standin of the case"""
        if not self.standin is None:
            return self.standin
        try:
            self.standin = MpulsCaseStandin(self.id)
            return self.standin
        except:
            self.standin = None
            raise
        return None

    def getTags(self):
        if self.tags is None:
            self.tags = []
        return self.tags

    def checkValidity(self):
        """Check validity and set tags based on configuration in tagging.xml"""
        self.check_validity_for_ids([self.id])

    @classmethod
    def check_validity_for_ids(cls, caseids):
        """Check validitiy for the datasets for the ids in the sequence caseids
        """
        if not g.mpuls_config.is_enabled('case-module', 'checktags'):
            return
        setter = get_TagSetter(session['USER_AUTHORIZED'], config)
        for caseid in caseids:
            log.debug('Tagging case: %s' % caseid)
            setter.perform(caseid)

    def getValidity(self):
        '''Will return the id of the max validity based on set tags'''
        tags = set(self.getTags())
        # Try to find the highest validity
        validity = '0' # Default validity (Not valid)
        for v in g.mpuls_config.get('validities', 'enabled'):
            req_tags = set(v.get('tags', []))
            if tags.issuperset(req_tags):
                # case has all requried tags for this validity
                # -> level up to next validity
                validity = v.get('id')
            else:
                break
        return validity

    def getDocuments(self):
        """Returns a list of documents which are attached to this case"""
        try:
            # Load list of documents
            self.documents = listDocuments(self.id)
            return self.documents
        except:
            self.documents = []
            raise
        return []

    def user_is_editor(self, uid):
        """Returns true if the given user id is the id of the editor of the case"""
        return self.editor_id == uid

    def user_may_read(self, user):
        """Return whether the user may read the case data.
        This default implementation always returns True. This assumes
        that the database will not let the user see the case and its id
        if the user does not have read permission on the case.

        This default behavior may change in the future, but initially,
        simply returning True is the most backwards compatible
        implementation.

        Derived classes may override this method if they wish to
        restrict access at the web-application level.
        """
        return True

    def user_may_edit(self, user):
        """Return whether the user may edit the case data.

        This default implementation always returns True. This assumes
        that the database will not let the user modify the case if the
        user does not have write permission on the case.

        This default behavior may change in the future, but initially,
        simply returning True is the most backwards compatible
        implementation.

        Derived classes may override this method if they wish to
        restrict access at the web-application level.
        """
        return True

    def getEditor(self):
        """Returns the editor of the case"""
        return UserObject(self.editor)

    def setEditor(self, userid):
        """sets the editor of the case to the given userid"""
        user = UserObject(userid)
        try:
            db.execute(SET_EDITOR_SQL, {'caseid': self.id, 'userid': userid})
        except Exception:
            log.exception("Error setting editor of case %r to %r"
                          % (self.id, userid))
        return self.editor

    def copy(self, new_values=None):
        """Copy the case and all its contents into a new case.

        If the new_values parameter is given it should be a dictionary
        with values to be set in the copy. The keys of this dictionary
        should be names of only to top-level, i.e. non-repeat-group,
        fields and the values should be strings. This method constructs
        a new dictionary where the keys have been converted to instance
        specific keys with get_field_identifier so that it can be passed
        to the new instance tree's setData method.

        The method returns the instance tree of the copy.
        """
        factory = self._get_instance_factory()
        self._load_formed_instance_tree()
        new_it = factory.copy_instance_tree(self.formed_instance)

        if new_values:
            new_id = new_it.getRootNode().getIdentifier()
            new_it.setData(dict((get_field_identifier(new_id, key), value)
                                for key, value in new_values.items()))
            new_it.commit()

        return new_it

    def getMeta(self):
        """Returns the meta object for the case identified with its id"""
        return Meta(self.id)

    def markDelete(self):
        """Mark case for deletion"""
        self.getState().setState(3)

    def delete(self):
        """Delete the case from the data base.
        The method raises exceptions if the case cannot be deleted.
        """
        if g.mpuls_config.is_enabled('case-module', 'meta-server'):
            meta = self.getMeta()
            if meta.is_linked():
                meta.delete_projectdata()

        db.execute(DELETE_CASE_SQL, {'id': self.id})

    def markAnonymize(self):
        """Mark case for anonymization"""
        if g.mpuls_config.is_enabled('case-module', 'meta-server'):
            meta = self.getMeta()
            if meta.is_linked():
                meta.push()

        self.getState().setState(4)

    def anonymize(self):
        """Deleted personal data in the case"""
        if g.mpuls_config.is_enabled('case-module', 'meta-server'):
            meta = self.getMeta()
            if meta.is_linked():
                meta.push()
            if g.mpuls_config.is_enabled('module', 'meta-transmit-reportquota'):
                get_meta_client().transmit_reportquota()

        cur = None
        conn = db.getConnection()
        try:
            cur = conn.cursor()
            cur.execute(ANONYMIZE_CASE, {'id': self.id})
        finally:
            db.recycleConnection(conn, cur)
        state = self.getState()
        state.setState(5)

    def restore(self):
        #If case has no EE then set EE to undefined on restore
        if g.mpuls_config.is_enabled('case-module', 'privacy'):
            ps = self.getPrivacyStatement()
            if not ps.isSigned():
                ps.keepUndefined()
        self.getState().setState(1)

    def formletter(self, id):
        """Returns a dictionary with fields required for a formletter""" 
        fields = {}
        order = []
        letter = None
        conn, cur = None, None

        # Search for formletter
        for letter in g.mpuls_config.get('case', 'formletter'):
            if letter.get('id') == str(id): break

        try:
            try:
                conn = db.getConnection()
                cur = conn.cursor()
                d = {'id': self.id}
                for sql in letter.get('sql'):
                    cur.execute(sql, d)
                    desc = [t[0] for t in cur.description]
                    order.extend(desc)
                    result = cur.fetchone()
                    if result:
                        for n, k in enumerate(desc):
                            fields[k] = result[n]
            finally:
                db.recycleConnection(conn, cur)
        except:
            log.exception(_('Error on fetching data for formletter'))
        return fields, order

    def getHashFields(self):
        """Return a dictionary of fields to be used for the meta case hashvalue.

        The dictionary should contain the following keys: firstname,
        birthname, gender and birthdate. firstname and birthname are
        strings, birthdate is a date object and gender should be an int
        with 0 meaning 'female', 1 meaning male and -1 for the case
        where the gender hasn't been set yet.

        This default implementation always raises NotImplemented.
        mpuls-Applications should override this method if they need
        support for the meta-server.
        """
        raise NotImplementedError

    def is_hashable(self):
        """Returns a true if the case has all information to build a hash."""
        fields = self.getHashFields()
        for k, f in fields.iteritems():
            if f in [-1, "", None]:
                return False
        return True


class SessionCase:

    """This class holds all needed information for the currently selected case
    """

    def __init__(self, id=None, page_id=None, state=None,
                 privacy_statement=False, **kw):
        if id is None:
            self.id = None
        else:
            self.id = int(id)

        self.page_id = page_id
        self.state = state
        self.privacy_statement = privacy_statement

        for field in self._session_fields():
            setattr(self, field.alias, kw.get(field.alias, field.default))

    def setData(self, params):
        for field in self._session_fields():
            setattr(self, field.alias,
                    field.convert(save_getone(params, field.name,
                                              field.default)))
    def getData(self):
        data = {}
        for field in self._session_fields():
            value = self.get_value(field.alias)
            if value is not None:
                data[field.name] = value
        return data

    def _session_fields(self):
        return g.case_factory.get_session_fields()

    def get_value(self, name):
        return getattr(self, name)

    def getState(self):
        return self.state

    # TODO: the [gs]etPageId functions are deprecated and should be not used, as
    # it is not assured that page_id will always be populated. For this, use
    # session['navigation.selectedpage'], which has the same page_id.
    # This means that this information is saved twice and a better solution
    # should be found.
    def setPageId(self, id):
        self.page_id = id

    def getPageId(self):
        return self.page_id

    def hasPrivacyStatement(self):
        return self.privacy_statement


class MpulsCaseFactory:

    """Factory for case object. This factory provides methods to either load
    existing cases from db or create new cases"""

    def __init__(self, case_cls=None):
        if case_cls is None:
            case_cls = Case
        self.case_cls = case_cls

    def get_search_retrieve_fields(self):
        """Frontend for the same method of the factory's case class"""
        return self.case_cls.get_search_retrieve_fields()

    def get_search_match_fields(self):
        """Frontend for the same method of the factory's case class"""
        return self.case_cls.get_search_match_fields()

    def get_session_fields(self):
        """Frontend for the same method of the factory's case class"""
        return self.case_cls.get_session_fields()

    def _build_preset(self, result):
        return dict((name, result[name])
                    for name in self.case_cls.get_preset_fields()
                    if result.has_key(name))

    def loadById(self, id, uncommitted=None):
        """Load a case with the given id. Return a case object.
        The parameter uncommitted should be a dictionary with
        uncommitted fields. It will be passed through to the case class
        constructor.
        """
        # ideally, id should always be an int.  However, sometimes it's
        # given as a unicode object, so we explicitly convert to int.
        # FIXME: the callers should be changed so that id is always an int.
        id = int(id)

        result = None
        conn, c = None, None
        try:
            conn = db.getConnection()
            c = conn.cursor(cursor_factory=psycopg2.extras.DictCursor)
            fields = self.case_cls.get_preload_fields()
            # Ensure that at least one field (id) is loaded if no fields are
            # configured for the case in JSON config.
            if "id" not in fields:
                fields.append('id')
            query = LOAD_CASE_SQL % dict(fields=", ".join(fields))
            fields = {'id': id}
            log.debug("Query: %s", query)
            c.execute(query, fields)
            result = c.fetchone()
            if not result:
                raise LoadCaseNotExistsError("Es konnte keine Fallakte geladen"
                                             " werden.")
        finally:
            db.recycleConnection(conn, c)

        return self.case_cls(id, preset=self._build_preset(result),
                             tags=result['tags'], uncommitted=uncommitted)

    def loadFromQueryResult(self, result):
        """Create a Case object from a DB query result row.

        The result should be a dict-like object mapping the names used
        in the formed tree to the values.  Which fields can be used is
        determined by the list returned by the Case's get_preset_fields
        method.  In addition to those fields, the row must contain the
        following fields that will be used to instantiate the State
        object: s_id, status, zugriff
        """
        return self.case_cls(result["id"],
                             preset=self._build_preset(result),
                             state_id=result["s_id"],
                             state_status=result["status"],
                             state_access=result["zugriff"],
                             state_phase=result["phase"],
                             tags = result["tags"],
                             editor_id = result["bearbeiter_id"])

    def createNew(self, init=True, uuid=None, data=None):
        """Create a new case in the database and return the case object.
        The new case will use the given uuid and contain the data given
        in the data parameter.  If the init parameter is true the case
        will get initialized with some standard information like the
        editor information.
        """
        factory = self._get_instance_factory()
        instance_tree = factory.createNewInstanceTree(uuid)
        id = instance_tree.getRootNode().getIdentifier()

        if data:
            self._updateMasterOnCreate(instance_tree, data)

        if init == True:
            self._initNewCase(id)

        case = self.case_cls(id)
        case.checkValidity()
        # Set hightes reachable phase id phase change mode is half or
        # fullautomatic
        changemode = g.mpuls_config.get('phases', 'changemode')
        if changemode != PC_MANUAL: 
            phases = PhaseFactory().load(case.id)
            reachable_phase = phases.get_max_phase_id()
            case.setPhase(reachable_phase)
        return case

    def _updateMasterOnCreate(self, instance_tree, data):
        id = instance_tree.getRootNode().getIdentifier()
        try:
            formed_data = {}
            nkey = None
            for key, value in data.iteritems():
                nkey = get_field_identifier(id, key, instance_tree)
                if not nkey:
                    continue
                formed_data[nkey] = value

            instance_tree.setData(formed_data)

            session['uncommitted_fields'] = instance_tree.commit()
            session.save()
        except SetDataException, err:
            # FIXME: We should handle this exception in some way.
            # For now, though, it's better to log it instead of silently
            # suppressing it.
            log.exception("SetDataException in _updateMasterOnCreate ignored")
            raise

    def _initNewCase(self, id):
        """Additional initialization of the new case.
        The base-class method does nothing.  Derived classes should
        override this method if necessary.
        """
        pass

    def _get_instance_factory(self):
        return g.get_instance_factory()


class MpulsCaseOverview:

    def __init__(self):
        self.ds_list = []
        self.pages = 0

    def search(self, search_str=None):
        self.ds_list = []
        result = g.case_search.perform(search_str)
        factory = g.case_factory
        for row in result:
            case = factory.loadFromQueryResult(row)
            self.ds_list.append(case)
        return self.ds_list

    def numDatasets(self):
        return len(self.ds_list)

    def getDatasets(self):
        return self.ds_list

    def get_count(self):
        return self.pages

    def set_count(self, pages_count):
        self.pages = pages_count

class MpulsCaseState(object):

    """This class store the state and the last accesstime of the case.
    Possible states are:
        * Open (1)
        * Closes (2)
        * outstanding deletion (3)
        * outstanding anonymision (4)
        * anonymised (5)
    """

    def __init__(self, id=None, master_id=None, state=None, access=None,
                 phase=None):
        self.id = id
        self.master_id = master_id
        self.state = state
        self.access = access
        self.phase = phase
        self._last_date = None

        if self.master_id is not None and self.id is None:
            self._load()

    def getLastDateFromDB(self):
        if self._last_date:
            return self._last_date
        else:
            # Reload all the data. As fetching the last appointment
            # for the case is the most expensive part of the operation getting
            # some more values does not hurt.
            self._load()
            return self._last_date

    last_date = property(getLastDateFromDB)

    def setData(self, id, master_id, state, access):
        self.master_id = master_id
        self.id = id
        self.state = state
        self.access = access

    def _load(self):
        conn, cur = None, None
        try:
            conn = db.getConnection()
            cur = conn.cursor(cursor_factory=psycopg2.extras.DictCursor)
            cur.execute(LOAD_CASESTATE_SQL, {'id': self.master_id})
            result = cur.fetchone()
            self.id = result.get('id')
            self.state = result.get('status')
            self.access = result.get('zugriff')
            self._last_date = result.get('letzer_termin')
            self.phase = result.get('phase')
        except:
            log.exception('Failed loading State')
            raise
        finally:
            db.recycleConnection(conn, cur)

    def getStateLabel(self):
        state_labels = {'1': _('Editable'),
                        '2': _('Editable'),
                        '3': _('Outstanding deleted'),
                        '4': _('Outstanding made anonymous'),
                        '5': _('Make anonymous'),
                        }
        return state_labels.get(str(self.state))

    def isCmFinished(self):
        return str(self.phase) in g.mpuls_config.get('phases', 'finished')

    def getState(self):
        return self.state

    def setState(self, value):
        try:
            db.execute(SET_CASESTATE_SQL, {'status': value, 'id': self.id})
            self.state = value
            return True
        except:
            log.exception("Error while setting state of case %d"
                          % (self.master_id,))
        return False

    def getPhase(self):
        return self.phase

    def setPhase(self, phase):
        try:
            db.execute(SET_CASEPHASE_SQL, {'phase': phase, 'id': self.id})
            self.phase = phase
            return True
        except:
            log.exception("Error while setting phase of case %r to %r"
                          % (self.id, phase))
        return False

    def getPhaseDescription(self):
        desc = []
        phases = PhaseFactory().load(self.master_id)
        phase = phases.get_current_phase()
        desc.append(g.mpuls_config.get('phases',
                                       'description')[0].get(phase.phase))

        # If phase is "unknown" do not append "running/finished"
        if phase.id == "-1":
            return " ".join(desc)
        if isinstance(phase, StartPhasePart):
            desc.append(_('Running'))
        else:
            desc.append(_('Finished'))
        return " ".join(desc)

    def getPhaseName(self):
        phases = PhaseFactory().load(self.master_id)
        return phases.get_current().getDescription()

    def getAccessTime(self):
        return self.access

    def getLastDate(self):
        """Returns the last occuring date from appointments, aids, or
        the last accesstime
        """
        return self.last_date

    def getMaxSavetime(self):
        return Agency().getMaxSavetime()

    def check_remindcondition(self):
        conn, cur = None, None
        remindercondition = Agency().getReminderCondition()
        try:
            try:
                conn = db.getConnection()
                cur = conn.cursor()
                sql = CHECK_REMIND_CONDITION % remindercondition
                fields = {'id': self.master_id}
                cur.execute(sql, fields)
                conn.commit()
                if cur.fetchone():
                    return True
                return False
            finally:
                db.recycleConnection(conn, cur)
        except:
            log.exception('Can not check remind condition')
        return False

    def daysToMaxSavetime(self):
        maxage = self.getMaxSavetime()
        if self.isCmFinished():
            age = datetime.now() - self.getAccessTime()
        else:
            age = datetime.now() - self.getLastDate()

        if self.check_remindcondition():
            return int(maxage) - age.days
        else:
            return 15 # showning counting days until generation of a reminder
                      # starts 14 days before maxsavetime is reached in
                      # template. We do not want such counter until the
                      # remindcondiotion is not true



class MPulsXMLImporter(DBXMLImporter):

    """MPuls specific extension for the libformed XML importer.

    This class extends the base class with the following functionality:

     - Import of logbook entries

     - Triggering validity checks after import.

       To achieve this, the class keeps track of the ids of the imported
       datasets and only when all datasets have been imported
       successfully, it triggers the validity checks.
    """

    UUID_ANONYMIZED_SQL = """SELECT is_uuid_anonymized(%(uuid)s)"""
    STATE_FROM_UUID_SQL = """
SELECT s.status
FROM ka_status_tbl_view s JOIN master_tbl_view m ON m.id = s.master_id
WHERE m.uuid_id = %(uuid)s
"""

    def __init__(self, meta_tree, extra_factories=None,
                 allow_logbook=False, allow_documents=False):
        if extra_factories is None:
            extra_factories = dict()
        if allow_logbook:
            extra_factories["tagebuch"] = get_logbook_import_factory()
        if allow_documents:
            extra_factories["anlagen"] = get_documents_import_factory()
        DBXMLImporter.__init__(self, meta_tree, extra_factories)
        self.imported_ids = []

    def import_xml(self, db, infile):
        """Extend the base class method to reformat rule violation exceptions
        """
        try:
            DBXMLImporter.import_xml(self, db, infile)
        except XMLImportException, e:
            if e.set_data_errors:
                raise XMLImportException(self.format_errors(e.uuid,
                                                            e.set_data_errors))
            raise

        # We should only get here if there were no errors during import.
        # If there were any errors, an XMLImportException will have been
        # raised and the DB changes will have been rolled back and no further
        # handling of the imported cases will be necessary anyway.
        self.postprocess_imported_cases()

    def postprocess_imported_cases(self):
        """Apply various postprocessing functions to the imported case.
        Called after successfully importing all cases. The method can
        use self.imported_ids as a list with the ids of the newly
        imported or modified cases.

        The default implementation handles automatic phase changes if
        they are configured and triggers tagging on all the imported
        cases. Derived classes may extend or override this method to
        change this.
        """
        changemode = g.mpuls_config.get('phases', 'changemode')
        if changemode != PC_MANUAL: 
            # Trigger setting max phase after importing when changemode is not
            # manual. This may be not to performant many cases are important as
            # each case needs to be loaded two time to set the phase. First
            # time to load the phasemodel second time to load the case.
            for case_id in self.imported_ids:
                phases = PhaseFactory().load(case_id)
                reachable_phase = phases.get_max_phase_id()
                case = g.case_factory.loadById(case_id)
                # Setting the phase includes tagging of the case.
                case.setPhase(reachable_phase)
        else:
            # Trigger the tagging of the newly imported data sets.
            g.case.check_validity_for_ids(self.imported_ids)

    def format_errors(self, uuid, error_items):
        out = []
        out.append(h.escape(_("The case with UUID '%s' could not be"
                              " imported due to rule violations") % uuid))
        out.append("<ul>\n")
        for key, errors in error_items.iteritems():
            node_name = key.split(':')[0]
            for error in errors:
                element = g.formedTree.findByName(node_name)
                out.append("<li><strong>%s:</strong> %s\n"
                           % (h.escape(element.getDescription()),
                              h.escape(error.msg)))
        out.append("</ul>\n")
        return h.literal("".join(out))

    def import_case(self, factory, cursor, components):
        if self.import_allowed(cursor, components):
            tree = DBXMLImporter.import_case(self, factory, cursor, components)
            if "tagebuch" in components:
                # only try to import logbook if there actually was a
                # logbook for the case
                import_logbook(tree, cursor, components["tagebuch"])
            if "anlagen" in components:
                import_documents(tree, cursor, components["anlagen"])
            self.imported_ids.append(tree.getRootNode().getIdentifier())
            return tree
        return None

    def import_allowed(self, cursor, components):
        uuid = components["master"].uuid
        if uuid is None:
            # There's no uuid associated with the case to be imported,
            # so we cannot really do any checks and always allow the
            # import.
            return True
        if self.uuid_anonymized(cursor, uuid):
            raise XMLImportException(_(u"The case with UUID '%s' was anonymized"
                                       u" already.") % uuid)
        if not self.state_allows_import(cursor, uuid):
            raise XMLImportException (_(u"The case with UUID '%s' is either"
                                        u" marked for anonymization or marked"
                                        u" for deletion.") % uuid)
        return True

    def uuid_anonymized(self, cursor, uuid):
        """Return whether the case for the uuid has been anonymized already.
        Cases that have been anonymized may not be imported.
        """
        cursor.execute(self.UUID_ANONYMIZED_SQL, dict(uuid=uuid))
        row = cursor.fetchone()
        if not row:
            return False
        return row[0]

    def state_allows_import(self, cursor, uuid):
        """Return whether the current state of the case allows an update.
        If a case exists and it's either marked for deletion or marked
        for anonymization it may not be updated.  Otherwise an update is
        OK.
        """
        cursor.execute(self.STATE_FROM_UUID_SQL, dict(uuid=uuid))
        row = cursor.fetchone()
        if not row:
            # case doesn't exist, so import is possible as far as the state
            # is concerned
            return True
        return row[0] not in (4, 3)

    def version_supported(self, version):
        accepted = self.structure_tree.get_version()
        if accepted  != version:
            raise XMLImportException(_("The XML file has the version '%s',"
                                       " but supported is only version '%s'")
                                     % (version, accepted))
