#!/usr/bin/env python

import sys
from optparse import OptionParser
from datetime import timedelta

import psycopg2 as dbapi


try:
    # Python >= 2.5 includes SQLite 3
    import sqlite3 as db
except ImportError:
    # installed separately in earlier versions
    from pysqlite2 import dbapi2 as db

from itertools import izip

DEFAULT_PORT = 5436

# The relations to be copied, ordered
# topological not to violate foreign key constraints.
DEFAULT_RELATIONS = [
    "master_tbl",
    "ka_status_tbl",
    ]

SELECT_TMPL = "SELECT * FROM %(relation)s"
INSERT_TMPL = "INSERT INTO %(relation)s (%%(fields)s) VALUES (%%(placeholders)s)"

SELECT_LOOGBOOK = '''
SELECT id, master_id, datum, bearbeiter, art, dauer
FROM ka_logbuch_tbl
'''

INSERT_LOGBOOK = '''
INSERT INTO ka_logbuch_tbl (id, master_id, datum, bearbeiter, art, dauer)
       VALUES (%(id)s, %(master_id)s, %(datum)s, %(bearbeiter)s, %(art)s, %(dauer)s)
'''

SELECT_DATADUMP = "SELECT value FROM dbinfo WHERE key = 'datadump'"
UPDATE_DATADUMP = "UPDATE dbinfo SET value = %(value)s WHERE key = 'datadump'"


SELECT_NM_TAGS_CASES = '''
SELECT id, master_id, tag_id, created FROM nm_tags_cases
'''

INSERT_NM_TAGS_CASES = '''
INSERT INTO nm_tags_cases (id, master_id, tag_id, created)
       VALUES (%(id)s, %(master_id)s, %(tag_id)s, %(created)s)
'''

def usage(msg="", exit_code=0):
    print >> sys.stderr, """%s
usage:
%s -i <sqlite db> -i <postgres db> [-p <port>]
""" % (msg, sys.argv[0])
    sys.exit(exit_code)


def copy_relation(relation, in_cur, out_cur):
    relation = {'relation': relation}

    select_stmnt = SELECT_TMPL % relation
    insert_tmpl  = INSERT_TMPL % relation

    stmnt = None
    fields = []

    in_cur.execute(select_stmnt)

    while True:
        row = in_cur.fetchone()
        if not row:
            break

        if stmnt is None:
            placeholders = []
            values = {}
            for idx, x in enumerate(row):
                name = in_cur.description[idx][0]
                values[name] = x
                fields.append(name)
                placeholders.append("%%(%s)s" % name)
            stmnt = insert_tmpl % {
                'fields'      : ','.join(fields),
                'placeholders': ','.join(placeholders)
                }
        else:
            values = dict(izip(fields, row))

        out_cur.execute(stmnt, values)

def copy_datadump(in_cur, out_cur):
    print >> sys.stderr, "copy datadump"

    in_cur.execute(SELECT_DATADUMP)
    while True:
        row = in_cur.fetchone()
        if not row:
            break
        out_cur.execute(UPDATE_DATADUMP, {'value': row[0]})

def copy_logbook(in_cur, out_cur):
    print >> sys.stderr, "copy ka_logbuch_tbl"

    in_cur.execute(SELECT_LOOGBOOK)
    while True:
        row = in_cur.fetchone()
        if not row:
            break
        id, master_id, datum, bearbeiter, art, dauer = row

        if dauer is not None:
            dauer = timedelta(seconds=dauer)
        out_cur.execute(INSERT_LOGBOOK, {
            'id'        : id,
            'master_id' : master_id,
            'datum'     : datum,
            'bearbeiter': bearbeiter,
            'art'       : art,
            'dauer'     : dauer
            })


def copy_nm_tags_cases(in_cur, out_cur):
    print >> sys.stderr, "copy nm_tags_cases"

    in_cur.execute(SELECT_NM_TAGS_CASES)

    while True:
        row = in_cur.fetchone()
        if not row:
            break

        id, master_id, tag_id, created = row
        out_cur.execute(INSERT_NM_TAGS_CASES, {
            'id'        : id,
            'master_id' : master_id,
            'tag_id'    : tag_id,
            'created'   : created
            })

def parse_relation_list(option, opt, value, parser):
    relations = getattr(parser.values, option.dest)
    if value:
    	setattr(parser.values, option.dest, relations + value.split(','))


def main():
    parser = OptionParser()
    parser.set_defaults(port=DEFAULT_PORT, relations=DEFAULT_RELATIONS)
    parser.add_option("-i", "--in", dest="indb",
                      help="Name of input database file")
    parser.add_option("-o", "--out", dest="outdb",
                      help="Name of output PostgreSQL database")
    parser.add_option("-p", "--port", type="int",
                      help="Port number")
    parser.add_option("-r", "--relations", action="callback",
                      callback=parse_relation_list, type="string",
                      help="Comma-separated list of relations")
    parser.add_option("-l", "--copy-logbook",
                      action="store_true", dest="copylogbook", default=False,
                      help="Copy logbook")

    opts, args = parser.parse_args()

    if not opts.indb:
        usage("error: missing input db", 1)
    if not opts.outdb:
        usage("error: missing output db", 1)
    if not opts.port:
        usage("error: missing port", 1)

    in_con = in_cur = out_con = out_cur = None
    try:
        in_con = db.connect(opts.indb)
        in_cur = in_con.cursor()

        out_con = dbapi.connect(database=opts.outdb, port=opts.port)
        out_cur = out_con.cursor()

        try:
            for relation in opts.relations:
                copy_relation(relation, in_cur, out_cur)
            copy_datadump(in_cur, out_cur)
            if opts.copylogbook:
                copy_logbook(in_cur, out_cur)
            copy_nm_tags_cases(in_cur, out_cur)
            out_con.commit()
        except:
            out_con.rollback()
            raise
    finally:
        for c in (in_cur, in_con, out_cur, out_con):
            if c:
                c.close()


if __name__ == '__main__':
    main()
