# -*- coding: utf-8 -*-
#
# Copyright (c) 2007 - 2011 by Intevation GmbH
#
# FormEd is free software under the terms of the
# GNU General Public License Version 3. See file
# LICENSE coming with the source of FormEd for details.
#
# Authors:
# Torsten Irlaender   <torsten.irlaender@intevation.de>
# Sascha L. Teichmann <teichmann@intevation.de>
# Frank Koormann      <frank.koormann@intevation.de>
# Björn Ricks         <bjoern.ricks@intevation.de>
# Roland Geider       <roland.geider@intevation.de>
#

import sys
import codecs
from time import gmtime, strftime

from formed.meta import data
from formed.meta.nodecomponents import Node, Leaf, \
VISIT_IGNORE_CHILDREN, VISIT_CONTINUE
from formed.meta.structure import structure_path
from formed.export.rules import ExportSQLForVirtualFields

from formed.export.sql import *

class ExportCreateViews:

    def __init__(self, document):
        self.document = document
        self.root = document.root

    @staticmethod
    def getDescription():
        return "Export document as SQL statements"

    def printRepeats(self, tables):
        arr = []
        for t in tables:
            root = t.root
            if not root:
                continue
            parent = t.parent
            joins = []
            while parent:
                joins.append(parent.name)
                parent = parent.parent
            linear = list(root.firstLevel(lambda c:
                                          isinstance(c, (data.RepeatNode,
                                                         data.PageNode))))
            print "%s:" % t.name
            arr.append((t, joins, linear))
            if joins:
                print "\tjoins: %s" % " -> ".join(joins)
            if linear:
                print "\tlinear: %s" % ", ".join([p.getName() for p in linear])
        return arr



    def doExport(self, path):
        root = self.root

        out = None
        tableGrants = ""
        creator = SchemaCreator(root)
        creator.createSchema()
        tables = creator.topological()
        arr = self.printRepeats(tables)

        caseAnonymizer = CaseAnonymizer()

        try:
            out = codecs.open(path, "wb", "UTF-8")
            pfeil = out.write
            pfeil(u'--\n-- Schema generated by Libformed\n-- %s\n'
                  % strftime("%a, %Y-%m-%d %H:%M:%S +0000", gmtime()))
            pfeil('-- # Revision: $Revision$\n--\n')
            pfeil('-- # tables: %d\n--\n' % len(tables))

            pfeil('BEGIN;')
            pfeil('SET ROLE :db_adm;')

            pfeil('\n--\n-- GRANTS ON TABLES\n--\n')
            pfeil(tableGrants)

            # generate ro-views for master table and repeat groups
            pfeil("--\n--\n-- views for master table and repeat groups\n--\n")
            virtual_sql_export = ExportSQLForVirtualFields(self.document)
            for raw_table, joins, linear in arr:
                oldjoins = joins
                table = raw_table.name
                tab = sqlName(table)
                pfeil("CREATE OR REPLACE VIEW %s_view\nAS SELECT \n" % tab)

                str_columns = []
                for column in raw_table.columns:
                    if isinstance(column, VirtualColumn):
                        sql = virtual_sql_export.doExport(column.name)
                        if sql:
                            str_columns.append("     ( %s ) as %s" % (sql, column.name))
                        else:
                            str_columns.append("     -- missing definition for %s" % column.name)
                            str_columns.append("     NULL:VARCHAR as %s" % column.name)
                    else:
                        str_columns.append("     %s.%s AS %s" % (tab, 
                                                                 column.name,
                                                                 column.name))
                pfeil(",\n".join(str_columns))
                pfeil("\nFROM %s\n" % tab)

                njoins = joins
                oldTable = table

                while njoins:
                    found = False
                    for t in arr:
                        raw_table2, joins, linear2 = t
                        nname = raw_table2.name
                        if nname == njoins[0]:
                            found = True
                            break
                    if not found:
                        break

                    njoins = joins

                    pfeil("    INNER JOIN %s\n" % nname)
                    pfeil("    ON         %s.%s_id = %s.id\n"
                          % (oldTable, removeTbl(nname), nname))
                    oldTable = nname
                pfeil(VIEW_WHERE_CLAUSE)
                pfeil("\n")


                pfeil(UPDATE_RULE % (tab, tab, tab))
                pfeil(",\n".join(["    %s = NEW.%s" % (c.name, c.name)
                                  for c in raw_table.columns if c.accessible]))
                pfeil("\n")
                pfeil("WHERE id = NEW.id;\n\n")

                pfeil(FORBID_DELETE_RULE % (tab, tab))
                pfeil("\n")
                pfeil(FORBID_INSERT_RULE % (tab, tab))
                pfeil("\n")

                vn = "%s_view" % tab
                pfeil(GRANT_SELECT % (vn, ':cm_group'))
                pfeil(GRANT_UPDATE % (vn, ':cm_group'))
                pfeil(GRANT_SELECT % (vn, ':admin_group'))
                pfeil("\n")

                # anonymization view
                self._writeAnonymizationView(pfeil, raw_table, tab)

                caseAnonymizer.addAnoymizer(RelationAnonymizer(raw_table,
                                                               table,oldjoins))

            caseAnonymizer.writeFunction(pfeil)
            # Create Master Table Evaluation View
            self._writeEvalTotalView(pfeil)
            pfeil('SET ROLE postgres;')
            pfeil('COMMIT;\n')
        finally:
            if out:
                try:
                    out.close()
                except:
                    pass

    def _writeEvalTotalView(self, pfeil):
        collector = EvaluateColumnsCollector()
        self.root.visit(collector.visitor)
        columns = collector.columns
        statustblname = "ka_status_tbl"

        pfeil("--\n-- view for master table evaluation\n--\n")
        pfeil("-- this table should be copied to 04views.sql\n--\n")
        pfeil("-- \n")
        pfeil("-- Notice: the evaluation server might need more fields than\n")
        pfeil("-- are generated here and will generate SQL errors if not\n")
        pfeil("-- edited. Likely candidates are fkz and bearbeiter_id\n")
        pfeil("-- \n")

        pfeil("-- CREATE OR REPLACE VIEW %s\n" % EVALUATION_VIEW)
        pfeil("--    AS SELECT\n")
        pfeil("--     m.id AS id,\n")
        if columns:
            pfeil("--     m.phase AS phase,\n")
        else:
            pfeil("--     m.phase AS phase\n")
        pfeil(",\n".join(["--     m.%s AS %s" % (column, column)
                    for column in columns]))
        pfeil("\n")
        pfeil("--    FROM master_tbl m JOIN %s s\n" % statustblname)
        pfeil("--        ON m.id = s.master_id\n")
        pfeil("--        WHERE\n")
        pfeil("--            s.status <> 3 AND\n")
        pfeil("--            (m.bearbeiter_id = (getuserid())::integer)\n")
        pfeil("--            OR pg_has_role('admin','MEMBER');\n")
        pfeil("--\n")
        pfeil("-- " + GRANT_SELECT % (EVALUATION_VIEW, ':cm_group'))
        pfeil("-- " + GRANT_SELECT % (EVALUATION_VIEW, ':admin_group'))
        pfeil("\n")

    def _writeAnonymizationView(self, pfeil, raw_table, table):
        anonym = []

        for c in raw_table.columns:
            if not c.func:
                anonym.append(c.name)
            else:
                if c.func == "suppress_value":
                    if c.anon_value == 'NULL':
                        datatype = c.datatype.split()[0] # a bit hackish
                        anonym.append('NULL::%s AS %s' % (datatype, c.name))
                    else:
                        anonym.append('%s AS %s' % (c.anon_value, c.name))
                else:
                    args = [x.strip() for x in c.func.split(':')]
                    if len(args) > 1:
                        anonym.append("%s(%s, %s) AS %s"
                                      % (args[0], c.name, ", ".join(args[1:]),
                                         c.name))
                    else:
                        anonym.append("%s(%s) AS %s"
                                      % (args[0], c.name, c.name))
        pfeil("CREATE OR REPLACE VIEW %s_anonym_view\nAS SELECT\n"
              "    %s FROM %s_view;\n\n"
              % (table, ",\n    ".join(anonym), table))

        vn = "%s_anonym" % table
        pfeil(FORBID_DELETE_RULE % (vn, vn))
        pfeil("\n")
        pfeil(FORBID_INSERT_RULE % (vn, vn))
        pfeil("\n")

        vn = "%s_anonym_view" % table
        pfeil(GRANT_SELECT % (vn, ':cm_group'))
        pfeil(GRANT_SELECT % (vn, ':admin_group'))
        pfeil("\n")

# vim:set ts=4 sw=4 si et sta sts=4:
