# -*- coding: utf-8 -*-
#
# Copyright (c) 2007 - 2011 by Intevation GmbH
#
# FormEd is free software under the terms of the
# GNU General Public License Version 3. See file
# LICENSE coming with the source of FormEd for details.
#
# Authors:
# Torsten Irlaender   <torsten.irlaender@intevation.de>
# Sascha L. Teichmann <teichmann@intevation.de>
# Frank Koormann      <frank.koormann@intevation.de>
# Björn Ricks         <bjoern.ricks@intevation.de>
#

import sys
import codecs
from time import gmtime, strftime

from formed.meta import data
from formed.meta.nodecomponents import Node, Leaf, \
VISIT_IGNORE_CHILDREN, VISIT_CONTINUE
from formed.meta.structure import structure_path
from formed.export.rules import ExportSQLForVirtualFields

VIEW_WHERE_CLAUSE = \
"""WHERE
    master_tbl.bearbeiter_id = getuserid()
    OR master_tbl.id IN (
        SELECT nm.master_id from nm_benutzer_master_tbl nm
        WHERE nm.benutzer_id IN (
            SELECT nm2.gruppe_id from nm_benutzer_gruppe_tbl nm2
            WHERE nm2.benutzer_id = getuserid()
        )
    )
    OR pg_has_role('admin','MEMBER');
"""

FORBID_DELETE_RULE = \
"""CREATE OR REPLACE RULE %s_delete_rule AS ON DELETE TO %s_view
    DO INSTEAD NOTHING;
"""

FORBID_INSERT_RULE = \
"""CREATE OR REPLACE RULE %s_insert_rule AS ON INSERT TO %s_view
    DO INSTEAD NOTHING;
"""

FORBID_UPDATE_RULE = \
"""CREATE OR REPLACE RULE %s_update_rule AS ON UPDATE TO %s_view
    DO INSTEAD NOTHING;
"""

UPDATE_RULE = \
"""CREATE OR REPLACE RULE %s_update_rule AS ON UPDATE TO %s_view
DO INSTEAD UPDATE %s SET
"""

GRANT_SELECT = \
"""GRANT SELECT ON %s TO GROUP %s;
"""

GRANT_INSERT = \
"""GRANT INSERT ON %s TO GROUP %s;
"""

GRANT_UPDATE = \
"""GRANT UPDATE ON %s TO GROUP %s;
"""

ANON_FUNCTION_HEADER = \
"""--
-- Function to anonymize a case and all of its repeat groups
--
CREATE OR REPLACE FUNCTION anonymize_case(mid int)
RETURNS void AS
$$
BEGIN
    PERFORM anonymize_uuid(mid);
"""

ANON_UPDATE = \
"""    UPDATE %s_view SET
        %s
    WHERE %s;

"""

ANON_FUNCTION_FOOTER = \
"""END;
$$ LANGUAGE plpgsql VOLATILE SECURITY DEFINER;

REVOKE ALL ON FUNCTION anonymize_case(int) FROM public;
GRANT EXECUTE ON FUNCTION anonymize_case(int) TO GROUP :admin_group;

"""

EVALUATION_VIEW= 'master_tbl_eval_total_view'

def removeTbl(s):
    return s.replace("_tbl", "")


def sqlQuote(name):
    return name.replace("'", "''")


def sqlName(name):
    return name.replace("-", "_")


class EvaluateColumnsCollector(object):

    def __init__(self, root=None):
        self.columns = []
        self.root = root

    def visitor(self, nc):
        if isinstance(nc, Node):
            if isinstance(nc, (data.RadioNode, data.ChoiceNode)):
                self.checkEvaluate(nc)
                return VISIT_IGNORE_CHILDREN
            if isinstance(nc, data.RepeatNode):
                return VISIT_IGNORE_CHILDREN
        elif (isinstance(nc, Leaf)
              and not isinstance(nc, (data.RuleLeaf, data.ExternalChoiceListLeaf,
                                      data.InfoLeaf))):
            self.checkEvaluate(nc)

        return VISIT_CONTINUE

    def checkEvaluate(self, nc):
        flags = nc.getFlags()
        if flags.find('evaluation') >= 0:
            self.columns.append(sqlName(nc.getName()))

class RelationAnonymizer:

    def __init__(self, raw_table, table, joins):
        self.name = table
        anonym = []
        self.joins = joins

        for c in raw_table.columns:
            if not c.func:
                # no set needed
                continue
            if c.func == "suppress_value":
                anonym.append('%s = %s' % (c.name, c.anon_value))
            else:
                args = [x.strip() for x in c.func.split(':')]
                if len(args) > 1:
                    anonym.append("%s = %s(%s, %s)"
                                  % (c.name, args[0], c.name,
                                     ", ".join(args[1:])))
                else:
                    anonym.append("%s = %s(%s)" % (c.name, args[0], c.name))

        if not anonym: # empty -> not sets
            self.processed = True
            self.set_columns = ""
        else:
            self.processed = False
            self.set_columns = ",\n        ".join(anonym)

    def writeUpdate(self, pfeil, where_clause="master_id = mid"):
        if self.processed: return
        self.processed = True
        if len(self.joins) == 2:
            where_clause = \
                ("""%s_id IN (SELECT id FROM %s WHERE master_id = mid)"""
                 % (self.joins[0][:-4], self.joins[0]))
        elif len(self.joins) > 2:
            # TODO: x->y->master_tbl
            # should result in
            #  x_id IN (SELECT x.id FROM x JOIN y
            #           ON x.y_id = y.id AND y.master_id = mid)
            raise NotImplementedError
        pfeil(ANON_UPDATE % (self.name, self.set_columns, where_clause))


class CaseAnonymizer:

    def __init__(self):
        self.anonymizers = []

    def addAnoymizer(self, anonymizer):
        self.anonymizers.append(anonymizer)

    def _findMaster(self):
        for rel in self.anonymizers:
            if rel.name == 'master_tbl':
                return rel
        return None

    def writeFunction(self, pfeil):
        master = self._findMaster()
        if master is None:
            print >> sys.stderr, "master_tbl not found"
            return
        pfeil(ANON_FUNCTION_HEADER)
        pfeil("    -- anonymize master table\n")
        master.writeUpdate(pfeil, "id = mid")

        pfeil("    -- anonymize repeat groups\n")
        for anonymizer in self.anonymizers:
            anonymizer.writeUpdate(pfeil)

        pfeil(ANON_FUNCTION_FOOTER)


class Column:

    def __init__(self, name, datatype, accessible=True, func=None,
                 anon_value='NULL'):
        self.name = name
        self.datatype = datatype
        self.accessible = accessible
        self.func = func
        self.anon_value = anon_value

    def definition(self):
        return "%s %s" % (self.name, self.datatype)

class VirtualColumn(Column):
    pass

class Table:

    def __init__(self, name, parent=None, root=None):
        self.name = name
        self.columns = []
        self.constraints = []
        self.inserts = []
        self.depends = set()
        self.parent = parent
        self.root = root

    def addDependency(self, dependency):
        self.depends.add(dependency)

    def appendColumn(self, column):
        self.columns.append(column)

    def appendConstraint(self, constraint):
        self.constraints.append(constraint)

    def appendInsert(self, ins):
        self.inserts.append(ins)

    def getName(self):
        return self.name

    def getSchema(self):
        out = ["CREATE TABLE %s (\n    " % self.name]
        out.append(",\n    ".join([column.definition()
                                   for column in self.columns if not isinstance(column, VirtualColumn)]))
        if self.constraints:
            out.append(",\n    ")
            out.append(",\n    ".join(self.constraints))

        out.append("\n);")
        return "".join(out)

    def getInserts(self):
        if not self.inserts:
            return ""
        out = []
        for ins in self.inserts:
            out.append("INSERT INTO %s %s;" % (self.name, ins))
        return "\n".join(out)

    def equalContent(self, other):
        # XXX: This is a bit to simple for general case
        if len(self.inserts) != len(other.inserts):
            return False
        ains = [a.lower() for a in self.inserts]
        bins = [b.lower() for b in other.inserts]
        ains.sort()
        bins.sort()
        return ains == bins


class SchemaCreator:

    def __init__(self, root):
        self.root = root
        self.tableStack = []
        self.tables = []

    def findSameContent(self, table):
        for t in self.tables:
            if t.equalContent(table):
                return t
        return None

    def createSchema(self):
        table = Table("master_tbl", root=self.root)
        table.appendColumn(Column("id", "SERIAL PRIMARY KEY NOT NULL", False))
        table.appendColumn(Column("uuid_id", "CHAR(36)", False))
        table.appendColumn(Column("bearbeiter_id", "int NOT NULL", False))

        self.maybe_add_fkz_column(table)

        table.appendConstraint("UNIQUE (uuid_id)")
        self.tableStack = [table]
        self._createRecursive(self.root)
        self.tables.append(self.tableStack.pop())

    def maybe_add_fkz_column(self, table):
        """Add column 'fkz' to table, if it wouldn't be added automatically.

        The master_tbl.fkz column is a special case because there is
        code related to the evaluation server that requires it, and
        therefore it must be present in all mpuls applications. Some
        meta trees explicitly define such a field (WASKA, WASKO), others
        do not, so it must be added in some cases.

        This method adds the fkz column only if the meta tree doesn't
        contain an fkz field that would end up in master_tbl and that is
        a text field. If the meta tree describes an fkz field for the
        master_tbl which is not text, that's likely an error, so a
        warning is printed to stderr.
        """
        for nc in self.root.walk():
            if nc.getName() == "fkz":
                # if nc is not directly in the master_tbl, i.e. if it is
                # contained in a repeat group, ignore it.
                if len(structure_path(nc)) > 1:
                    continue

                if isinstance(nc, data.TextLeaf):
                    return
                print >> sys.stderr, ("Tree contains node component named 'fkz'"
                                      " which is not a TextLeaf. Column 'fkz'"
                                      "may be duplicated.")
                break

        table.appendColumn(Column("fkz", "VARCHAR(60)", False))

    def topological(self):
        free, bound = [], []
        for t in self.tables:
            if not t.depends:
                free.append(t)
            else:
                bound.append(t)

        out = []
        while free:
            f = free.pop()
            out.append(f)
            tmp = []
            for i in range(len(bound)):
                t = bound[i]
                try:
                    t.depends.remove(f.name)
                    if not t.depends:
                        tmp.insert(0, i)
                except:
                    pass
            for i in tmp:
                free.append(bound.pop(i))

        if bound:
            print >> sys.stderr, \
                "WARNING: cyclic dependencies on %d tables." % len(bound)
            for t in bound:
                deps = ", ".join(sorted(t.depends))
                print >> sys.stderr, "  %s: %s" % (t.name, deps)
            # append them for debugging purposes
            out.extend(bound)
        return out

    def _createRecursive(self, node):
        for child in node.children:
            self._createChild(child)

    def _createDate(self, date):
        name = date.getName()
        if not name:
            return
        name = sqlName(name)
        table = self.tableStack[-1]
        table.appendColumn(Column(name, "DATE", func=date.getFunction()))

    def _createBool(self, bool):
        name = bool.getName()
        if not name:
            return
        name = sqlName(name)
        table = self.tableStack[-1]
        kvs = {-1: 'Keine Angabe', 0: 'Nein', 1: 'Ja'}
        max_length = max([len(v) for v in kvs.itervalues()]) + 1

        choiceTable = Table(name + '_tbl')
        choiceTable.appendColumn(Column('id', 'SERIAL PRIMARY KEY NOT NULL'))
        choiceTable.appendColumn(Column('value',
                                        'VARCHAR(%d) NOT NULL' % max_length))

        for k in sorted(kvs.iterkeys()):
            choiceTable.appendInsert("(id, value) VALUES (%d, '%s')"
                                     % (k, sqlQuote(kvs[k])))

        other = self.findSameContent(choiceTable)

        if other:
            into = other.name
        else:
            into = name + "_tbl"
            self.tables.append(choiceTable)

        table = self.tableStack[-1]

        table.appendColumn(Column(name, 'INTEGER DEFAULT -1',
                                  func=bool.getFunction()))

        table.appendConstraint("FOREIGN KEY (%s) REFERENCES %s (id)"
                               % (name, into))
        table.addDependency(into)

    def _createText(self, text):
        name = text.getName()
        if not name:
            return
        name = sqlName(name)
        table = self.tableStack[-1]
        length = text.getMaxLength()
        if length:
            table.appendColumn(Column(name, "VARCHAR(%s)" % str(length),
                                      func=text.getFunction()))
        else:
            table.appendColumn(Column(name, "VARCHAR", func=text.getFunction()))

    def _createVirtual(self, text):
        name = text.getName()
        if not name:
            return
        name = sqlName(name)
        table = self.tableStack[-1]
        table.appendColumn(VirtualColumn(name, 
                                  "TEXT", #FIXME: datatype is used for
                                          #anonymization. Return NULL::TEXT here
                                          #but virtual may represent all kind of
                                          #datatypes
                                  accessible=False, 
                                  func=text.getFunction()))

    def _createTextArea(self, text):
        name = text.getName()
        if not name:
            return
        name = sqlName(name)
        table = self.tableStack[-1]
        table.appendColumn(Column(name, "TEXT", func=text.getFunction()))

    def _createInteger(self, integer):
        name = integer.getName()
        if not name:
            return
        name = sqlName(name)
        table = self.tableStack[-1]
        table.appendColumn(Column(name, "INTEGER", func=integer.getFunction()))

    def _createNumeric(self, numeric):
        name = numeric.getName()
        if not name:
            return
        name = sqlName(name)
        table = self.tableStack[-1]
        table.appendColumn(Column(name, "NUMERIC", func=numeric.getFunction()))

    def _createChoice(self, choice):
        name = choice.getName()
        if not name:
            return
        name = sqlName(name)

        if choice.getMultiple():
            print >> sys.stderr, "WARNING: No support for multiple choices."
            return

        kvs = {}
        stack = choice.children[:]
        default_value = None
        try:
            while stack:
                c = stack.pop()
                if isinstance(c, data.ExternalChoiceListLeaf):
                    cc = c.getChildren()
                    if cc:
                        stack.extend(cc)
                elif isinstance(c, data.GroupNode):
                    stack.extend(c.children)
                elif isinstance(c, data.ConditionalNode):
                    stack.extend(c.children)
                elif isinstance(c, data.BoolLeaf):
                    desc = c.getDescription()
                    if c.getChecked() and default_value == None:
                        default_value = c.getValue()
                        is_checked = True
                    elif c.getChecked():
                        print >> sys.stderr, "WARNING: Check set twice in '%s'" % name
                    if not desc:
                        desc = ""
                    kvs[int(c.getValue())] = desc
        except ValueError:
            print >> sys.stderr, "WARNING choice '%s' contains none integer values" % name
            return

        if not kvs:
            print >> sys.stderr, "WARNING choice '%s' contains no values" % name
            return

        max_length = max([len(v) for v in kvs.itervalues()]) + 1

        choiceTable = Table(name + '_tbl')
        choiceTable.appendColumn(Column('id', 'SERIAL PRIMARY KEY NOT NULL'))
        choiceTable.appendColumn(Column('value',
                                        'VARCHAR(%d) NOT NULL' % max_length))

        for k in sorted(kvs.iterkeys()):
            choiceTable.appendInsert("(id, value) VALUES (%d, '%s')"
                                     % (k, sqlQuote(kvs[k])))

        if choiceTable.inserts:
            other = self.findSameContent(choiceTable)
        else:
            other = None

        if other:
            into = other.name
        else:
            into = name + "_tbl"
            self.tables.append(choiceTable)

        table = self.tableStack[-1]

        if default_value == None:
            default_value = "-1"
        table.appendColumn(Column(name, 'INTEGER DEFAULT '
                                        + str(default_value) ,
                                  func=choice.getFunction(), anon_value=-1))


        table.appendConstraint("FOREIGN KEY (%s) REFERENCES %s (id)"
                               % (name, into))
        table.addDependency(into)


    def _createRadioGroup(self, radio):
        name = radio.getName()
        if not name:
            return
        name = sqlName(name)
        selected = radio.getSelected()
        if selected:
            found = False
            for child in radio.children:
                if child.getName() == selected:
                    found = True
                    break
            if not found:
                print >> sys.stderr, \
                    "WARNING: Don't found selected in radio '%s'" % name
                selected = None

        max_length = max([len(c.getDescription()) for c in radio.children]) + 1

        radioTable = Table(name + '_tbl')
        radioTable.appendColumn(Column('id', 'INTEGER PRIMARY KEY NOT NULL'))
        radioTable.appendColumn(Column('value',
                                       'VARCHAR(%d) NOT NULL' % max_length))

        default_idx = None

        for i, child in enumerate(radio.children):
            cname = child.getName()
            try:
                value = int(child.getValue())
            except ValueError:
                print >> sys, "WARNING: cannot convert radio '%s' value '%s' to integer" % (
                    cname, child.getValue())
                continue

            if default_idx is None and (child.getChecked()
                                        or cname == selected):
                default_idx = value

            radioTable.appendInsert("(id, value) VALUES (%d, '%s')"
                                    % (value, sqlQuote(child.getDescription())))

        other = self.findSameContent(radioTable)

        if other:
            into = other.name
            print >> sys.stderr, "Found radio '%s" % into
        else:
            into = name + "_tbl"
            self.tables.append(radioTable)

        table = self.tableStack[-1]

        table.appendColumn(Column(name,
                                  (default_idx is not None
                                   and ('INTEGER DEFAULT %d' % default_idx)
                                   or 'INTEGER'),
                                  func=radio.getFunction()))

        table.appendConstraint("FOREIGN KEY (%s) REFERENCES %s (id)"
                               % (name, into))

        table.addDependency(into)

    def _createRepeatNode(self, group):
        name = group.getName()

        if name:
            name = sqlName(name)
            parent = self.tableStack[-1]
            table = Table("%s_tbl" % name, parent, group)
            table.appendColumn(Column("id", "SERIAL PRIMARY KEY NOT NULL",
                                      False))
            table.appendColumn(Column("%s_id" % removeTbl(parent.name),
                                      "INTEGER NOT NULL", False))
            table.appendColumn(Column("uuid_id", "CHAR(36)", False))
            table.appendConstraint("FOREIGN KEY (%s_id) REFERENCES %s (id)"
                                   " ON DELETE CASCADE"
                                   % (removeTbl(parent.name), parent.name))
            table.appendConstraint("UNIQUE (uuid_id)")
            table.addDependency(parent.name)
            self.tableStack.append(table)

            for child in group.children:
                self._createChild(child)

            self.tables.append(self.tableStack.pop())
        else:
            for child in group.children:
                self._createChild(child)

    def _createChild(self, child):
        if isinstance(child, Leaf):
            if isinstance(child, data.NumericLeaf):
                self._createNumeric(child)
            elif isinstance(child, data.IntLeaf):
                self._createInteger(child)
            elif isinstance(child, data.DateLeaf):
                self._createDate(child)
            elif isinstance(child, data.TextLeaf):
                self._createText(child)
            elif isinstance(child, data.TextAreaLeaf):
                self._createTextArea(child)
            elif isinstance(child, data.BoolLeaf):
                self._createBool(child)
            elif isinstance(child, data.VirtualLeaf):
                self._createVirtual(child)
        elif isinstance(child, Node):
            if isinstance(child, data.RepeatNode):
                self._createRepeatNode(child)
            elif isinstance(child, data.RadioNode):
                self._createRadioGroup(child)
            elif isinstance(child, data.ChoiceNode):
                self._createChoice(child)
            else:
                self._createRecursive(child)


class ExportAsSQL:

    def __init__(self, document):
        self.document = document
        self.root = document.root

    @staticmethod
    def getDescription():
        return "Export document as SQL statements"

    def printRepeats(self, tables):
        arr = []
        for t in tables:
            root = t.root
            if not root:
                continue
            parent = t.parent
            joins = []
            while parent:
                joins.append(parent.name)
                parent = parent.parent
            linear = list(root.firstLevel(lambda c:
                                          isinstance(c, (data.RepeatNode,
                                                         data.PageNode))))
            print "%s:" % t.name
            arr.append((t, joins, linear))
            if joins:
                print "\tjoins: %s" % " -> ".join(joins)
            if linear:
                print "\tlinear: %s" % ", ".join([p.getName() for p in linear])
        return arr



    def doExport(self, path):
        root = self.root

        out = None
        tableGrants = ""
        creator = SchemaCreator(root)
        creator.createSchema()
        tables = creator.topological()
        arr = self.printRepeats(tables)

        caseAnonymizer = CaseAnonymizer()

        try:
            out = codecs.open(path, "wb", "UTF-8")
            pfeil = out.write
            pfeil(u'--\n-- Schema generated by Libformed\n-- %s\n'
                  % strftime("%a, %Y-%m-%d %H:%M:%S +0000", gmtime()))
            pfeil('-- # Revision: $Revision$\n--\n')
            pfeil('-- # tables: %d\n--\n' % len(tables))

            pfeil('BEGIN;')

            pfeil('\n--\n-- DROP TABLES\n--\n')
            for t in reversed(tables):
                pfeil('-- DROP TABLE %s CASCADE;\n' % t.name)

            pfeil('\n--\n-- CREATE TABLES\n--\n')
            for table in tables:
                pfeil("%s\n" % table.getSchema())

            pfeil('\n--\n-- INSERT DATA\n--\n')
            for table in tables:
                inserts = table.getInserts()
                if inserts:
                    pfeil("%s\n\n" % inserts)
                    tableGrants = tableGrants \
                        + (GRANT_SELECT % (table.getName(), ':cm_group')) \
                        + (GRANT_SELECT % (table.getName(), ':admin_group'))

            pfeil('\n--\n-- GRANTS ON TABLES\n--\n')
            pfeil(tableGrants)

            # generate ro-views for master table and repeat groups
            pfeil("--\n--\n-- views for master table and repeat groups\n--\n")
            virtual_sql_export = ExportSQLForVirtualFields(self.document)
            for raw_table, joins, linear in arr:
                oldjoins = joins
                table = raw_table.name
                tab = sqlName(table)
                pfeil("CREATE OR REPLACE VIEW %s_view\nAS SELECT \n" % tab)

                str_columns = []
                for column in raw_table.columns:
                    if isinstance(column, VirtualColumn):
                        sql = virtual_sql_export.doExport(column.name)
                        if sql:
                            str_columns.append("     ( %s ) as %s" % (sql, column.name))
                        else:
                            str_columns.append("     -- missing definition for %s" % column.name)
                            str_columns.append("     NULL:VARCHAR as %s" % column.name)
                    else:
                        str_columns.append("     %s.%s AS %s" % (tab, 
                                                                 column.name,
                                                                 column.name))
                pfeil(",\n".join(str_columns))
                pfeil("\nFROM %s\n" % tab)

                njoins = joins
                oldTable = table

                while njoins:
                    found = False
                    for t in arr:
                        raw_table2, joins, linear2 = t
                        nname = raw_table2.name
                        if nname == njoins[0]:
                            found = True
                            break
                    if not found:
                        break

                    njoins = joins

                    pfeil("    INNER JOIN %s\n" % nname)
                    pfeil("    ON         %s.%s_id = %s.id\n"
                          % (oldTable, removeTbl(nname), nname))
                    oldTable = nname
                pfeil(VIEW_WHERE_CLAUSE)
                pfeil("\n")


                pfeil(UPDATE_RULE % (tab, tab, tab))
                pfeil(",\n".join(["    %s = NEW.%s" % (c.name, c.name)
                                  for c in raw_table.columns if c.accessible]))
                pfeil("\n")
                pfeil("WHERE id = NEW.id;\n\n")

                pfeil(FORBID_DELETE_RULE % (tab, tab))
                pfeil("\n")
                pfeil(FORBID_INSERT_RULE % (tab, tab))
                pfeil("\n")

                vn = "%s_view" % tab
                pfeil(GRANT_SELECT % (vn, ':cm_group'))
                pfeil(GRANT_UPDATE % (vn, ':cm_group'))
                pfeil(GRANT_SELECT % (vn, ':admin_group'))
                pfeil("\n")

                # anonymization view
                self._writeAnonymizationView(pfeil, raw_table, tab)

                caseAnonymizer.addAnoymizer(RelationAnonymizer(raw_table,
                                                               table,oldjoins))

            caseAnonymizer.writeFunction(pfeil)
            # Create Master Table Evaluation View
            self._writeEvalTotalView(pfeil)
            pfeil('COMMIT;\n')
        finally:
            if out:
                try:
                    out.close()
                except:
                    pass

    def _writeEvalTotalView(self, pfeil):
        collector = EvaluateColumnsCollector()
        self.root.visit(collector.visitor)
        columns = collector.columns
        statustblname = "ka_status_tbl"

        pfeil("--\n-- view for master table evaluation\n--\n")
        pfeil("-- this table should be copied to 04views.sql\n--\n")
        pfeil("-- \n")
        pfeil("-- Notice: the evaluation server might need more fields than\n")
        pfeil("-- are generated here and will generate SQL errors if not\n")
        pfeil("-- edited. Likely candidates are fkz and bearbeiter_id\n")
        pfeil("-- \n")

        pfeil("-- CREATE OR REPLACE VIEW %s\n" % EVALUATION_VIEW)
        pfeil("--    AS SELECT\n")
        pfeil("--     m.id AS id,\n")
        if columns:
            pfeil("--     m.phase AS phase,\n")
        else:
            pfeil("--     m.phase AS phase\n")
        pfeil(",\n".join(["--     m.%s AS %s" % (column, column)
                    for column in columns]))
        pfeil("\n")
        pfeil("--    FROM master_tbl m JOIN %s s\n" % statustblname)
        pfeil("--        ON m.id = s.master_id\n")
        pfeil("--        WHERE\n")
        pfeil("--            s.status <> 3 AND\n")
        pfeil("--            (m.bearbeiter_id = (getuserid())::integer)\n")
        pfeil("--            OR pg_has_role('admin','MEMBER');\n")
        pfeil("--\n")
        pfeil("-- " + GRANT_SELECT % (EVALUATION_VIEW, ':cm_group'))
        pfeil("-- " + GRANT_SELECT % (EVALUATION_VIEW, ':admin_group'))
        pfeil("\n")

    def _writeAnonymizationView(self, pfeil, raw_table, table):
        anonym = []

        for c in raw_table.columns:
            if not c.func:
                anonym.append(c.name)
            else:
                if c.func == "suppress_value":
                    if c.anon_value == 'NULL':
                        datatype = c.datatype.split()[0] # a bit hackish
                        anonym.append('NULL::%s AS %s' % (datatype, c.name))
                    else:
                        anonym.append('%s AS %s' % (c.anon_value, c.name))
                else:
                    args = [x.strip() for x in c.func.split(':')]
                    if len(args) > 1:
                        anonym.append("%s(%s, %s) AS %s"
                                      % (args[0], c.name, ", ".join(args[1:]),
                                         c.name))
                    else:
                        anonym.append("%s(%s) AS %s"
                                      % (args[0], c.name, c.name))
        pfeil("CREATE OR REPLACE VIEW %s_anonym_view\nAS SELECT\n"
              "    %s FROM %s_view;\n\n"
              % (table, ",\n    ".join(anonym), table))

        vn = "%s_anonym" % table
        pfeil(FORBID_DELETE_RULE % (vn, vn))
        pfeil("\n")
        pfeil(FORBID_INSERT_RULE % (vn, vn))
        pfeil("\n")

        vn = "%s_anonym_view" % table
        pfeil(GRANT_SELECT % (vn, ':cm_group'))
        pfeil(GRANT_SELECT % (vn, ':admin_group'))
        pfeil("\n")

# vim:set ts=4 sw=4 si et sta sts=4:
