# -*- coding: utf-8 -*-
# Copyright 2009 Intevation GmbH, Germany, <info@intevation.de>
# Authors:
# Sascha L. Teichmann <teichmann@intevation.de>
#

import sys
import traceback

from formed.meta.data import CaseNode, ConditionalNode, VirtualLeaf
from formed.meta.exprtree import VariableProvider
from formed.meta.semantic import checkAndConvert, check, SemanticError, \
     ErrorItem
from formed.instance.tree import InstanceTree, InstanceNode, \
     RepeatInstanceContainerNode, RepeatInstanceNode, SetDataException,\
     IDENTIFIER_SEPARATOR


def extract_meta_path(nc):
    path = []
    while nc:
        path.insert(0, nc)
        if isinstance(nc, CaseNode):
            break
        nc = nc.parent
    return path


class RuleEvalContext(VariableProvider):

    def __init__(self, instance_tree, reference, shadow=None):
        VariableProvider.__init__(self)

        self.instance_tree = instance_tree
        self.reference = reference
        if shadow is None:
            shadow = {}
        self.shadow = shadow

    def getValueForVar(self, var, reference=None):
        # TODO: Do some caching here!
        if reference is None:
            reference = self.reference
        item = self.instance_tree.getItem(reference, var)
        if item:
            id = item.getIdentifier()
            try:
                value = self.shadow[id]
            except KeyError:
                value = item.getValue()
            return (id, value)
        else:
            return (var, None)

    def getAllValuesForVar(self, var):
        # TODO: Do some caching here!
        items = []
        shadow = self.shadow
        for item in self.instance_tree.getAllItemsForName(var):
            id = item.getIdentifier()
            try:
                value = shadow[id]
            except KeyError:
                value = item.getValue()
            items.append((id, value))
        return items

    def get_repeat_group_instances(self, name):
        """Return the ids of the instances of the named repeat group.
        """
        item = self.instance_tree.getItem(self.reference, name)
        if isinstance(item, RepeatInstanceNode):
            item = item.parent
        if not isinstance(item, RepeatInstanceContainerNode):
            raise ValueError("RG node for %r is %r, which is not an instance of"
                             " RepeatInstanceContainerNode"
                             % (name, item))
        return [child.getIdentifier() for child in item.getChildren()]


def best_changes(changes, applicable):
    queue = [tuple(changes)]
    already = set([queue[0]])
    last_length = len(changes) + 1
    total_fails = dict([(c, 0) for c in changes])
    while queue:
        changes = queue.pop(0)
        failed = applicable(changes)
        if not failed:
            return changes
        for fail in failed:
            total_fails[fail] += 1
        L = len(changes)
        if L < last_length:
            last_length = L
            already.clear()
        if L < 2:
            continue
        xchanges = sorted(list(changes),
                          lambda a, b: cmp(total_fails[b], total_fails[a]))
        for i in xrange(L):
            reduced = xchanges[0:i] + xchanges[i+1:]
            reduced.sort()
            reduced = tuple(reduced)
            if reduced not in already:
                already.add(reduced)
                queue.append(reduced)
    return []

class ArrayIterator(object):

    '''Iterator which can be ask "has_next()"'''

    def __init__(self, array, idx=0):
        self.array = array
        self.idx = idx

    def fork(self):
        return ArrayIterator(self.array, self.idx)

    def has_next(self):
        return self.idx < len(self.array)

    def current(self):
        return self.array[self.idx]

    def current_is_last(self):
        return self.idx == len(self.array) - 1

    def next(self):
        if not self.has_next():
            raise StopIteration()
        x = self.array[self.idx]
        self.idx += 1
        return x


def evaluate(variables, formed, rule_available, variable_provider_factory,
             name_from_key, find_best_changes=True):
    """Determine the maximum set of changes that still conforms to the rules.

    The parameter variables should be a dictionary mapping the names of
    formed instance items to the corresponding values.  The keys in this
    dictionary are used in several places:

      - To determine the formed meta name of an item by passing it to
        name_from_key (see below)

      - As keys in the return values described below

    The parameter formed is the formed meta tree for the case.

    The parameter variable_provider_factory is a callable that is called
    with an instance item's key and a dictionary of tentative new values
    and should return an instance of a VariableProvider.

    The parameter rule_available is a callable that is called with a
    rule and a variable provider returned from the
    variable_provider_factory and should return a boolean indicating
    whether the rule is available given the values in the variable
    provider.

    The parameter name_from_key is a callable that returns the name of
    the formed meta item given the key of the item in the variables
    dictionary.

    If find_best_changes is true (the default), and rules are violated
    the function tries to find a maximal

    The function returns a tuple with two items, changes and errors.
    The errors item is a dictionary containing information about the
    error rules that were violated.  It maps the keys from the variables
    dictionary for which errors were found to lists of ErrorItem
    instances which describe the errors.

    The contents of the changes item depend on the find_best_changes
    parameter.  If find_best_changes is true (the default) and rules are
    violated, the changes item contains a maximal subset of the original
    changes described by the variables parameter which satisfies all
    rules.  Otherwise the changes item simply contains all of the
    variables, including the ones that lead to rule violations.
    """

    errors = {}

    def applicable(changes):
        shadow = dict(changes)
        failed = []
        for item in changes:
            key, value = item
            tkey = name_from_key(key)
            rules = [rule for rule in formed.name2rules.get(tkey, ())
                          if rule.isError()]
            if not rules:
                continue
            context = variable_provider_factory(key, shadow)
            rules = [rule for rule in rules
                          if rule_available(rule, context)]
            for rule in rules:
                try:
                    result = rule.getExpr().evaluate_instances(context)
                except:
                    traceback.print_exc(file=sys.stderr)
                    result = [(False, {key: None})]
                for success, variables in result:
                    if not success and key in variables:
                        erroritems = errors.setdefault(key, [])
                        msg = rule.getValue()
                        found = False
                        for i in erroritems:
                            if i.msg == msg:
                                found = True
                                break
                        if not found:
                            erroritems.append(ErrorItem(ErrorItem.RULE, key,
                                                        msg, value))
                        failed.append(item)
        return failed

    changes = variables.items()
    if find_best_changes:
        changes = best_changes(variables.items(), applicable)
    else:
        applicable(changes)
    return (changes, errors)

def rule_available(formed, rule, eval_context):
    """Return whether the rule is currently available.
    A rule is available if its isAlwaysAvailable method returns
    true.  Otherwise the rule is available if it does not depend on
    any field contained in an inactive conditional.  A conditional
    is active if its expression evaluates to true.  The expressions
    are evaluated using eval_context.
    """
    if rule.isAlwaysAvailable():
        return True
    for var in rule.getExpr().getDependencies():
        for conditional in formed.name2conditionals[var]:
            available = conditional.getExpr().evaluate(eval_context)
            if not available:
                return False
    return True


class AbstractInstanceTree(InstanceTree):

    def __init__(self, formed, root=None, uncommited=None):
        InstanceTree.__init__(self, formed, root)
        self.new_values = {}
        if uncommited is None:
            uncommited = {}
        self.uncommited = uncommited

    def getItem(self, identifier, name=None):
        if not identifier:
            return None

        tokens = identifier.split(IDENTIFIER_SEPARATOR)
        if len(tokens) < 2:
            return None
        try:
            nc = self.formed.name2nc[name and name or tokens[0]]
        except KeyError:
            return None

        meta_path = extract_meta_path(nc)

        meta_iter = ArrayIterator(meta_path)
        cur_meta = meta_iter.next()

        tokens.pop(0)
        tokens.reverse()

        ids_iter = ArrayIterator(tokens)
        cur_id = ids_iter.next()

        children = [self.root]

        while children:
            instance = children.pop()
            if instance.getMeta() != cur_meta:
                # only interessed in matching metas
                continue

            if isinstance(instance, RepeatInstanceContainerNode):
                # no ids left -> we search the container
                if cur_id is None:
                    return instance
                # keep meta and search same meta in children
                children = instance.getChildren()[:]
                continue

            # Is it a repeat group?
            # yes: compare its internal identifier with the current token
            if instance.isRepeat():
                if (cur_id is not None
                    and str(instance.getInternalIdentifier()) == cur_id):
                    # found matching id
                    if not meta_iter.has_next():
                        # no further meta -> it is the search node
                        return instance

                    cur_meta = meta_iter.next()

                    if not ids_iter.has_next():
                        # last token -> no further rgs inside
                        cur_id = None
                    else:
                        # next rg id
                        cur_id = ids_iter.next()

                    # advance search in children of this rg
                    children = instance.getChildren()[:]
                continue

            # Is it a instance node?
            if isinstance(instance, InstanceNode):
                # no more meta -> found it
                if not meta_iter.has_next():
                    return instance
                cur_meta = meta_iter.next()
                # advance search in children
                children = instance.getChildren()[:]
                continue

            # It is an lnstance leaf
            if not meta_iter.has_next():
                # if no further metas -> found
                return instance

        return None

    def getAllItemsForName(self, name):
        if not name:
            return
        try:
            nc = self.formed.name2nc[name]
        except KeyError:
            return None

        meta_path = extract_meta_path(nc)
        meta_iter = ArrayIterator(meta_path)

        found = []

        stack = [(meta_iter, [self.root])]

        while stack:
            m_iter, children = stack.pop()

            while children:
                inst = children.pop()
                if inst.getMeta() != m_iter.current():
                    continue

                if m_iter.current_is_last():
                    found.append(inst)
                    break

                if isinstance(inst, RepeatInstanceContainerNode):
                    stack.extend([(m_iter.fork(), [c])
                                  for c in inst.getChildren()])
                    continue

                if isinstance(inst, InstanceNode):
                    m_iter.next()
                    children = inst.getChildren()[:]
                    continue

                # should not happen: same meta but remaining path
                break

        return found

    def evaluate_conditional(self, conditional, ident=None):

        # Conditional may be an instance or meta node (in choicelists e.g)
        if isinstance(conditional, ConditionalNode): 
            expr = conditional.getExpr()
            identifier = ident 
        else:
            expr = conditional.getMeta().getExpr()
            identifier = conditional.getIdentifier() 
        
        if expr is not None:
            return expr.evaluate(RuleEvalContext(self, identifier))
        return False

    def rule_available(self, rule, eval_context):
        return rule_available(self.formed, rule, eval_context)

    def evaluate(self, variables, find_best_changes=True):
        return evaluate(variables, self.formed, self.rule_available,
                        lambda key, shadow: RuleEvalContext(self, key, shadow),
                        lambda key: key.split(IDENTIFIER_SEPARATOR, 1)[0],
                        find_best_changes=find_best_changes)

    def setData(self, data, convert=True, find_best_changes=True):
        name2nc = self.formed.name2nc

        # TODO: Use a ConditionalVisitor to filter out
        # the fields which are not mutable in current state.

        ndata = {}
        widgets = {}
        for key, value in data.iteritems():
            tkey = key.split(IDENTIFIER_SEPARATOR, 1)[0]
            try:
                widgets[tkey] = name2nc[tkey]
                ndata[key] = value
            except KeyError:
                #print >> sys.stderr, "missing key: %s" % repr(key)
                pass

        semantic_errors = {}
        values = {}

        # try to convert to native types
        for key, value in data.iteritems():
            tkey = key.split(IDENTIFIER_SEPARATOR, 1)[0]
            w = widgets[tkey]
            # Ignore virtual fields as can not be saved.i
            if isinstance(w, VirtualLeaf): continue
            try:
                if convert:
                    value = value.strip()
                    if not value:
                        v = None
                    else:
                        v = self.check_and_convert(w, value)
                else:
                    v = self.check(w, value)
                values[key] = v
            except SemanticError, err:
                semantic_errors.setdefault(key, []) \
                    .append(ErrorItem(ErrorItem.SEMANTIC, key, err.value,
                                      value))

        # save old invalid semantic errors from all uncommitted values
        # that have not been passed in again in data.
        old_semantic_errors = {}
        for k, items in self.uncommited.iteritems():
            if k not in data:
                l = [i for i in items if i.type == ErrorItem.SEMANTIC]
                if l:
                    old_semantic_errors[k] = l

        # extract rule violations from uncommited
        for k, items in self.uncommited.iteritems():
            if k in values:
                continue # ignore because a new value is given
            for i in items:
                if i.type == ErrorItem.RULE:
                    values[k] = i.bad
                    break

        changes, eval_errors = self.evaluate(values,
                                            find_best_changes=find_best_changes)

        # merge accepted changes
        self.new_values.update(changes)

        # merge semantic with eval errors
        for k, v in semantic_errors.iteritems():
            eval_errors.setdefault(k, []).extend(v)

        # merge old semantic with eval errors
        for k, v in old_semantic_errors.iteritems():
            eval_errors.setdefault(k, []).extend(v)

        self.uncommited = eval_errors

        if eval_errors:
            raise SetDataException(eval_errors)

    def check_and_convert(self, nc, value):
        return checkAndConvert(nc, value)

    def check(self, nc, value):
        return check(nc, value)

    def force_data(self, data):
        """Adds the data in data without any checks.
        This method should only be used in special circumstances, as one
        would normally want to at least check the values.
        """
        self.new_values.update(data)

    def commit(self):
        return self.uncommited

    def hasErrors(self, identifier=None):
        if identifier is None:
            return bool(self.uncommited)
        return identifier in self.uncommited

    def getErrors(self, identifier=None):
        if identifier is None:
            return self.uncommited
        return self.uncommited.get(identifier)


# vim:set ts=4 sw=4 si et sta sts=4 enc=utf-8 fenc=utf-8 :
