# -*- coding: utf-8 -*-
# Copyright 2009 Intevation GmbH, Germany, <info@intevation.de>
# Authors:
# Sascha L. Teichmann <teichmann@intevation.de>
#

import sys
import traceback
from datetime import date
import logging

try:
    # 2.5 includes it
    import xml.etree.ElementTree as ET
except ImportError:
    # installed separately in earlier versions
    import elementtree.ElementTree as ET

import simplejson
import psycopg2 as dbapi

from formed.meta.nodecomponents import Node, VISIT_IGNORE_CHILDREN, \
     VISIT_CONTINUE
from formed.meta.data import RepeatNode, PageNode, ChoiceNode, RadioNode, \
     BoolLeaf, InfoLeaf, TextAreaLeaf, DateLeaf, IntLeaf, TextLeaf, RuleLeaf, \
     WidgetCollector, NumericLeaf, ReferenceNode, VirtualLeaf
from formed.instance.tree import RootNode, InstanceNode, \
     RepeatInstanceNode, RepeatInstanceContainerNode, PageInstanceNode, \
     InstanceLeaf
from formed.instance.factories import InstanceTreeFactory
from formed.instance.backends.database import DBException, DB
from formed.instance.backends.common import AbstractInstanceTree


log = logging.getLogger(__name__)


LIBFORMED_XML_CONTAINER_VERSION = "5"

# SQL statements
CREATE_CASE = '''SELECT create_master_tbl(%(uuid)s)'''
GET_CASE_STRUCTURE = '''SELECT get_case_structure(%(case_id)s)'''

LOAD_DATA = '''\
SELECT %(projection)s FROM %(relation)s WHERE id = %(dataset_id)s'''
UPDATE_DATA = '''UPDATE %s SET %s WHERE id = %%(id)s'''

CREATE_RG_INSTANCE = '''SELECT create_%s(%%(mid)s, %%(uuid)s)'''
DELETE_RG_INSTANCE = '''SELECT delete_%s(%%(rid)s)'''

SELECT_XML_FIELDS = "SELECT %s FROM %s WHERE id = %%(dataset_id)s"

XML_ENCODERS = {
    unicode: lambda s: s,
    date: lambda s: "%04d-%02d-%02d" % (s.year, s.month, s.day)
}


class PostgreSQLInterface(DB):

    """Extend the DBInterface with a direct connection to postgresql"""

    def __init__(self, dbname, user, password, host, port):
        DB.__init__(self)
        self.dbname = dbname
        self.user = user
        self.password = password.replace("\\", "\\\\")
        self.host = host
        self.port = int(port)
        self.connection = None

    def acquire_cursor(self):
        if self.connection is None:
            self.connection = dbapi.connect(database=self.dbname,
                                            host=self.host, port=self.port,
                                            user=self.user,
                                            password=self.password)
        return self.connection.cursor()

    def release_cursor(self, cur):
        cur.close()

    def do_rollback(self):
        # if self.connection is still None, no cursor was acquired and
        # there's nothing to do, actually.
        if self.connection is not None:
            self.connection.rollback()

    def do_commit(self):
        # if self.connection is still None, no cursor was acquired and
        # there's nothing to do, actually.
        if self.connection is not None:
            self.connection.commit()


class StructureNode(object):

    """Models the structure of the master table and the repeat groups. """

    def __init__(self, name, id, children):
        self.name = name
        self.id = id
        self.children = children

    def __repr__(self):
        return self.recursive_repr()

    def recursive_repr(self, depth=0):
        out = []
        indent = "  " * depth
        children = []
        for child in self.children:
            children.append(child.recursive_repr(depth + 1))

        out.append('%s{ "name": "%s", "id": %d' % (indent, self.name, self.id))
        if children:
            children = ",\n".join(children)
            out.append(', "children": [\n')
            out.append("%s" % children)
            out.append("\n%s]}" % indent)
        else:
            out.append(" }")

        return "".join(out)


def to_structure_node(dct):
    try:
        name = dct['name']
        db_id = dct.get('id')
        if db_id is not None:
            db_id = int(db_id)
        children = dct.get('children')
        if children is None:
            children = []
        return StructureNode(name, db_id, children)
    except KeyError:
        return dct


def get_case_structure(cur, case_id):
    try:
        cur.execute(GET_CASE_STRUCTURE, {'case_id': case_id})
        row = cur.fetchone()

        if not row:
            raise DBException("cannot load case structure")
        root = simplejson.loads(row[0], object_hook=to_structure_node)
        return root
    except:
        traceback.print_exc(file=sys.stderr)
        raise


def master_id_from_uuid(cur, uuid):
    """Return a master id for the given uuid.  Return None if not found."""
    cur.execute("SELECT get_masterid_from_uuid('master', %(uuid)s)",
                {"uuid": uuid})
    row = cur.fetchone()
    if not row:
        return None
    return row[0]


def loadRepeatData(db, meta, relation, dataset_id, load_uuid=False):
    collector = WidgetCollector(meta)
    meta.visit(collector.visitor)

    if not collector.widgets:
        return {}

    names = [nc.getName() for nc in collector.widgets]
    if load_uuid:
        names.append("uuid_id")
    projection = ",".join(names)
    select = LOAD_DATA % {'projection': projection,
                          'relation': relation + "_view",
                          'dataset_id': dataset_id}
    cur = None
    try:
        cur = db.acquire_cursor()
        cur.execute(select)
        row = cur.fetchone()
        if row:
            return dict(zip(names, row))
    finally:
        db.release_cursor(cur)

    return {}


def buildChildren(db, data, parent, structure_node):
    children = []
    for meta_child in parent.meta.children:
        if isinstance(meta_child, (ChoiceNode, RadioNode)):
            instance_child = InstanceLeaf(meta_child,
                                          data.get(meta_child.getName()))
            instance_child.parent = parent
            children.append(instance_child)
        elif isinstance(meta_child, RepeatNode):
            rg_children = []
            rg_name = meta_child.getName().replace("-", "_") + "_tbl"
            container = DBRepeatInstanceContainerNode(meta_child, rg_children,
                                                      rg_name, db)
            container.parent = parent
            children.append(container)

            rgs = [rg for rg in structure_node.children if rg.name == rg_name]
            for rg in rgs:
                rg_instance = LazyRepeatInstanceNode(meta_child, None,
                                                     rg.id, db, rg)
                rg_instance.parent = container
                rg_children.append(rg_instance)
        elif isinstance(meta_child, Node):
            sub = InstanceNode(meta_child)
            sub.parent = parent
            children.append(sub)
            buildChildren(db, data, sub, structure_node) # recurse
        elif isinstance(meta_child, (BoolLeaf, InfoLeaf, TextAreaLeaf,
                                     TextLeaf, IntLeaf, DateLeaf, NumericLeaf,
                                     VirtualLeaf)):
            instance_child = InstanceLeaf(meta_child,
                                          data.get(meta_child.getName()))
            instance_child.parent = parent
            children.append(instance_child)
        elif isinstance(meta_child, (RuleLeaf, ReferenceNode)):
            pass
        else:
            log.warning("formed.instance.backends.postgres.buildChildren:"
                        " unhandled child: %r", meta_child)

    parent.children = children


class DBRootNode(RootNode):

    def __init__(self, meta=None, children=None, identifier=None):
        RootNode.__init__(self, meta, children, identifier)

    def getInternalIdentifier(self):
        return self.identifier


class DBRepeatInstanceContainerNode(RepeatInstanceContainerNode):

    def __init__(self, meta, children=None, relation=None, db=None):
        RepeatInstanceContainerNode.__init__(self, meta, children)
        self.relation = relation
        self.db = db

    def getMasterIdentifier(self):
        cur = self.parent
        while cur:
            if cur.isRepeat():
                return cur.getInternalIdentifier()
            cur = cur.parent
        return None

    def createInstance(self, uuid=None):
        mid = self.getMasterIdentifier()
        if mid is None:
            return None
        mid = int(mid)
        stmnt = CREATE_RG_INSTANCE % self.relation
        cur = None
        self.db.begin()
        try:
            try:
                cur = self.db.acquire_cursor()
                cur.execute(stmnt, {'mid': mid, 'uuid': uuid})
                row = cur.fetchone()
                if not row:
                    raise DBException("Cannot fetch id for new repeat group")
                nid = str(row[0])
                self.db.commit()
            except:
                self.db.rollback()
                raise
        finally:
            self.db.release_cursor(cur)

        nnode = LazyRepeatInstanceNode(self.meta, None, nid, self.db,
                                       StructureNode(self.relation, nid, []))
        nnode.parent = self
        self.children.append(nnode)
        return nnode

    def deleteInstance(self, instance_node):
        try:
            self.children.remove(instance_node)
        except ValueError:
            return

        stmnt = DELETE_RG_INSTANCE % self.relation
        rid = int(instance_node.getInternalIdentifier())

        cur = None
        self.db.begin()
        try:
            try:
                cur = self.db.acquire_cursor()
                cur.execute(stmnt, {'rid': rid})
                self.db.commit()
            except:
                self.db.rollback()
                raise
        finally:
            self.db.release_cursor(cur)


class DBRepeatInstanceNode(RepeatInstanceNode):

    def __init__(self, meta, children, identifier):
        RepeatInstanceNode.__init__(self, meta, children, identifier)

    def getInternalIdentifier(self):
        return self.identifier


class LazyRepeatInstanceNode(RepeatInstanceNode):

    def __init__(self, meta, children, identifier, db, structure_node):
        RepeatInstanceNode.__init__(self, meta, children, identifier)
        self.db = db
        self.structure_node = structure_node
        self.uuid_id = None

    def __repr__(self):
        return ("<LazyRepeatInstanceNode 0x%x id=%r>"
                % (id(self), self.identifier))

    def getInternalIdentifier(self):
        return self.identifier

    def getUUID(self):
        # Call getChildren to trigger a lazy load to make sure the uuid
        # has been loaded
        self.getChildren()
        return self.uuid_id

    def loadData(self):
        data = loadRepeatData(self.db, self.meta, self.structure_node.name,
                              self.identifier, load_uuid=True)
        self.uuid_id = data.pop("uuid_id")
        return data

    def lazyLoad(self):
        buildChildren(self.db, self.loadData(), self, self.structure_node)

    def getChildren(self):
        if self.children is None:
            self.lazyLoad()
        return self.children


class DBInstanceTree(AbstractInstanceTree):

    def __init__(self, formed, root=None, db=None, uncommited=None):
        AbstractInstanceTree.__init__(self, formed, root, uncommited)
        self.db = db

    def commit(self):
        real_changes = {}
        for key, value in self.new_values.iteritems():
            real_item = self.getItem(key)
            if not real_item:
                continue
            if real_item.getValue() != value:
                real_changes[real_item] = value
                real_item.setValue(value)

        self.new_values.clear()

        if not real_changes:
            return self.uncommited

        tables = [(relationAndId(k, "_view"), (k.getMeta().getName(), v))
                  for k, v in real_changes.iteritems()]
        tables.sort(lambda a, b: cmp(a[0], b[0]))

        updates = {}

        cur = None
        self.db.begin()
        try:
            try:
                cur = self.db.acquire_cursor()
                rel_id = None
                for r_id, kv in tables:
                    if r_id != rel_id:
                        if updates:
                            sets = u",".join(["%s = %%(%s)s" % (k, k)
                                              for k in updates.iterkeys()])
                            updates['id'] = rel_id[1]
                            stmnt = UPDATE_DATA % (rel_id[0], sets)
                            cur.execute(stmnt, updates)
                            updates.clear()
                        rel_id = r_id
                    updates[kv[0]] = kv[1]

                if updates and rel_id is not None:
                    sets = u",".join(["%s = %%(%s)s" % (k, k)
                                      for k in updates.iterkeys()])
                    updates['id'] = rel_id[1]
                    stmnt = UPDATE_DATA % (rel_id[0], sets)
                    cur.execute(stmnt, updates)

                self.db.commit()
            except:
                self.db.rollback()
                raise
        finally:
            self.db.release_cursor(cur)

        return self.uncommited

    def hasErrors(self, identifier=None):
        if identifier is None:
            return bool(self.uncommited)
        return identifier in self.uncommited

    def getErrors(self, identifier=None):
        if identifier is None:
            return self.uncommited
        return self.uncommited.get(identifier)

def relationAndId(instance, view=''):
    cur = instance
    dataset_id = relation = None
    while cur:
        if cur.isRepeat():
            dataset_id = cur.getInternalIdentifier()
        if isinstance(cur, DBRepeatInstanceContainerNode):
            relation = cur.relation
            if dataset_id is not None:
                break
        cur = cur.parent

    if relation is None:
        relation = "master_tbl"

    return (relation + view, dataset_id)


class LazyPageNode(PageInstanceNode):

    def __init__(self, meta, parent, structure_node, db):
        PageInstanceNode.__init__(self, meta)
        self.parent = parent
        self.structure_node = structure_node
        self.db = db

    def loadData(self):
        relation, dataset_id = relationAndId(self)
        return loadRepeatData(self.db, self.meta, relation, dataset_id)

    def lazyLoad(self):
        buildChildren(self.db, self.loadData(), self, self.structure_node)

    def getChildren(self):
        if self.children is None:
            self.lazyLoad()
        return self.children


class InstanceTreeBuilder(object):

    def __init__(self, identifier, case_structure, factory, uncommited):
        self.identifier = identifier
        self.case_structure = case_structure
        self.factory = factory
        self.uncommited = uncommited

    def recursive_build(self, formed_node, instance_node, instance_children,
                        structure_node):
        formed_children = formed_node.children

        for formed_child in formed_children:
            if isinstance(formed_child, Node):
                if isinstance(formed_child, RepeatNode):
                    rg_name = formed_child.getName().replace("-", "_") + "_tbl"
                    rgs = [c for c in structure_node.children
                             if c.name == rg_name]
                    rg_container_children = []
                    container_node = \
                          DBRepeatInstanceContainerNode(formed_child,
                                                        rg_container_children,
                                                        rg_name,
                                                        self.factory.db)

                    container_node.parent = instance_node
                    instance_children.append(container_node)

                    for rg in rgs:
                        rg_instance = LazyRepeatInstanceNode(formed_child,
                                                             None, rg.id,
                                                             self.factory.db,
                                                             rg)
                        rg_instance.parent = container_node
                        rg_container_children.append(rg_instance)
                elif isinstance(formed_child, PageNode):
                    lazyPage = LazyPageNode(formed_child, instance_node,
                                            structure_node, self.factory.db)
                    instance_children.append(lazyPage)
                else:
                    sub_children = []
                    sub_node = InstanceNode(formed_child, sub_children)
                    sub_node.parent = instance_node
                    instance_children.append(sub_node)

                    self.recursive_build(formed_child, sub_node, sub_children,
                                         structure_node)

    def build(self):
        children = []
        root_meta = self.factory.formed.getRoot().getCaseNode()
        instance_root = DBRootNode(root_meta, children, self.identifier)

        self.recursive_build(root_meta, instance_root, children,
                             self.case_structure)

        tree = self._get_instance_tree(instance_root)
        instance_root.tree = tree

        return tree

    def _get_instance_tree(self, instance_root):
        return DBInstanceTree(self.factory.formed, instance_root,
                              self.factory.db, self.uncommited)



class PageFilteredWidgetCollector(WidgetCollector):

    def __init__(self, page_filter):
        WidgetCollector.__init__(self)
        self.page_filter = page_filter

    def visitor(self, nc):
        if isinstance(nc, VirtualLeaf):
            # ignore VirtualLeafs
            return VISIT_CONTINUE
        if isinstance(nc, PageNode):
            if not self.page_filter(nc):
                return VISIT_IGNORE_CHILDREN
        return WidgetCollector.visitor(self, nc)

    def handleRepeatNode(self, repeat):
        # visit all repeats
        return VISIT_CONTINUE


class DBFactory(InstanceTreeFactory):

    def __init__(self, formed, db):
        InstanceTreeFactory.__init__(self, formed)
        self.db = db

    def createNewInstanceTree(self, uuid=None):
        identifier = None
        cursor = None
        self.db.begin()
        try:
            cursor = self.db.acquire_cursor()
            try:
                cursor.execute(CREATE_CASE, {'uuid': uuid})
                row = cursor.fetchone()
                if not row:
                    raise DBException("Cannot fetch new id of new case")
                identifier = row[0]
                self.db.commit()
            except:
                self.db.rollback()
                raise
        finally:
            self.db.release_cursor(cursor)

        return self.loadInstanceTreeByIdentifier(identifier)

    def loadInstanceTreeByIdentifier(self, identifier, uncommited=None):
        cursor = None
        try:
            cursor = self.db.acquire_cursor()
            case_structure = get_case_structure(cursor, identifier)
        finally:
            self.db.release_cursor(cursor)

        return self.build_tree(identifier, case_structure, uncommited)

    def build_tree(self, identifier, case_structure, uncommited=None):
        tree_builder = InstanceTreeBuilder(identifier, case_structure, self,
                                           uncommited)
        return tree_builder.build()


    def loadInstanceTreeByUUID(self, uuid, uncommited=None):
        cursor = None
        try:
            cursor = self.db.acquire_cursor()
            identifier = master_id_from_uuid(cursor, uuid)
        finally:
            self.db.release_cursor(cursor)

        if identifier is None:
            return None

        return self.loadInstanceTreeByIdentifier(identifier,
                                                 uncommited=uncommited)

    def toXML(self, identifiers, selection=None, anonymize=False,
              external=None):

        if selection is not None:
            selection = set(selection)

        collector = PageFilteredWidgetCollector(
            lambda p: (selection is None or p.getName() in selection
                       or p.isInvisible()))

        self.formed.visit(collector.visitor)
        widgets = collector.widgets

        needed_repeats = {'master': []}

        for widget in widgets:
            p = widget.parent
            rel = None
            while p:
                if isinstance(p, RepeatNode):
                    rel = p.getName().replace('-', '_')
                    break
                p = p.parent

            if not rel:
                rel = 'master'

            fields = needed_repeats.setdefault(rel, [])
            fields.append(widget.getName())


        for key in needed_repeats.keys():
            fields = needed_repeats[key]
            if not anonymize:
                fields.insert(0, "uuid_id")
            fields = ",".join(fields)
            relation = ((anonymize and "%s_tbl_anonym_view" or "%s_tbl_view")
                        % key)
            sql = SELECT_XML_FIELDS % (fields, relation)
            needed_repeats[key] = sql

        cases = ET.Element("cases")
        cases.attrib["version"] = LIBFORMED_XML_CONTAINER_VERSION
        cases.attrib["formed-type"] = self.formed.getVersion()

        encode = XML_ENCODERS.get

        cursor = None
        try:
            cursor = self.db.acquire_cursor()
            for identifier in identifiers:
                case = ET.Element("case")
                cases.append(case)

                case_structure = get_case_structure(cursor, identifier)

                stack = [(case, case_structure)]

                while stack:
                    parent_node, relation = stack.pop()
                    name = relation.name.replace("_tbl", "")
                    sql = needed_repeats.get(name)
                    if sql is not None:
                        node = ET.Element(name)
                        parent_node.append(node)
                        if sql:
                            cursor.execute(sql, {'dataset_id': relation.id})
                            row = cursor.fetchone()
                            if row:
                                description = cursor.description
                                for idx, data in enumerate(row):
                                    if data is None:
                                        continue
                                    field_node = ET.Element(description[idx][0])
                                    node.append(field_node)
                                    field_node.text = \
                                           encode(data.__class__,
                                                  lambda s: str(s))(data)

                    if relation.children:
                        stack.extend([(node, c)
                                      for c in reversed(relation.children)])

                if external:
                    external(case, case_structure.id, cursor, selection,
                             anonymize)
        finally:
            self.db.release_cursor(cursor)

        return ET.ElementTree(cases)

    def copy_instance_tree(self, src, fields=None):
        """Create a new instance tree as a copy of the instance tree src.
        """
        dest = self.createNewInstanceTree()
        data = dict()
        self._copy_data_recursively(src.getRootNode(), dest.getRootNode(), data, fields)
        dest.force_data(data)
        uncommitted = dest.commit()
        # FIXME: what should we do with uncommitted fields?
        return dest

    def _copy_data_recursively(self, src, dest, data, fields=None):
        src_children = src.getChildren()
        dest_children = dest.getChildren()
        if len(dest_children) < len(src_children):
            dest_children = dest_children + [None] * (len(src_children)
                                                      - len(dest_children))
        for sc, dc in zip(src_children, dest_children):
            assert dc is None or sc.getMeta() is dc.getMeta()
            if sc.isLeaf():
                if fields is not None:
                    if dc.getIdentifier().split(':')[0] in fields:
                        data[dc.getIdentifier()] = sc.getValue()
                else:
                    data[dc.getIdentifier()] = sc.getValue()
            else:
                if sc.isRepeat() and dc is None:
                    dc = dest.createInstance()
                self._copy_data_recursively(sc, dc, data, fields)


# vim:set ts=4 sw=4 si et sta sts=4 enc=utf-8 fenc=utf-8 :
