# -*- coding: utf-8 -*-
# Copyright 2010 Intevation GmbH, Germany, <info@intevation.de>
# Authors:
# Bernhard Herzog <bernhard.herzog@intevation.de>
#

import re
from datetime import date
from xml.sax.handler import ContentHandler, feature_external_ges, \
     feature_external_pes
from xml.sax import make_parser
import logging

import formed.meta.data as data
import formed.meta.semantic as semantic
from formed.meta.structure import StructureTree
from formed.instance.simplified import StructureInstanceNode, \
     StructureInstanceTree


log = logging.getLogger(__name__)


class XMLImportException(Exception):

    """Raised by the import code if there's a problem.
    Instances have the following public attributes:

      message -- A string with a description of the exception

      uuid -- A string with the uuid of the case whose import failed

      set_data_errors -- A dictionary with the error descriptions from a
                         SetDataException
    """

    def __init__(self, message, uuid=None, set_data_errors=()):
        Exception.__init__(self, message)
        self.uuid = uuid
        self.set_data_errors = set_data_errors


re_date = re.compile("([0-9]{1,4})-([0-9]{1,2})-([0-9]{1,2})")


def decode_int(s):
    return int(s)

def decode_numeric(s):
    return float(s)

def decode_date(s):
    m = re_date.match(s)
    if not m:
        raise ValueError("%r cannot be decoded as a date" % s)
    return date(int(m.group(1)), int(m.group(2)), int(m.group(3)))

def decode_text(s):
    return s

def decode_uuid(s):
    """Decode the uuid s.
    UUIDs are case insensitive in the XML content. In the database and
    the instance trees they're always stored in lower case. Also, the
    UUID might have leading and trailing whitespace, so strip that, too.
    """
    return s.strip().lower()


DECODERS = {
    data.BoolLeaf: decode_int,
    data.ChoiceNode: decode_int,
    data.DateLeaf: decode_date,
    data.IntLeaf: decode_int,
    data.NumericLeaf: decode_numeric,
    data.RadioNode: decode_int,
    data.TextAreaLeaf: decode_text,
    data.TextLeaf: decode_text,
    }


def decode(meta_node, value):
    decoder = DECODERS.get(meta_node.__class__)
    if decoder is not None:
        value = decoder(value)
    else:
        log.warning("decode: no decoder for %r:%r", meta_node, value)
    return semantic.check(meta_node, value)


class CasesHandler(ContentHandler):

    def __init__(self, xml_importer):
        ContentHandler.__init__(self)
        self.xml_importer = xml_importer
        self.cases = None
        self.case = None
        self.formed_type = None

    def startElement(self, name, attrs):
        if name == "cases":
            if self.cases is not None:
                raise XMLImportException("unexpected 'cases' element")
            if attrs.get("version") != "5":
                raise XMLImportException("Only version 5 of the cases"
                                         " element is supported")
            self.formed_type = attrs.get("formed-type")
            self.xml_importer.version_supported(self.formed_type)
            self.cases = []
        elif name == "case":
            if self.case is not None:
                raise XMLImportException("unexpected 'case' element")
            self.case = self.get_case_handler(self.formed_type)
        else:
            if self.case is not None:
                self.case.startElement(name, attrs)
            else:
                raise XMLImportException("unexpected element: %r" % name)

    def endElement(self, name):
        if name == "cases":
            pass
        elif name == "case":
            if self.case is None:
                raise XMLImportException("unexpected 'case' end-tag")
            else:
                self.cases.append(self.case.get_instance())
                self.case = None
        else:
            if self.case is not None:
                self.case.endElement(name)
            else:
                raise XMLImportException("unexpected end-tag: %r" % name)

    def characters(self, content):
        if self.case:
            self.case.characters(content)

    def get_case_handler(self, formed_type):
        return CaseHandler(self.xml_importer.case_factories)


class ComponentFactory(object):

    """Factories to create StructureTree based trees."""

    def __init__(self, meta_tree, meta_root_node, node_factory, tree_factory):
        self.meta_tree = meta_tree
        self.meta_root_node = meta_root_node
        self.node_factory = node_factory
        self.tree_factory = tree_factory


class CaseHandler(ContentHandler):

    def __init__(self, component_factories):
        """Initialize the CaseHandler.

        The component_factories parameter should be a dictionary mapping
        names of child elements of the case element to the corresponding
        ComponentFactory instance."""
        self.component_factories = component_factories
        self.components = dict()
        self.nodestack = []
        self.cur_meta = self.cur_leaves = self.cur_rgs = None
        self.text_content = None
        self.factory = None

    def get_instance(self):
        """Return the instance read from the XML file.
        The return value is a dictionary mapping the keys from the
        component_factories dictionary given to the constructor to the
        instance tree created for that component if it actually was
        present in the XML file.
        """
        return self.components

    def push_node(self, node):
        self.cur_meta = node
        self.cur_leaves = dict()
        self.cur_rgs = dict()
        self.nodestack.append((self.cur_meta, self.cur_leaves, self.cur_rgs))

    def pop_node(self):
        node, leaves, rgs = self.nodestack.pop()
        for leafname in node.leaves:
            if leafname not in leaves:
                leaves[leafname] = None
        for rgname in node.repeat_groups:
            if rgname not in rgs:
                rgs[rgname] = dict()
        if self.nodestack:
            self.cur_meta, self.cur_leaves, self.cur_rgs = self.nodestack[-1]
        else:
            self.cur_meta = self.cur_leaves = self.cur_rgs = None
        return self.create_instance_node(node, leaves, rgs)

    def startElement(self, name, attrs):
        if name in self.component_factories:
            self.factory = self.component_factories[name]
            if self.nodestack:
                raise XMLImportException("Unexpected element %r" % name)
            self.push_node(self.factory.meta_root_node)
        else:
            if name in self.cur_meta.leaves or name == "uuid_id":
                self.text_content = []
            elif name in self.cur_meta.repeat_groups:
                if self.text_content is not None:
                    raise XMLImportException("Unexpected element: %r" % name)
                self.push_node(self.cur_meta.repeat_groups[name])
            else:
                raise XMLImportException("Unexpected element: %r" % name)

    def endElement(self, name):
        if name in self.component_factories:
            assert(self.component_factories[name] is self.factory)
            if len(self.nodestack) != 1:
                raise XMLImportException("Unexpected element %r" % name)
            node = self.pop_node()
            if not node.meta_node.is_root():
                raise XMLImportException("node %r is not root node" % name)
            self.components[name] = \
                     self.factory.tree_factory(self.factory.meta_tree, node)
        else:
            if name in self.cur_meta.leaves:
                leaf = self.cur_meta.leaves[name]
                # ignore virtualleafs which are calculated automatically
                if isinstance(leaf, data.VirtualLeaf):
                    return
                try:
                    self.cur_leaves[name] = decode(leaf,
                                                   self.get_text_content())
                except ValueError, e:
                    raise XMLImportException("Field %s contains an invalid " \
                                             "value. Details %s" % (name, e))
            elif name == "uuid_id":
                self.cur_leaves[name] = decode_uuid(self.get_text_content())
            elif name == self.cur_meta.name:
                if self.text_content is not None:
                    raise XMLImportException("Unexpected element: %r" % name)
                node = self.pop_node()
                self.cur_rgs.setdefault(name, []).append(node)
            else:
                raise XMLImportException("Unexpected element: %r" % name)

    def create_instance_node(self, structure_node, leaves, repeat_groups):
        # uuids are optional and only present if cases have not yet been
        # anonymized, either by anonymizing the case in the database or
        # by exporting an anonymized version.
        uuid = leaves.pop("uuid_id", None)
        return self.factory.node_factory(structure_node, None, uuid, leaves,
                                         repeat_groups)

    def get_text_content(self):
        content = "".join(self.text_content)
        self.text_content = None
        return content

    def characters(self, content):
        if self.text_content is not None:
            self.text_content.append(content)


class XMLImporter(object):

    def __init__(self, meta_tree, extra_factories=None):
        """Initialize the importer.
        The meta_tree parameter should be the formed meta tree matching
        the data to be read.
        """
        self.structure_tree = StructureTree(meta_tree)
        factories = {"master": ComponentFactory(self.structure_tree,
                                                self.structure_tree.root,
                                                StructureInstanceNode,
                                                StructureInstanceTree)}
        if extra_factories is not None:
            factories.update(extra_factories)
        self.case_factories = factories

    def get_cases_handler(self):
        return CasesHandler(self)

    def read_cases_from_xml(self, infile, chunksize=16*1024):
        """Read cases from the infile and yield each once it has been read.
        """
        handler = self.get_cases_handler()
        parser = make_parser()
        parser.setFeature(feature_external_ges, False)
        parser.setFeature(feature_external_pes, False)
        parser.setContentHandler(handler)

        while 1:
            chunk = infile.read(chunksize)
            if not chunk:
                break
            parser.feed(chunk)
            while handler.cases:
                yield handler.cases.pop(0)

    def version_supported(self, version):
        """Return whether the given version of the formed tree is supported.
        Derived class should override this to check the version and
        raise an XMLImportException if the version is not acceptable.
        """
        return True
