#!/usr/bin/env python
# -*- coding: latin-1 -*-
#
# Copyright 2007, 2008, 2009 Intevation GmbH, Germany, <info@intevation.de>
#
# This file is part of mpuls WASKA (CoMPUter-based case fiLeS -
# Web-Anwendungs-Server fuer Kompetenzagenturen).
#
# mpuls WASKA is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# mpuls WASKA is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public
# License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with mpuls WASKA. If not, see <http://www.gnu.org/licenses/>.
#
# mpuls WASKA has been developed on behalf of the
# Projekttraeger im Deutschen Zentrum fuer Luft- und Raumfahrt e.V. (PT-DLR)
# within the programme Kompetenzagenturen (Durchfuehrungsphase) funded by
# the Bundesministerium fuer Familie, Senioren, Frauen und Jugend and
# European Social Fund resources.
#
# Authors:
# Sascha L. Teichmann <teichmann@intevation.de>
#

import sys
import re
import codecs
import traceback
from types import StringTypes
from datetime import date, timedelta
# from shlex import split as lexsplit
from decimal import Decimal
import logging

from formed.meta import UNKNOWN_DATE, UNKNOWN_INT, UNKNOWN_STR
from formed.meta.lex import split


log = logging.getLogger(__name__)


FLT_RE = re.compile(r'^([-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)$')
INT_RE = re.compile(r'^([-+]?[0-9]+)$')
VAR_RE = re.compile(r'^\$(\w+)$')
ALL_VAR_RE = re.compile(r'^%(\w+)$')
STRING_RE = re.compile(r'^(\'(.*)\')|("(.*)")$')


class Node(object):

    def __init__(self):
        pass

    def eval(self, ctx):
        return None


class Unary(Node):

    def __init__(self, a=None):
        Node.__init__(self)
        self.a = a

    def __repr__(self):
        return "%s(%s)" % (self.__class__.__name__, repr(self.a))


class List(Node):

    def __init__(self, items):
        self.items = items

    def eval(self, ctx):
        return [item.eval(ctx) for item in self.items]

    def __repr__(self):
        return "List(%r)" % (self.items,)


class Binary(Unary):

    def __init__(self, a=None, b=None):
        Unary.__init__(self, a)
        self.b = b

    def __repr__(self):
        return "%s(%s, %s)" % (self.__class__.__name__, repr(self.a),
                               repr(self.b))


class OperatorUnary(Unary):

    def __init__(self, a, operator):
        Unary.__init__(self, a)
        self.operator = operator

    def eval(self, ctx):
        return self.operator(self.a.eval(ctx))

    def __repr__(self):
        return "%s[%s](%s)" % (self.__class__.__name__, self.operator.__name__,
                               repr(self.a))


class OperatorBinary(Binary):

    def __init__(self, a, b, operator):
        Binary.__init__(self, a, b)
        self.operator = operator

    def eval(self, ctx):
        return self.operator(self.a.eval(ctx), self.b.eval(ctx))

    def __repr__(self):
        return "%s[%s](%s, %s)" % (self.__class__.__name__,
                                   self.operator.__name__,
                                   repr(self.a), repr(self.b))


class NumericOperator(OperatorBinary):

    def eval(self, ctx):
        valid_types = (int, float, Decimal)
        a = self.a.eval(ctx)
        b = self.b.eval(ctx)
        if not isinstance(a, valid_types) or not isinstance(b, valid_types):
            raise ValueError("Invalid type of arguments %s with value %r " \
                             "and %s with value %r for operator %s" % (self.a,
                                 a, self.b, b, self.operator.__name__))

        return self.operator(a, b)


class Not(Unary):

    def eval(self, ctx):
        return not self.a.eval(ctx)


class And(Binary):

    def eval(self, ctx):
        ea = self.a.eval(ctx)
        if not ea:
            return False
        return bool(self.b.eval(ctx))


class Or(Binary):

    def eval(self, ctx):
        return self.a.eval(ctx) and True or bool(self.b.eval(ctx))


class Var(Node):

    def __init__(self, var_name):
        self.var_name = var_name

    def eval(self, ctx):
        return ctx.get_var(self.var_name)

    def __repr__(self):
        return "var(%s)" % (self.var_name)


class Const(Node):

    def __init__(self, const):
        self.const = const

    def eval(self, ctx):
        return self.const

    def __repr__(self):
        return "const(%s)" % repr(self.const)

class String(Node):

    def __init__(self, str):
        self.str = str

    def eval(self, ctx):
        return str(self)

    def __str__(self):
        return self.str[1:-1]

    def __repr__(self):
        return "string(%s)" % repr(self.str)

class Today(Node):

    def eval(self, ctx):
        return date.today()


class DateNode(Node):

    def __init__(self, year, month, day):
        self.day = day
        self.month = month
        self.year = year

    def eval(self, ctx):
        day = self.day.eval(ctx)
        month = self.month.eval(ctx)
        year = self.year.eval(ctx)
        return date(year, month, day)


class Out(Node):

    def __init__(self, parent):
        self.parent = parent

    def eval(self, ctx):
        e = self.parent.eval(ctx)
        print >> sys.stderr, "OUT: %s" % repr(e)
        return e


class IThenElse(Node):

    def __init__(self, q, a, b):
        self.q = q
        self.a = a
        self.b = b

    def eval(self, ctx):
        if self.q.eval(ctx):
            return self.a.eval(ctx)
        return self.b.eval(ctx)


class RGMap(Binary):

    def eval(self, ctx):
        rg_instances = ctx.get_rg_instances(self.a.eval(ctx))
        result = []
        for rg_inst in rg_instances:
            rgmapctx = RGMapEvalContext(ctx, rg_inst)
            result.append(self.b.eval(rgmapctx))
        return result


def ADD(a, b):   return a + b
def MINUS(a, b): return a - b
def MUL(a, b):   return a * b
def DIV(a, b):   return a / b
def MOD(a, b):   return a % b
def POW(a, b):   return a ** b
def EQ(a, b):    return a == b
def NE(a, b):    return a != b
def GT(a, b):    return a > b
def LT(a, b):    return a < b
def LE(a, b):    return a <= b
def GE(a, b):    return a >= b
def MIN(a, b):   return min(a, b)
def MAX(a, b):   return max(a, b)
def INT(a):      return int(a)
def FLOAT(a):    return float(a)
def STR(a):      return str(a)
def LEN(a):      return len(a)
def ABS(a):      return abs(a)
def NOT(a):      return not a
def ISSET(a):    return not a is None
def TYPE(a):     return type(a)
def CONTAINS(a, b): return a in b

def KNOWN(a):
    return not (a == UNKNOWN_INT or a == UNKNOWN_DATE
                or (type(a) in StringTypes and a.lower() == UNKNOWN_STR))

def YEARS(start, end):
    """Return the number of complete years between the dates start and end.
    This is useful for e.g. age calculations.  The order of the dates is
    irrelevant.
    """
    if end < start:
        start, end = end, start
    years = end.year - start.year
    if end.month < start.month or (end.month == start.month
                                   and end.day < start.day):
        years -= 1
    return years

def MONTH(start, end):
    """Return the number of month between the dates start and end."""
    if end < start:
        start, end = end, start
    month = end.month - start.month
    years = end.year - start.year
    month += years * 12
    if end.day < start.day:
        month -= 1
    return month

def DAYS(start, end):
    """Return the number of days years between the dates start and end.
    """
    if end < start:
        start, end = end, start
    td = end - start
    return td.days 


# special object that marks the end of a list on the stack
MARK = object()

def pop_to_mark(s):
    """Return all items on the stack obove the top-most MARK as a list.
    If the stack does not contain the MARK, the function raises a
    ValueError.
    """
    items = []
    while s:
        item = s.pop()
        if item is MARK:
            break
        items.append(item)
    else:
        raise ValueError("No MARK ('[') found on stack")
    items.reverse()
    return items

def list_sum(s):
    """Calculate the sum of a list e.g. [ 1 2 3 ] sum.
    If s is not a list a ValueError is raised.
    """
    if not isinstance(s, list):
        raise ValueError("Argument '%s' is not a list" % s)
    return sum(s)

def coalesce(a, b):
    if a is None:
        return b
    return a

NODE_FACTORIES = {
    '+':     lambda s: NumericOperator(s.pop(),   s.pop(), ADD),
    '-':     lambda s: OperatorBinary(s.pop(-2), s.pop(), MINUS),
    '*':     lambda s: NumericOperator(s.pop(),   s.pop(), MUL),
    '/':     lambda s: OperatorBinary(s.pop(-2), s.pop(), DIV),
    '%':     lambda s: OperatorBinary(s.pop(-2), s.pop(), MOD),
    '**':    lambda s: OperatorBinary(s.pop(-2), s.pop(), POW),
    '==':    lambda s: OperatorBinary(s.pop(),   s.pop(), EQ),
    '!=':    lambda s: OperatorBinary(s.pop(),   s.pop(), NE),
    '>':     lambda s: OperatorBinary(s.pop(-2), s.pop(), GT),
    '<':     lambda s: OperatorBinary(s.pop(-2), s.pop(), LT),
    '<=':    lambda s: OperatorBinary(s.pop(-2), s.pop(), LE),
    '>=':    lambda s: OperatorBinary(s.pop(-2), s.pop(), GE),
    'min':   lambda s: OperatorBinary(s.pop(), s.pop(),   MIN),
    'max':   lambda s: OperatorBinary(s.pop(), s.pop(),   MAX),
    'years': lambda s: OperatorBinary(s.pop(), s.pop(),   YEARS),
    'months':lambda s: OperatorBinary(s.pop(), s.pop(),   MONTH),
    'days':  lambda s: OperatorBinary(s.pop(), s.pop(),   DAYS),
    'int':   lambda s: OperatorUnary(s.pop(),             INT),
    'float': lambda s: OperatorUnary(s.pop(),             FLOAT),
    'str':   lambda s: OperatorUnary(s.pop(),             STR),
    'len':   lambda s: OperatorUnary(s.pop(),             LEN),
    'abs':   lambda s: OperatorUnary(s.pop(),             ABS),
    'not':   lambda s: OperatorUnary(s.pop(),             NOT),
    'isset': lambda s: OperatorUnary(s.pop(),             ISSET),
    'known': lambda s: OperatorUnary(s.pop(),             KNOWN),
    'type':  lambda s: OperatorUnary(s.pop(),             TYPE),
    'and':   lambda s: And(s.pop(-2), s.pop()),
    'or':    lambda s: Or(s.pop(-2), s.pop()),
    'today': lambda s: Today(),
    'date':  lambda s: DateNode(s.pop(), s.pop(), s.pop()),
    '.':     lambda s: Out(s.pop()),
    '?':     lambda s: IThenElse(s.pop(), s.pop(), s.pop()),
    '[':     lambda s: MARK,
    ']':     lambda s: List(pop_to_mark(s)),
    'contains':    lambda s: OperatorBinary(s.pop(), s.pop(), CONTAINS),
    'sum': lambda s: OperatorUnary(s.pop(), list_sum),
    'rgmap': lambda s: RGMap(s.pop(), s.pop()),
    'coalesce' : lambda s: OperatorBinary(s.pop(-2), s.pop(), coalesce),
    'true' : lambda s: Const(True),
    'false' : lambda s: Const(False),
}


class EvalContext(object):

    def __init__(self, var_provider):
        self.vars = dict()
        self.extra_vars = list()
        self.for_all = None
        self.var_provider = var_provider
        self.rgs = dict()

    def get_var(self, var):
        if var in self.vars:
            return self.vars[var][-1]
        try:
            return self.for_all[var][-1]
        except KeyError:
            pass
        value = self.var_provider.getValueForVar(var)
        self.vars[var] = value
        return value[-1]

    def add_extra_var(self, value):
        self.extra_vars.append(value)

    def get_all_vars(self):
        all_vars = self.vars.values()
        if self.for_all:
            all_vars.extend(self.for_all.itervalues())
        all_vars.extend(self.extra_vars)
        return all_vars

    def get_rg_instances(self, rg_name):
        rgs = []
        try:
            rgs = self.rgs[rg_name]
        except KeyError:
            rgs = self.var_provider.get_repeat_group_instances(rg_name)
            self.rgs[rg_name] = rgs
        return rgs


class RGMapEvalContext(object):

    def __init__(self, ctx, rg_instance):
        self.ctx = ctx
        self.rg_instance = rg_instance

    def get_var(self, var):
        value = self.ctx.var_provider.getValueForVar(var, self.rg_instance)
        self.ctx.add_extra_var(value)
        return value[-1]


class VariableProvider(object):

    def getValueForVar(self, var, reference=None):
        """Return id and value of the variable var.
        reference can be a RG instance identifier. In that case
        the value of var within the RG instance is returned.
        The return value is a tuple of the form (ID, VALUE) where ID is
        the item's id in the instance tree and VALUE is the value of the
        item.
        """
        return (var, None)

    def getAllValuesForVar(self, var):
        """Return all values for the variable.
        The return value is a sequence of the items returned by
        getValueForVar.  If the are no values for the variable, the
        return value is an empty sequence.
        """
        return ()

    def get_repeat_group_instances(self, name):
        """Return a sequence representing the instances of a repeat group.
        The parameter name should be the name of the repeat gorup.

        The return value is a sequence of objects each representing an
        instance of the named repeat group. The items are usable as the
        reference parameter for the getValueForVar method. Usually this
        means that the items will be the instance identifieres of the
        repeat groups.

        If name does not refer to a repeat group or the repeat group is
        not an immediate child of the repeat group relative to which the
        name is looked up, a ValueError is raised.
        """
        raise NotImplementedError


def product(args):
    result = [[]]
    for pool in args:
        result = [x+[y] for x in result for y in pool]
    for prod in result:
        yield prod


class Expr(object):

    def __init__(self, expr):
        self.expr = expr
        self.prog = None
        self.depends = None
        self.for_all = None

    def evaluate_instances(self, var_provider, exit_early=False):
        """Evaluate expression and return results for all instances.

        Expressions are defined with the names of the formed fields in
        the meta tree.  However, they are evaluated with values taken
        from an instance tree.  For many fields there is a one-to-one
        correspondence between meta tree fields and instance fields, but
        there is a one-to-many correspondence when repeat groups are
        involved.  Expressions that refer to fields in repeat groups
        have to be evaluated multiple times, once for each repeat group
        instance (and sometimes even more often when %-field references
        are used).  This is equivalent to assuming that every expression
        for the meta tree is actually instantiated multiple times with
        references to the various combinations of instance tree items it
        applies to.  This method evaluates the expression once for each
        such expression instance.

        The return value is a list of pairs of the form (result,
        variables), where result is the result of evaluating the
        expression for the variables given in the dictionary variables.
        The variables dictionary maps the identifiers of the instance
        tree items the expressions was evaluated with to the
        corresponding values.

        If exit_early is true (default False), exit as soon as an
        instance of the expression evaluates to false and return the
        list of results obtained so far.  The result that caused the
        early exit is the last item in that list.
        """
        if self.prog is None:
            self.compile()

        # fetch 'for all' variables
        get_value = var_provider.getAllValuesForVar
        for_all = [([(var, item) for item in get_value(var)])
                   for var in self.for_all]
        ctx = EvalContext(var_provider)

        all_results = []

        # do cross product
        eval = self.prog.eval
        for prod in product(for_all):
            ctx.for_all = dict(prod)
            try:
                result = eval(ctx)
                all_results.append((result, dict(ctx.get_all_vars())))
                if not result and exit_early:
                    return all_results
            except:
                log.exception("Exception while evaluating expr: %r",self.prog)
                raise
        return all_results

    def evaluate(self, var_provider):
        results = self.evaluate_instances(var_provider, exit_early=True)
        return results[-1][0]

    def compile(self):
        depends = set()
        for_all = set()

        stack = []

        # XXX: shlex does not support unicode -> ascii
        encoder = codecs.getencoder("ascii")
        for token in split(encoder(self.expr)[0]):
            token = unicode(token)
            m = ALL_VAR_RE.match(token)
            if m:
                var = m.group(1)
                stack.append(Var(var))
                for_all.add(var)
                depends.add(var)
                continue
            m = VAR_RE.match(token)
            if m:
                var = m.group(1)
                stack.append(Var(var))
                depends.add(var)
                continue
            m = INT_RE.match(token)
            if m:
                stack.append(Const(int(m.group(1))))
                continue
            m = FLT_RE.match(token)
            if m:
                stack.append(Const(float(m.group(1))))
                continue
            m = STRING_RE.match(token)
            if m:
                stack.append(String(m.group(0)))
                continue
            try:
                stack.append(NODE_FACTORIES[token](stack))
            except KeyError:
                raise ValueError("Unkown token %r in Expr %r" % (token,
                                 self.expr))

        # At this point the expression must have been reduced to exactly
        # one compiled expression object on the stack.  Otherwise, it is
        # likely that the expression contains errors such as a
        # misspelled operator name or a missing boolean operator.
        if len(stack) != 1:
            raise ValueError("Expr %r does not reduce to a single stack item"
                             % self.expr)
        self.prog = stack[0]
        self.depends = depends
        self.for_all = for_all

    def getDependencies(self):
        if self.depends is None:
            self.compile()
        return self.depends



#def main():
#    for arg in sys.argv[1:]:
#        expr = Expr(arg)
#        vars = { 'a': 3 }
#        result = expr.evaluate(vars)
#        print str(result)
#
#if __name__ == "__main__":
#    main()

# vim:set ts=4 sw=4 si et sta sts=4:
