#!/bin/sh

usage()
{
cat << EOF
usage: $0 options source

This script builds a release tarball from given source directory.

OPTIONS:
   -h       Show this message
   -n       Name of the release (Defaults to name of archived source-revision)
   -o       Builddir (Default /tmp)
   -r       Revision number used to build the release (Default: tip of working directory)
   -c       Build release as tar.gz file (Default: False)
EOF
}

# DEFAULTS
VERBOSE=0
NAME=
BUILDDIR=/tmp
REVISION=
COMPRESS=0

# Parse Options
while getopts "hcn:o:r:" OPTION
do
        case $OPTION in
            h)
                usage
                exit 1
                ;;
            n)
                NAME=$OPTARG
                ;;
            v)
                VERBOSE=1
                ;;
            o)
                BUILDDIR=$OPTARG
                ;;
            r)
                REVISION=$OPTARG
                ;;
            c)
                COMPRESS=1
                ;;
            ?)
                # Will be set if the optstring begins with a ":". In this case
                # all options not defined in the optstring will OPTION set to
                # "?". Otherwise the shell reports an error on the unknown
                # option.
                usage
                exit
                ;;
        esac
done

# Forget about options. arguments now begin a position 1
shift $(($OPTIND - 1))
SOURCE=${1}

# Check Arguments
if [[ -z $SOURCE ]]
then
    echo "Error! Argument required!";
    usage;
    exit 1;
fi

# Check if source dir exists
if [ ! -d $SOURCE ]
then
    echo "Error! Can not find source dir!";
    usage;
    exit 1;
else
    DIR=`readlink -f ${SOURCE}`
fi

# Enter source dir and get revision.
cd ${SOURCE}
if [ ! -z ${REVISION} ]
then
    echo "Updating to revision ${REVISION}";
    hg update -r ${REVISION};
    if [ ! $? -eq 0 ]; then
        exit $?
    fi
else
    REVISION=`hg tip| head -n 1|cut -d " " -f4|cut -d: -f1`
fi

# Build releasename
if [ -z $NAME ]; then
    echo "Error! Please provide name of the distribution!";
    exit 1;
fi
RELEASENAME=${NAME}-${REVISION}

# Archive
hg archive -S -t files ${BUILDDIR}/${RELEASENAME}
# Finally update to the last revision again
hg update

# Do translations
cd ${BUILDDIR}/${RELEASENAME}
cd libformed
python setup.py compile_catalog
cd ..

cd application
for f in $(find . -name \*.po); 
    do msgfmt ${f} -o $(echo ${f} | sed -e "s/\.po$/.mo/g"); 
done
cd ..

cd base
python setup.py compile_catalog
DATE=$(date +%d.%m.%Y)
sed -i -e s/DISTRIBUTION/${NAME}/g mpulsweb/lib/config.py
sed -i -e s/REVISION/${REVISION}/g mpulsweb/lib/config.py
sed -i -e s/RELEASEDATE/${DATE}/g mpulsweb/lib/config.py
cd ..

# Compress release
if [ ${COMPRESS} -eq 1 ]
then
    echo "Compressing archive"
    cd ${BUILDDIR}
    tar czf ${RELEASENAME}.tar.gz ${RELEASENAME}
    RELEASENAME=${RELEASENAME}.tar.gz
fi

echo "Archive build in ${BUILDDIR}/${RELEASENAME}"
